/**
Action Script library. Button class implementation.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_button.hh"
#include <X11/cursorfont.h>

static funcPtr callbacks[]= { 
	&ASButton::init,		// 0
	&ASButton::getDepth,	// 1
	&ASButton::onDragOut,	// 2
	&ASButton::onDragOver,	// 3
	&ASButton::onKeyUp,		// 4
	&ASButton::onKillFocus,	// 5
	&ASButton::onPress,		// 6
	&ASButton::onRelease,	// 7
	&ASButton::onReleaseOutside,// 8
	&ASButton::onRollOut,	// 9
	&ASButton::onRollOver,	// 10
	&ASButton::onSetFocus,	// 11
	// Need to be at the end.
	&ASButton::addProperty,		
	&ASButton::registerClass,	
	&ASButton::AStoString,		
	&ASButton::unwatch,			
	&ASButton::valueOf,			
	&ASButton::watch		
};
extern int newX11Cursor;		// The cursor shape

/**
 * The constructor
 **/
ASButton::ASButton(Script *s, Graphics *g) : ASObject(s,g,L"Button") {
	selectedMovie = L"";
	hitBounds = new StringRectMap();
}

ASButton::ASButton(ASButton *p) : ASObject(p) {
	selectedMovie = L"";
	hitBounds = new StringRectMap();
}

/**
 * The destructor
 **/
ASButton::~ASButton() {
	activeCharacters.clear();
	hitBounds->clear();
	delete hitBounds;
	shapes.clear();
}

/**
 * Deep copy constructor
 **/
ASInterface* ASButton::getObjectDeepCopy(bool deep) {
	ASButton *newObj = new ASButton(this);
	newObj->tag = tag;
	newObj->buttonState = buttonState;
	//newObj->hitBounds->clear(); // = hitBounds;
	newObj->activeCharacters = activeCharacters;
	newObj->invMatrixes.clear(); // = invMatrixes;
	newObj->selectedMovie = selectedMovie;
	newObj->shapes = shapes;
	newObj->rebuildBounds();
	return newObj;
}

/**
 * Return a function descriptor
 */
Function *ASButton::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"Button") 		{func->setFunctionIndex(0); func->setNumParams(0); }
	if (memberName == L"getDepth") 		{func->setFunctionIndex(1); func->setNumParams(0); }
	if (memberName == L"onDragOut") 	{func->setFunctionIndex(2); func->setNumParams(1); }
	if (memberName == L"onDragOver") 	{func->setFunctionIndex(3); func->setNumParams(1); }
	if (memberName == L"onKeyUp") 		{func->setFunctionIndex(4); func->setNumParams(1); }
	if (memberName == L"onKillFocus") 	{func->setFunctionIndex(5); func->setNumParams(1); }
	if (memberName == L"onPress") 		{func->setFunctionIndex(6); func->setNumParams(1); }
	if (memberName == L"onRelease") 	{func->setFunctionIndex(7); func->setNumParams(1); }
	if (memberName == L"onReleaseOutside"){func->setFunctionIndex(8); func->setNumParams(1); }
	if (memberName == L"onRollOut") 	{func->setFunctionIndex(9); func->setNumParams(1); }
	if (memberName == L"onRollOver") 	{func->setFunctionIndex(10); func->setNumParams(1); }
	if (memberName == L"onSetFocus") 	{func->setFunctionIndex(11); func->setNumParams(1); }
	func = finishFunctionSetup(memberName,12,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASButton::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 *	Button init, nothing to do so far
 **/
void ASButton::init() {
}

/**
 * Returns the depth of a button instance.
 **/
void ASButton::getDepth() {
	puts("TODO");
	script->push(new IntegerObject(0)); // FIXME!
}

/**
 * Contruct a button from a defineButton Tag
 **/
void ASButton::setButtton(DefineButton2Tag *t) {
	this->tag = t;
	setMemberNoClone(L"trackAsMenu",new BooleanObject(t->getTrackAsMenu()));
	buttonState = buttonUp;
	rebuildBounds();
	needRedraw = true;
}

void ASButton::rebuildBounds() {
	hitBounds->clear();
	// Register button active area.
	for (list<BUTTONRECORD*>::iterator it = tag->getCharacters()->begin(); it != tag->getCharacters()->end(); it++) {
		BUTTONRECORD *br = *it;
		wstring movieName = encodeCharacterName(br->characterID);
		ASInterface *amc = getChildByName(movieName);
		if (amc == NULL) {
			amc = (ASInterface *)getObjectStore()->getGlobalVariable(movieName);
			getObjectStore()->setSpriteVariable(this,movieName,amc);
		}
		if (amc->getIsReferenced() ) amc = amc->getObjectDeepCopy(false);
		shapes[br->characterID] = amc;
		if (br->buttonStateHitTest) {
			(*hitBounds)[movieName] = *amc->getBounds();
		}
	}
}

/**
 * Convert to a button to graphic instructions. 
 **/
void ASButton::toGraphic() {
	if (needRedraw) {
		// Look for the shape corresponding to the button state
		int shapeId = 0;
		activeCharacters.clear();
		for (list<BUTTONRECORD*>::iterator it = tag->getCharacters()->begin(); it != tag->getCharacters()->end(); it++) {
			BUTTONRECORD *br = *it;
			switch (buttonState) {
				case buttonUp:	if (br->buttonStateUp) 	{shapeId = br->characterID; activeCharacters.push_back(br);} break;
				case buttonOver:if (br->buttonStateOver){shapeId = br->characterID; activeCharacters.push_back(br);} break;
				case buttonDown:if (br->buttonStateDown){shapeId = br->characterID; activeCharacters.push_back(br);} break;
			}
			if (br->buttonStateHitTest) activeCharacters.push_back(br); // Insert it since we need the full transform matrix to convert user click back
		}
		int i=0;
		for (list<BUTTONRECORD*>::iterator it = activeCharacters.begin() ; it != activeCharacters.end(); it++, i++) {
			BUTTONRECORD* b= *it;
			ASInterface *amc = shapes[b->characterID];
			if (b->placeDepth) {
				amc->setTransformMatrix(&(b->placeMatrix));
				amc->setHasTransformMatrix(true);
			} else {
				amc->setHasTransformMatrix(false);
			}
			if (b->hasColorTransform) {
				amc->setCxFormWithAlpha(&(b->colorTransform));
				amc->setHasCxFormWithAlpha(true);
			} else {
				amc->setHasCxFormWithAlpha(false);
			}
		}
	}
	needRedraw = false;
}

/**
 * Button event callback
 **/
void ASButton::callEvent(string eventName) {
	puts("ASButton::callEvent");
}

/**
 * Button event call back - internal call
 **/
bool ASButton::callInternalEvent(string eventName, void *data) {
	//puts("ASButton::callInternalEvent");
	funcPtr f = eventDispatcher(eventName);
	if (f != NULL) (this->*f)();
	return checkNeedRedraw();
}

// CallBacks
/**
 * Invoked when the mouse button is pressed over the button and the pointer then rolls outside the button.
 **/
void ASButton::onDragOut() {};

/**
 * Invoked when the user presses and drags the mouse button outside and then over the button.
 **/
void ASButton::onDragOver(){
};

/**
 * Invoked when a key is released.
 **/
void ASButton::onKeyUp(){};

/**
 * Invoked when focus is removed from a button.
 **/
void ASButton::onKillFocus(){};

/**
 * Invoked when the mouse is pressed while the pointer is over a button.
 **/
void ASButton::onPress(){
	puts("ASButton::onPress EVENT");
	processActions(buttonDown);
	buttonState = buttonDown;
	needRedraw = true;
};

/**
 * Invoked when the mouse is released while the pointer is over a button.
 **/
void ASButton::onRelease(){
	puts("ASButton::onRelease EVENT");
	processActions(buttonUp);
	buttonState = buttonUp;
	needRedraw = true;
};

/**
 * Invoked when the mouse is released while the pointer is outside the button after the button is pressed 
 * while the pointer is inside the button.
 **/
void ASButton::onReleaseOutside(){};

/**
 * Invoked when the pointer rolls outside of a button area.
 **/
void ASButton::onRollOut(){
	if (buttonState != buttonUp) {
		processActions(buttonUp);
		buttonState = buttonUp;
		newX11Cursor = XC_left_ptr;
		needRedraw = true;
	}
};

/**
 * Invoked when the mouse pointer rolls over a button.
 **/
void ASButton::onRollOver(){
	if (buttonState != buttonOver) {
		processActions(buttonOver);
		buttonState = buttonOver;
		newX11Cursor = XC_hand2;
		needRedraw = true;
	}
};

/**
 * Invoked when a button has input focus and a key is released.
 **/
void ASButton::onSetFocus(){};


/**
 * Handler event dispatcher
 **/
funcPtr	ASButton::eventDispatcher(string eventName) {
	if (eventName == "onDragOut") 	return &ASButton::onDragOut;
	if (eventName == "onDragOver") 	return &ASButton::onDragOver;
	if (eventName == "onKeyUp") 	return &ASButton::onKeyUp;
	if (eventName == "onKillFocus") return &ASButton::onKillFocus;
	if (eventName == "onPress") 	return &ASButton::onPress;
	if (eventName == "onRelease") 	return &ASButton::onRelease;
	if (eventName == "onReleaseOutside") return &ASButton::onReleaseOutside;
	if (eventName == "onRollOut") 	return &ASButton::onRollOut;
	if (eventName == "onRollOver") 	return &ASButton::onRollOver;
	if (eventName == "onSetFocus") 	return &ASButton::onSetFocus;
	return NULL;
}

/**
 * Process button transition logic
 **/
void ASButton::processActions(int newState) {
	printf("buttonDown:%d/%d  ButtonID=%d\n",buttonState,newState,tag->getButtonId());
///????	ASInterface *amc = getMovieFromButtonState(buttonState);
//wcout << amc->getMovieName() << L"\n";
	script->setCurrentContext((ScriptObject*)getObjectStore()->getGlobalVariable(L"_global"));
	for (list<BUTTONCONDACTION*>::iterator it=tag->getActions()->begin();it != tag->getActions()->end(); it++) {
		BUTTONCONDACTION *bc = *it;
		if (buttonState == 	buttonUp) {
			if (newState == buttonDown) {
				if (bc->getCondOverUpToOverDown()) execActionScript(bc->getCode());
			} 
		}	
		if (buttonState == 	buttonOver) {
			if (newState == buttonDown) {
				if (bc->getCondOverUpToOverDown()) execActionScript(bc->getCode());
			} 
		}	
		if (buttonState == 	buttonDown) {
			if (newState == buttonUp) {
				if (bc->getCondOverDownToOverUp())	execActionScript(bc->getCode());
			} 
		}
/*
	TODO:        ----------- IMPLEMENT missing transitions -----------

CondIdleToOverDown
CondOutDownToIdle
CondOutDownToOverDown
CondOverDownToOutDow n
CondOverDownToOverUp
CondOverUpToOverDown
CondOverUpToIdle
CondIdleToOverUp
*/
	}
}

/**
 * Check if we need to redraw the button.
 * The redraw action is given by:
 * - the redraw state of the button
 * - the redraw state of each of its component
 **/
bool ASButton::checkNeedRedraw() {
	if (needRedraw) return true;
	for (list<BUTTONRECORD*>::iterator it = activeCharacters.begin();it != activeCharacters.end(); it++) {
		BUTTONRECORD* b= *it;
		wstring movieName = encodeCharacterName(b->characterID);
		ASInterface *a = getChildByName(movieName);
		if (a->isDirty()) return true;
	}
	return false;
}

/**
 * Returns if the mouse location is inside the hit test area
 **/
bool ASButton::isInSelectionBox(int x, int y) {
	bool res = false;
	for (StringRectMap::const_iterator it = hitBounds->begin();it != hitBounds->end(); ++it) {
		wstring movName = (*it).first;
		RECT bounds = (*it).second;
		float invMat[4][4];
		float xt=x,yt=y;
		if (invMatrixes[movName] != NULL) {
			float *m = (float*)invMatrixes[movName];
			invert4x4Matrix(m, (float*)invMat);
			doTransform(xt,yt,(float*)invMat);					
		} else {
			// No placeObject have been defined. Use identity matrix
			xt = x; yt = y;
		}
		res = xt > bounds.xmin && xt < bounds.xmax && yt > bounds.ymin && yt < bounds.ymax;
		// If we are inside the bounding box, we refine by looking at the geometry definition.

		if (res) {
			ASInterface *amc = getChildByName(movName);
			res = amc->pointIsInsideGeometry(xt,yt);
			if (res) {
				selectedMovie = movName;
				return true;
			}
		}
	}
	return res;
}

/**
 * Set the inverse transform Matrix for a given shape in the button
 **/
void ASButton::setInverseMatrix(wstring childName, float *invM) {
	if (invMatrixes[childName] != NULL) delete []invMatrixes[childName];
	float *t = new float[16];
	for (int i=0; i< 16; i++) t[i] = invM[i];
	invMatrixes[childName] = t;
}

/**
 * Returns true, if this button handles the keyboard key passed as a parameter
 **/
bool ASButton::handleThisKey(int key) {
	bool res = false;

	for (list<BUTTONCONDACTION*>::iterator it=tag->getActions()->begin();it != tag->getActions()->end(); it++) {
		BUTTONCONDACTION *bc = *it;
		if (key == bc->getCondKeyPress()) {
			execActionScript(bc->getCode());
			res = true;
			break;
		}
	}
	return res;
}

/**
 * Return button members
 **/
ValueObject*ASButton::getMember(wstring memberName) {
	ValueObject *res = getMemberCommon(memberName);
	if (res != NULL) return res;
 	if (res == NULL && memberName == L"_currentframe") {
/////////		res = new IntegerObject(currentFrame +1);
	} else if (memberName == L"_totalframes") {
/////////		res = new IntegerObject(frameCount);		
	} else if (memberName == L"_alpha") { 
///////////		res = new IntegerObject(100);	// It's not CXFORMWITHALPHA. WHat's that alpha value?
	} else if (memberName == L"_width") { 
		v0 =  100000;
		vf = -100000;
		res = new FloatObject(getWidth());
	} else if (memberName == L"_height") { 
		v0 =  100000;
		vf = -100000;
		res = new FloatObject(getHeight());
	}  else if (memberName == L"_target") { 
//////////		res = new StringObject(L"/"+movieName);	// FIXME: nesting
	} else if (memberName == L"_framesloaded") { 
//////////		res = new IntegerObject(frameCount);	// Frame loaded? Not correct.
	} 	else if (memberName == L"_droptarget") { 
		res = new StringObject(L"");			// ?????
	} else if (memberName == L"_url") { 
///////////		res = new StringObject(L"undefined");	// TODO.
	} else if (memberName == L"_focusrect") { 
		res = new StringObject(L"null");		// Focus rect for buttons, but when not null?????
	}  else {
		res = defaultGetMember2(memberName);
	}
	return res;
}

/**
 * Returns the width of a button. That's the sum of test areas
 **/
float ASButton::getWidth() {
	for (list<BUTTONRECORD*>::iterator it = tag->getCharacters()->begin(); it != tag->getCharacters()->end(); it++) {
		BUTTONRECORD *br = *it;
		wstring movieName = encodeCharacterName(br->characterID);
		if (br->buttonStateHitTest) {
			RECT rect = (*hitBounds)[movieName];
			ASInterface *amc = shapes[br->characterID];
			if (rect.xmin != rect.xmax) {
				float xmin = rect.xmin;
				if(amc->getHasTransformMatrix()) xmin = xmin + amc->getTransformMatrix()->translateX; 
				float xmax = rect.xmax;
				if(amc->getHasTransformMatrix()) xmax = xmax + amc->getTransformMatrix()->translateX; 
				v0 = v0 < xmin ? v0 : xmin;
				vf = vf > xmax ? vf : xmax;
			} 
		}
	}
	return (vf-v0)/20.0;
}

/**
 * Returns the height of a button. That's the sum of test areas
 **/
float ASButton::getHeight() {
	for (list<BUTTONRECORD*>::iterator it = tag->getCharacters()->begin(); it != tag->getCharacters()->end(); it++) {
		BUTTONRECORD *br = *it;
		wstring movieName = encodeCharacterName(br->characterID);
		if (br->buttonStateHitTest) {
			RECT rect = (*hitBounds)[movieName];
			ASInterface *amc = shapes[br->characterID];
			if (rect.ymin != rect.ymax) {
				float ymin = rect.ymin;
				if(amc->getHasTransformMatrix()) ymin = ymin + amc->getTransformMatrix()->translateY; 
				float ymax = rect.ymax;
				if(amc->getHasTransformMatrix()) ymax = ymax + amc->getTransformMatrix()->translateY; 
				v0 = v0 < ymin ? v0 : ymin;
				vf = vf > ymax ? vf : ymax;
			} 
		}
	}
	return (vf-v0)/20.0;
}

/**
 * Returns the movie selected by the buttonState
 **/
ASInterface *ASButton::getMovieFromButtonState(int buttonState) {
	ASInterface *res = NULL;
	for (list<BUTTONRECORD*>::iterator it = tag->getCharacters()->begin(); it != tag->getCharacters()->end(); it++) {
		BUTTONRECORD *br = *it;
		if ((br->buttonStateDown && buttonState == buttonDown) || 
			(br->buttonStateOver && buttonState == buttonOver) ||
			(br->buttonStateUp && buttonState == buttonUp) ||
			(br->buttonStateHitTest && buttonState == buttonHit) ) {
			if (!br->buttonStateHitTest || res == NULL) {
				res = getChildByName(encodeCharacterName(br->characterID));
			}
		}
	}
	return res;
}

