/**
Action Script library. Date library implementation
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_date.hh"

typedef void (ASDate::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASDate::init,				// 0
	&ASDate::getDate,			// 1
	&ASDate::getDay,			// 2
	&ASDate::getFullYear,		// 3
	&ASDate::getHours,			// 4
	&ASDate::getMilliseconds,	// 5
	&ASDate::getMinutes,		// 6
	&ASDate::getMonth,			// 7
	&ASDate::getSeconds,		// 8
	&ASDate::getTime,			// 9
	&ASDate::getTimezoneOffset,	// 10
	&ASDate::getUTCDate,		// 11	
	&ASDate::getUTCDay,			// 12
	&ASDate::getUTCFullYear,	// 13
	&ASDate::getUTCHours,		// 14
	&ASDate::getUTCMilliseconds,// 15
	&ASDate::getUTCMinutes,		// 16
	&ASDate::getUTCMonth,		// 17
	&ASDate::getUTCSeconds,		// 18
	&ASDate::getYear,			// 19
	&ASDate::setDate,			// 20
	&ASDate::setFullYear,		// 21
	&ASDate::setHours,			// 22
	&ASDate::setMilliseconds,	// 23
	&ASDate::setMinutes,		// 24
	&ASDate::setMonth,			// 25
	&ASDate::setSeconds,		// 26
	&ASDate::setTime,			// 27
	&ASDate::setUTCDate,		// 28
	&ASDate::setUTCFullYear,	// 29
	&ASDate::setUTCHours,		// 30
	&ASDate::setUTCMilliseconds,// 31
	&ASDate::setUTCMinutes,		// 32
	&ASDate::setUTCMonth,		// 33
	&ASDate::setUTCSeconds,		// 34
	&ASDate::setYear,			// 35
	&ASDate::AStoString,		// 36
	&ASDate::UTC,				// 37
	// Need to be at the end.
	&ASDate::addProperty,		
	&ASDate::registerClass,	
	&ASDate::AStoString,		
	&ASDate::unwatch,			
	&ASDate::valueOf,			
	&ASDate::watch			
};

/**
 * The constructor
 **/
ASDate::ASDate(Script *s, Graphics *g) : ASObject(s,g,L"Date") {
	timeShift = 0;
}

/**
 * The destructor
 **/
ASDate::~ASDate(){
}

/**
 * Return a function descriptor
 */
Function *ASDate::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"Date") 			{func->setFunctionIndex(0); func->setNumParams(7); }
	if (memberName == L"getDate") 		{func->setFunctionIndex(1); func->setNumParams(0); }
	if (memberName == L"getDay") 		{func->setFunctionIndex(2); func->setNumParams(0); }
	if (memberName == L"getFullYear") 	{func->setFunctionIndex(3); func->setNumParams(0); }
	if (memberName == L"getHours") 		{func->setFunctionIndex(4); func->setNumParams(0); }
	if (memberName == L"getMilliseconds"){func->setFunctionIndex(5); func->setNumParams(0); }
	if (memberName == L"getMinutes") 	{func->setFunctionIndex(6); func->setNumParams(0); }
	if (memberName == L"getMonth") 		{func->setFunctionIndex(7); func->setNumParams(0); }
	if (memberName == L"getSeconds") 	{func->setFunctionIndex(8); func->setNumParams(0); }
	if (memberName == L"getTime") 		{func->setFunctionIndex(9); func->setNumParams(0); }
	if (memberName == L"getTimezoneOffset"){func->setFunctionIndex(10); func->setNumParams(0); }
	if (memberName == L"getUTCDate") 	{func->setFunctionIndex(11); func->setNumParams(0); }
	if (memberName == L"getUTCDay") 	{func->setFunctionIndex(12); func->setNumParams(0); }
	if (memberName == L"getUTCFullYear") {func->setFunctionIndex(13); func->setNumParams(0); }
	if (memberName == L"getUTCHours") 	{func->setFunctionIndex(14); func->setNumParams(0); }
	if (memberName == L"getUTCMilliseconds"){func->setFunctionIndex(15); func->setNumParams(0); }
	if (memberName == L"getUTCMinutes") {func->setFunctionIndex(16); func->setNumParams(0); }
	if (memberName == L"getUTCMonth") 	{func->setFunctionIndex(17); func->setNumParams(0); }
	if (memberName == L"getUTCSeconds") {func->setFunctionIndex(18); func->setNumParams(0); }
	if (memberName == L"getYear") 		{func->setFunctionIndex(19); func->setNumParams(0); }
	if (memberName == L"setDate") 		{func->setFunctionIndex(20); func->setNumParams(1); }
	if (memberName == L"setFullYear") 	{func->setFunctionIndex(21); func->setNumParams(3); }
	if (memberName == L"setHours") 		{func->setFunctionIndex(22); func->setNumParams(1); }
	if (memberName == L"setMilliseconds"){func->setFunctionIndex(23); func->setNumParams(1); }
	if (memberName == L"setMinutes") 	{func->setFunctionIndex(24); func->setNumParams(1); }
	if (memberName == L"setMonth") 		{func->setFunctionIndex(25); func->setNumParams(2); }
	if (memberName == L"setSeconds") 	{func->setFunctionIndex(26); func->setNumParams(1); }
	if (memberName == L"setTime") 		{func->setFunctionIndex(27); func->setNumParams(1); }
	if (memberName == L"setUTCDate") 	{func->setFunctionIndex(28); func->setNumParams(1); }
	if (memberName == L"setUTCFullYear"){func->setFunctionIndex(29); func->setNumParams(3); }
	if (memberName == L"setUTCHours") 	{func->setFunctionIndex(30); func->setNumParams(4); }
	if (memberName == L"setUTCMilliseconds"){func->setFunctionIndex(31); func->setNumParams(1); }
	if (memberName == L"setUTCMinutes") {func->setFunctionIndex(32); func->setNumParams(3); }
	if (memberName == L"setUTCMonth") 	{func->setFunctionIndex(33); func->setNumParams(2); }
	if (memberName == L"setUTCSeconds") {func->setFunctionIndex(34); func->setNumParams(2); }
	if (memberName == L"setYear") 		{func->setFunctionIndex(35); func->setNumParams(1); }
	if (memberName == L"toString") 		{func->setFunctionIndex(36); func->setNumParams(1); }
	if (memberName == L"UTC") 			{func->setFunctionIndex(37); func->setNumParams(7); }
	finishFunctionSetup(memberName,38,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASDate::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 * Action Script Date constructor
 * new Date()
 * new Date(year, month [, date [, hour [, minute [, second [, millisecond ]]]]])
 **/
void ASDate::init() {
	timeShift = 0;
	if (numPassedParams == 0) return;
	year=month=date=hour=minute=second=millisecond =0;
	if (numPassedParams > 0) year 		= int(script->getRegister(1)->toNumeric());
	if (numPassedParams > 1) month 		= int(script->getRegister(2)->toNumeric());
	if (numPassedParams > 2) date 		= int(script->getRegister(3)->toNumeric());
	if (numPassedParams > 3) hour 		= int(script->getRegister(4)->toNumeric());
	if (numPassedParams > 4) minute 	= int(script->getRegister(5)->toNumeric());
	if (numPassedParams > 5) second 	= int(script->getRegister(6)->toNumeric());
	if (numPassedParams > 6) millisecond= int(script->getRegister(7)->toNumeric());
	setLocalSystemDate();
}

/**
 * returns the day of the month 
 **/
void ASDate::getDate() {
	getLocalSystemDate();
	script->push(new IntegerObject(date));
}

/**
 * returns the day of the week (0 for Sunday, 1 for Monday, and so on)
 **/
void ASDate::getDay() {
	getLocalSystemDate();
	script->push(new IntegerObject(dayWeek));
}

/**
 * returns the full year (a four-digit number, for example, 2000)
 **/
void ASDate::getFullYear() {
	getLocalSystemDate();
	script->push(new IntegerObject(year+1900));
}

/**
 * returns the hour (an integer from 0 to 23) 
 **/
void ASDate::getHours() {
	getLocalSystemDate();
	script->push(new IntegerObject(hour));
}

/**
 *  returns the milliseconds
 **/
void ASDate::getMilliseconds() {
	getLocalSystemDate();
	script->push(new IntegerObject(millisecond));
}

/**
 * returns the minutes (an integer from 0 to 59)
 **/
void ASDate::getMinutes() {
	getLocalSystemDate();
	script->push(new IntegerObject(minute));
}

/**
 * returns the month (0 for January, 1 for February, and so on) 
 **/
void ASDate::getMonth() {
	getLocalSystemDate();
	script->push(new IntegerObject(month));
}

/**
 * returns the seconds (an integer from 0 to 59) 
 **/
void ASDate::getSeconds() {
	getLocalSystemDate();
	script->push(new IntegerObject(second));
}

/**
 *  returns the number of milliseconds since midnight January 1, 1970 UTC
 **/
void ASDate::getTime() {
	struct timeval tEnd;
	gettimeofday(&tEnd,NULL);
	long a  = tEnd.tv_sec*1000 + tEnd.tv_usec/1000;
	script->push(new IntegerObject(a));
}

/**
 * returns the difference, in minutes, between the computer's local time and universal time.
 **/
void ASDate::getTimezoneOffset() {
	time_t t = time(&t);
	struct tm *t0 = localtime(&t);
	struct tm *t1 = gmtime(&t);
	time_t a = mktime(t0) - mktime(t1);
	script->push(new IntegerObject(a/60));
}

/**
 * returns the day of the month (an integer from 1 to 31)
 **/
void ASDate::getUTCDate() {
	getUTCSystemDate();
	script->push(new IntegerObject(date));
}

/**
 *  returns the day of the week (0 for Sunday, 1 for Monday, and so on)
 **/
void ASDate::getUTCDay() {
	getUTCSystemDate();
	script->push(new IntegerObject(dayWeek));
}

/**
 * returns the four-digit year of the specified 
 **/
void ASDate::getUTCFullYear() {
	getUTCSystemDate();
	script->push(new IntegerObject(year+1900));
}

/**
 * returns the hours of the specified Date object, according to universal time
 **/
void ASDate::getUTCHours() {
	getUTCSystemDate();
	script->push(new IntegerObject(hour));
}

/**
 * returns the milliseconds 
 **/
void ASDate::getUTCMilliseconds() {
	getUTCSystemDate();
	script->push(new IntegerObject(millisecond));
}

/**
 * returns the minutes 
 **/
void ASDate::getUTCMinutes() {
	getUTCSystemDate();
	script->push(new IntegerObject(minute));
}

/**
 * returns the month (0 for January, 1 for February, and so on) 
 **/
void ASDate::getUTCMonth() {
	getUTCSystemDate();
	script->push(new IntegerObject(month));
}

/**
 *
 **/
void ASDate::getUTCSeconds() {
	getUTCSystemDate();
	script->push(new IntegerObject(second));
}

/**
 * Year minus 1900
 **/
void ASDate::getYear() {
	getUTCSystemDate();
	script->push(new IntegerObject(year));
}

/**
 * Set local date
 **/
void ASDate::setDate() {
	getLocalSystemDate();
	date = int(script->getRegister(1)->toNumeric());
	setLocalSystemDate();
	script->push(new IntegerObject(getLocalTimeInMillis()));
}

/**
 * Set year 
 **/
void ASDate::setFullYear() {
	getLocalSystemDate();
	year = int(script->getRegister(1)->toNumeric()) - 1900;
	setLocalSystemDate();
	script->push(new IntegerObject(getLocalTimeInMillis()));
}

/**
 * Set hours
 **/
void ASDate::setHours() {
	getLocalSystemDate();
	hour = int(script->getRegister(1)->toNumeric());
	setLocalSystemDate();
	script->push(new IntegerObject(getLocalTimeInMillis()));
}

/**
 * set millisecond
 **/
void ASDate::setMilliseconds() {
	getLocalSystemDate();
	millisecond = int(script->getRegister(1)->toNumeric());
	setLocalSystemDate();
	script->push(new IntegerObject(getLocalTimeInMillis()));
}

/**
 * Set minutes
 **/
void ASDate::setMinutes() {
	getLocalSystemDate();
	minute = int(script->getRegister(1)->toNumeric());
	setLocalSystemDate();
	script->push(new IntegerObject(getLocalTimeInMillis()));
}

/**
 * Set month
 **/
void ASDate::setMonth() {
	getLocalSystemDate();
	month = int(script->getRegister(1)->toNumeric());
	if (numPassedParams > 1) date  = int(script->getRegister(2)->toNumeric());
	setLocalSystemDate();
	script->push(new IntegerObject(getLocalTimeInMillis()));
}

/**
 * Set seconds
 **/
void ASDate::setSeconds() {
	getLocalSystemDate();
	second = int(script->getRegister(1)->toNumeric());
	setLocalSystemDate();
	noReturnValue();
	script->push(new IntegerObject(getLocalTimeInMillis()));
}

/**
 * Sets the date in milliseconds. Returns the new time in milliseconds.
 **/
void ASDate::setTime() {
	time_t t = time_t(script->getRegister(1)->toNumeric());
	struct tm *ti = localtime(&t);
	second = ti->tm_sec;
	minute = ti->tm_min;
	hour = ti->tm_hour;
	date = ti->tm_mday;
	month = ti->tm_mon;
	year = ti->tm_year;
	script->push(new IntegerObject(t*1000));
}

/**
 * set UTC date
 **/
void ASDate::setUTCDate() {
	getUTCSystemDate();
	date = int(script->getRegister(1)->toNumeric());
	setUTCSystemDate();
	script->push(new IntegerObject(getUTCTimeInMillis()));
}

/**
 * set year 
 **/
void ASDate::setUTCFullYear() {
	getUTCSystemDate();
	year = int(script->getRegister(1)->toNumeric()) - 1900;
	if (numPassedParams > 1) month = int(script->getRegister(2)->toNumeric());
	if (numPassedParams > 2) date  = int(script->getRegister(3)->toNumeric());
	setUTCSystemDate();
	script->push(new IntegerObject(getUTCTimeInMillis()));
}

/**
 * Set UTC hour
 **/
void ASDate::setUTCHours() {
	getUTCSystemDate();
	hour = int(script->getRegister(1)->toNumeric());
	setUTCSystemDate();
	script->push(new IntegerObject(getUTCTimeInMillis()));
}

/**
 * Set UTC milliseconds
 **/
void ASDate::setUTCMilliseconds() {
	getUTCSystemDate();
	millisecond = int(script->getRegister(1)->toNumeric());
	setUTCSystemDate();
	script->push(new IntegerObject(getUTCTimeInMillis()));
}

/**
 * Set UTC minute
 **/
void ASDate::setUTCMinutes() {
	getUTCSystemDate();
	minute = int(script->getRegister(1)->toNumeric());
	setUTCSystemDate();
	script->push(new IntegerObject(getUTCTimeInMillis()));
}

/**
 * Set UTC month
 **/
void ASDate::setUTCMonth() {
	getUTCSystemDate();
	month = int(script->getRegister(1)->toNumeric());
	if (numPassedParams > 1) date  = int(script->getRegister(2)->toNumeric());
	setUTCSystemDate();
	script->push(new IntegerObject(getUTCTimeInMillis()));
}

/**
 * Set UTC seconds
 **/
void ASDate::setUTCSeconds() {
	getUTCSystemDate();
	second = int(script->getRegister(1)->toNumeric());
	if (numPassedParams > 1) millisecond  = int(script->getRegister(2)->toNumeric());
	setUTCSystemDate();
	script->push(new IntegerObject(getUTCTimeInMillis()));
}

/**
 * Set UTC year since 1900
 **/
void ASDate::setYear() {
	getUTCSystemDate();
	year = int(script->getRegister(1)->toNumeric());
	setUTCSystemDate();
	script->push(new IntegerObject(getUTCTimeInMillis()));
}

/**
 * returns a string value for the specified date object in a readable format
 **/
void ASDate::AStoString() {
	time_t t = time(&t) + timeShift;
	string s = string(asctime(gmtime(&t)));
	script->push(new StringObject(stringTowstring(s.c_str())));
}

/**
 * init Date in UTC
 **/
void ASDate::UTC() {
	if (numPassedParams == 0) return;
	year=month=date=hour=minute=second=millisecond =0;
	if (numPassedParams > 0) year 		= int(script->getRegister(1)->toNumeric());
	if (numPassedParams > 1) month 		= int(script->getRegister(2)->toNumeric());
	if (numPassedParams > 2) date 		= int(script->getRegister(3)->toNumeric());
	if (numPassedParams > 3) hour 		= int(script->getRegister(4)->toNumeric());
	if (numPassedParams > 4) minute 	= int(script->getRegister(5)->toNumeric());
	if (numPassedParams > 5) second 	= int(script->getRegister(6)->toNumeric());
	if (numPassedParams > 6) millisecond= int(script->getRegister(7)->toNumeric());
	setUTCSystemDate();
	script->push(new IntegerObject(getUTCTimeInMillis()));
}

/**
 * Set the date (localtime format)
 * We set the class date not the system date!
 **/
void ASDate::setLocalSystemDate() {
	struct tm ti;
	ti.tm_sec = second;
	ti.tm_min = minute;
	ti.tm_hour = hour;
	ti.tm_mday = date;
	ti.tm_mon = month;
	ti.tm_year = year;
	time_t newdate= mktime(&ti);
	time_t now = time(&now);
	timeShift = newdate - now; 
}

/**
 * Set the date UTC
 * We set the class date not the system date!
 **/
void ASDate::setUTCSystemDate() {
	setLocalSystemDate();
	time_t t = time(&t);
	struct tm *t0 = localtime(&t);
	struct tm *t1 = gmtime(&t);
	timeShift += mktime(t1) - mktime(t0);
}

/*
 * get the date in UTC
 **/
void ASDate::getUTCSystemDate() {
	struct timeval tEnd;
	gettimeofday(&tEnd,NULL);
	millisecond = tEnd.tv_usec;
	time_t t = time(&t);
	t += timeShift;
	struct tm *ti = gmtime(&t);
	second = ti->tm_sec;
	minute = ti->tm_min;
	hour = ti->tm_hour;
	date = ti->tm_mday;
	month = ti->tm_mon;
	year = ti->tm_year;
	dayWeek = ti->tm_wday;
}

/*
 * get the Local date
 **/
void ASDate::getLocalSystemDate() {
	struct timeval tEnd;
	gettimeofday(&tEnd,NULL);
	millisecond = tEnd.tv_usec;
	time_t t = time(&t);
	t += timeShift;
	struct tm *ti = localtime(&t);
	second = ti->tm_sec;
	minute = ti->tm_min;
	hour = ti->tm_hour;
	date = ti->tm_mday;
	month = ti->tm_mon;
	year = ti->tm_year;
	dayWeek = ti->tm_wday;
}

/**
 * Return local time in milliseconds
 **/
long ASDate::getLocalTimeInMillis() {
	time_t t = time(&t);
	t += timeShift;
	struct tm *ti = localtime(&t);
	return mktime(ti);
}

/**
 * Return UTC time in milliseconds
 **/
long ASDate::getUTCTimeInMillis() {
	time_t t = time(&t);
	t += timeShift;
	struct tm *ti = localtime(&t);
	return mktime(ti);
}
