/**
Action Script library. Error class implementation
Copyright (C) 2005 Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_error.hh"

typedef void (ASError::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASError::init,			// 0
	// Need to be at the end.
	&ASError::addProperty,		
	&ASError::registerClass,	
	&ASError::AStoString,		
	&ASError::unwatch,			
	&ASError::valueOf,			
	&ASError::watch			
};

/**
 * The constructor
 **/
ASError::ASError(Script *s, Graphics *g) : ASObject(s,g,L"Error") {
}

/**
 * The destructor
 **/
ASError::~ASError() {
}

/**
 * Return a function descriptor
 */
Function *ASError::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"Error") 	{func->setFunctionIndex(0); func->setNumParams(1); }
	finishFunctionSetup(memberName,1,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASError::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}
/**
 * Argument message
 **/
void ASError::init() {
	if (numPassedParams > 0) {
		errorMessage = script->getRegister(1)->toString();
	}
}

/**
 * Return error message as String
 **/
void ASError::AStoString() {
	script->push(new StringObject(errorMessage));
}
