/**
Action Script library.	Non Action Script class - Used as a base built-in classe.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_interface.hh" 
#include "../lib/swf_engine.hh"
#include "../lib/swf_function.hh"
#include "../lib/event_manager.hh"

/**
 * The constructor
 ***/
ASInterface::ASInterface(Script *s,Graphics *g,wstring className) : ScriptObject(className) {
	this->className = className;
	this->script = s;
	this->graphics = g;
	objectUnid = long(this);		// Use class pointer as unique UID
	objectInstance = NULL;
	hasTransformMatrix = false;
	hasCxFormWithAlpha = false;
	movieName = L"undefined";
	isStopped = false;
	isReferenced = false;
	clipDepth = 0;
	isLoaded = false;
	parent = NULL;
	getObjectStore()->registerNewObject(this);
	isDuplicatedObject = false;
}

/**
 * Deep copy cloning
 **/
ASInterface::ASInterface(ASInterface *p) : ScriptObject(p->className) {
	this->className = p->className;
	this->movieName = p->movieName;
	this->script = p ->script,
	this->objectInstance = p->objectInstance;
	this->graphics = p->graphics;
	this->numPassedParams = p->numPassedParams;
	this->objectUnid = long(this); //p->objectUnid;
	this->currentFrame = 0; //p->currentFrame;
	this->transformMatrix = p->transformMatrix;
	this->cxFormWithAlpha = p->cxFormWithAlpha;
	this->hasTransformMatrix = p->hasTransformMatrix;
	this->hasCxFormWithAlpha = p->hasCxFormWithAlpha;
	this->bounds = p->bounds;
	this->isStopped = p->isStopped;
	this->isReferenced = false;
	this->needRedraw = true;
	memcpy(this->invMatrix,p->invMatrix,sizeof(invMatrix));
	this->clipDepth = 0; //p->clipDepth;
	this->objectInstanceName = p->getThisObjectInstanceName();
	this->members = new ObjectMembers();
	this->characterId = p->characterId;
	this->parent = NULL;
	this->isLoaded = false;
	this->isDuplicatedObject = false;
	// Need to clone members as well

	StringValueObjectHashMap *mem = p->getMembers()->getMembers();
	for (StringValueObjectHashMap::iterator it = mem->begin(); it != mem->end(); it++) {
		if ((*it).second != NULL) this->members->setMember((*it).first,(*it).second);
	}

	// Register this newly object as a cloned object of the given characterId.
	getObjectStore()->addClonedObject(p->characterId,this);
	//printf("New OBJ : %d",characterId);
}

/**
 * The destructor
 **/
ASInterface::~ASInterface() {
	//printf("~ASInterface");
	getObjectStore()->removeClonedObject(characterId,this);
	getObjectStore()->unRegisterNewObject(this);
	if (objectInstance != NULL) delete objectInstance;
	graphics = NULL;	
	script = NULL;
}

/**
 * Return a pre-configured Function definition for built in func
 **/
Function* ASInterface::getBaseFunction(wstring methodName) {
	Function *func = new Function();
	func->setFunctionName(methodName);
	func->setPreloadParentFlag(false);
	func->setPreloadRootFlag(false);
	func->setSuppressSuperFlag(true);
	func->setPreloadSuperFlag(false);
	func->setSuppressArgumentsFlag(true);
	func->setPreloadArgumentsFlag(false);
	func->setSuppressThisFlag(true);
	func->setPreloadThisFlag(false);
	func->setPreloadGlobalFlag(false);
	func->setIsFunction2(true);
	func->setIsNative(true);
 	func->setFunctionIndex(-1);
	func->setClassName(className);
	return func;
}

/**
 * Complete function callback initialization
 **/
Function* ASInterface::finishFunctionSetup(wstring memberName, int baseObjectIdx, Function *func) {
	if (memberName == L"addProperty") 	{func->setFunctionIndex(baseObjectIdx+0); func->setNumParams(3); }
	if (memberName == L"registerClass") {func->setFunctionIndex(baseObjectIdx+1); func->setNumParams(2); }
	if (memberName == L"toString") 		{func->setFunctionIndex(baseObjectIdx+2); func->setNumParams(10); } // max is 1?
	if (memberName == L"unwatch") 		{func->setFunctionIndex(baseObjectIdx+3); func->setNumParams(1); }
	if (memberName == L"valueOf") 		{func->setFunctionIndex(baseObjectIdx+4); func->setNumParams(0); }
	if (memberName == L"watch") 		{func->setFunctionIndex(baseObjectIdx+5); func->setNumParams(3); }
	if (func->getFunctionIndex() == -1) {
		delete func;
		return NULL;
	}
	wstring *s = new wstring[func->getNumParams()];
	UI8 *paramRegisters = new UI8[func->getNumParams()];
	for (int i=0; i< func->getNumParams(); i++) {
		s[i] = L"p"+i;
		paramRegisters[i] = i+1;
	}
	func->setParams(s);
	func->setParamRegisters(paramRegisters);
	func->setRegisterCount(1+func->getNumParams());
	return func;
}

/**
 * Returns if an object is visible
 **/
bool ASInterface::isVisible() {
	ValueObject *v = defaultGetMember(L"_visible");
	bool res = v->toString() == L"true";
	delete v;
	return res;
}

/**
 * Set the object visible (it has been placed)
 **/
void ASInterface::setVisible(bool val) {
	setMemberNoClone(L"_visible", new BooleanObject(val));
}

/**
 * Set rendering inverse transform matrix.
 * This matrix is used when the user click on an input area to find click position in object coordinate space.
 * A typical usage is dynamic text editing.
 * childName is here ignored.
 **/
void ASInterface::setInverseMatrix(wstring childName, float *m) { 
	memcpy(invMatrix,m,16*sizeof(float));
}

void ASInterface::callEvent(string eventName) {
	puts("ASInterface::callEvent should be OVERLOADED!!!!");
}

bool ASInterface::callInternalEvent(string eventName, void *data) {
	puts("ASInterface::callInternalEvent should be OVERLOADED!!!!");
	return false;
}

void ASInterface::setTransformMatrixClones(MATRIX *c) {
	list<ScriptObject*> l = getObjectStore()->getClonedObjects(characterId);
	for (list<ScriptObject*>::const_iterator it=l.begin(); it != l.end(); ++it) {
		ASObject *aos = (ASObject*) *it;
		//if (aos != this) aos->setTransformMatrixRecurs(c);
		aos->setTransformMatrix(c);
	}
}
void ASInterface::setHasTransformMatrixClones(bool p) {
	list<ScriptObject*> l = getObjectStore()->getClonedObjects(characterId);
	for (list<ScriptObject*>::const_iterator it=l.begin(); it != l.end(); ++it) {
		ASObject *aos = (ASObject*) *it;
		//if (aos != this) aos->setHasTransformMatrixRecurs(p);
		aos->setHasTransformMatrix(p);
	}
}

void ASInterface::setTransformMatrix(MATRIX *c) { 
	transformMatrix = *c;
}
void ASInterface::setHasTransformMatrix(bool p) { 
	hasTransformMatrix = p; 
}

/**
 * Execute action script code
 **/
void ASInterface::execActionScript(UI8 *code) {
	//puts("ASInterface::execActionScript");
	SwfEngine *e = script->getEngine();
	while (*code != 0) e->actionDispatch(&code);
	//puts("End AS");
}

/**
 * Returns true if the mouse location is inside the object
 * This method will be overidden for buttons since buttons may have multiple selection areas.
 **/
bool ASInterface::isInSelectionBox(int x, int y) {
	float invMat[4][4];
	invert4x4Matrix(invMatrix, (float*)invMat);
	float xt=x,yt=y;
	doTransform(xt,yt,(float*)invMat);
	bool res = xt > bounds.xmin && xt < bounds.xmax && yt > bounds.ymin && yt < bounds.ymax;
	if (res) res = pointIsInsideGeometry(xt,yt);
	return res;
}

/**
 * Mouse to object Transform
 **/
void ASInterface::mouseTransform(int x, int y, float xt, float yt) {
	float invMat[4][4];
	invert4x4Matrix(invMatrix, (float*)invMat);
	xt=x,yt=y;
	doTransform(xt,yt,(float*)invMat);
}

/**
 * Test if a point is inside geometry definition
 *
 * Loop over all triangles inside the shape and find if the point is inside.
 **/
bool ASInterface::pointIsInsideGeometry(float xt, float yt) {
	bool res = false;
	if (needRedraw) toGraphic();
	listDrawingSequence *dr = graphics->getSequence(this->getUnid());
	for (listDrawingSequence::iterator it=dr->begin();it != dr->end(); it++) {
		DrawingSequence *d = (*it);
		for (list<DrawingEntry*>::iterator it= d->getDrawingEntries()->begin(); it != d->getDrawingEntries()->end(); it++) {
			DrawingEntry *de = *it;
			if (de->drawingType == DrawingTriangles || de->drawingType == DrawingFontChars) {
				float v[3][2];
				for (int i=0; i< de->nbElements; i++) {
					int idx = de->indexes[i*3+0];
					v[0][0] = d->getVertexes()[idx*2];
					v[0][1] = d->getVertexes()[idx*2+1];
					idx = de->indexes[i*3+1];
					v[1][0] = d->getVertexes()[idx*2];
					v[1][1] = d->getVertexes()[idx*2+1];
					idx = de->indexes[i*3+2];
					v[2][0] = d->getVertexes()[idx*2];
					v[2][1] = d->getVertexes()[idx*2+1];
					res |= triangle_contains_point_2d(v,xt,yt);
					if (res) return true;
				}
			}
		}
	}
	return res;
}

/**
 * Copy iniObject members
 **/
void ASInterface::populateMembersFromInitObject(ScriptObject *initObject) {
	StringValueObjectHashMap *members = initObject->getMembers()->getMembers();
	for (StringValueObjectHashMap::const_iterator i=members->begin(); i!=members->end(); ++i) 
		setMember(i->first,i->second);
}

/**
 * Return members that are common to sprites/movieclips/buttons
 **/
ValueObject* ASInterface::getMemberCommon(wstring memberName) {
	wstring name = toUpper(memberName);
	ValueObject *res = NULL;
	if (name == L"_X") {
		float x = 0;
		if (getHasTransformMatrix()) x = float(transformMatrix.translateX)/20.0;
		res = new FloatObject(x);
	} else if (name == L"_Y") { 
		float y = 0;
		if (getHasTransformMatrix()) y = float(transformMatrix.translateY)/20.0;
		res = new FloatObject(y);
	} else if (name == L"_XSCALE") { 
		float s = 100;
		if (getHasTransformMatrix()) s = float(transformMatrix.scaleX)/655.36;
		res = new FloatObject(s);
	} else if (name == L"_YSCALE") { 
		float s = 100;
		if (getHasTransformMatrix()) s = float(transformMatrix.scaleY)/655.36;
		res = new FloatObject(s);
	} else if (memberName == L"_visible"){
		res = members->getMember(L"_visible");
	} else if (memberName == L"_rotation") { 
		float r = 0;
		if (getHasTransformMatrix()) r = getTransformationMatrixAngle(&transformMatrix);
		res = new FloatObject(r);
	} else if (memberName == L"_name") {
		res = new StringObject(movieName);		
	}else if (memberName == L"_quality") { 
		res = new StringObject(L"HIGH");
	} else if (memberName == L"_xmouse") {
		double x = script->getEngine()->getEventManager()->getMouseX();
		double xt = (x - transformMatrix.translateX)*65536/transformMatrix.scaleX / 20.0;
		res = new FloatObject(xt);
	} else if (memberName == L"_ymouse") { 
		double y = script->getEngine()->getEventManager()->getMouseY();
		double yt = (y - transformMatrix.translateY)*65536/transformMatrix.scaleY / 20.0;
		res = new FloatObject(yt);
	} else if (memberName == L"_soundbuftime") { 
		res = new IntegerObject(1);				// ??????
	} else if (memberName == L"_highquality") {
		res = new IntegerObject(1);
	} 
	return res;
}


/**
 * Should be implemented for some objects only like sprites
 * 
 * TODO:
 **/
void ASInterface::actionNextFrame() { };
void ASInterface::actionPreviousFrame() {};
void ASInterface::actionGotoFrame(int frameNum) {};

void ASInterface::setChildrenProperty(wstring member, ValueObject *value) { 
	wcout << L"BUG! please overload setChildrenProperty " << *this << L"\n" ; exit(0);
}

bool ASInterface::handleThisKey(int key) {
	wcout << L"BUG! please overload handleThisKey " << *this << L"\n" ; exit(0);
}

ASInterface*ASInterface::getChildByName(wstring movieName){
	return (ASInterface *) script->getObjectStore()->getSpriteVariable(this,movieName);
}

ASInterface*ASInterface::getRootMovieByName(wstring movieName) {
//??	ScriptObject *global = script->getEngine()->getGlobal();
//??	return (ASInterface *) script->getObjectStore()->getSpriteVariable(global,movieName);
	ValueObject *s = getObjectStore()->getGlobalVariable(movieName);
	return (ASInterface *)s;
}

void ASInterface::computeBounds() {
	wcout << L"BUG! please overload computeBounds " << *this << L"\n" ; exit(0);
}

