/**
Action Script library.Non Action Script class - Used as a base built-in classe.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#ifndef _AS_INTERFACE_HH
#define _AS_INTERFACE_HH
#include "../lib/object_types.hh"
#include "../lib/script.hh"

typedef stdext::hash_map<string, Function*, HashString,  HashStringCompare> StringFuncHashMap;
typedef stdext::hash_map<int, MATRIX ,  HashInt,HashIntCompare> IntMatrixHashMap;
typedef stdext::hash_map<int, CXFORMWITHALPHA , HashInt, HashIntCompare> IntCxAlphaHashMap;
typedef stdext::hash_map<int, bool, HashInt, HashIntCompare> IntBoolHashMap;

class Graphics;

class ASInterface : public ScriptObject{
public:
	ASInterface(ASInterface*);
	ASInterface(Script *s,Graphics *g,wstring className);
	~ASInterface();
	virtual Function *getFunctionMember(wstring memberName) =0;
	virtual void init() = 0;
	long	getUnid() { return objectUnid; }

	// MovieClip transforms
	void	setTransformMatrix(MATRIX *c);
	void	setCxFormWithAlpha(CXFORMWITHALPHA *c) { cxFormWithAlpha = *c; }
	MATRIX *getTransformMatrix() { return &transformMatrix; }
	CXFORMWITHALPHA *getCxFormWithAlpha() { return &cxFormWithAlpha; }
	void	setHasTransformMatrix(bool p);
	bool	getHasTransformMatrix() { return hasTransformMatrix; }
	void	setHasCxFormWithAlpha(bool p) { hasCxFormWithAlpha = p; }
	bool	getHasCxFormWithAlpha() { return hasCxFormWithAlpha; }

	void	setBounds(RECT *c) { bounds = *c; needRedraw=true; }
	RECT	*getBounds() { return &bounds; }
	virtual bool isInSelectionBox(int x, int y);
	virtual ASInterface*getNewInstance(Script *s,Graphics *g)=0;
	virtual void	actionNextFrame();
	virtual void	actionPreviousFrame();
	virtual void	actionGotoFrame(int frameNum);
	bool	isVisible();
	void	setVisible(bool val);
	virtual void setInverseMatrix(wstring childName, float *);
	void 	setClipDepth(int p) { clipDepth = p; }
	int		getClipDepth() { return clipDepth; }

	virtual void callEvent(string eventName);
	virtual bool callInternalEvent(string eventName, void *data);
	bool	pointIsInsideGeometry(float xt, float yt);
	void	mouseTransform(int x, int y, float xt, float yt);
	bool	getIsStopped() { return isStopped; }
	void	setIsStopped(bool p) { isStopped = p; }
	void	setReferenced(bool p) { isReferenced = p; }
	bool	getIsReferenced() { return isReferenced; }
	virtual ASInterface* getObjectDeepCopy(bool deep) { wcout << L"BUG! please overload getObjectDeepCopy " << *this << L"\n" ; exit(0);}
	void	setLoaded(bool p) { isLoaded = p; }
	bool	getIsLoaded() { return isLoaded; }
	virtual	void setChildrenProperty(wstring member, ValueObject *value);

	void			setTransformMatrixClones(MATRIX *c);
	void			setHasTransformMatrixClones(bool p);
	void			populateMembersFromInitObject(ScriptObject *p);
	ValueObject*	getMemberCommon(wstring memberName);
	ASInterface*	getParent() { return parent; }
	void			setParent(ASInterface *p) { parent = p; }
	int				getCurrentFrame() { return currentFrame; }
	virtual bool	handleThisKey(int key);
	ASInterface*	getChildByName(wstring movieName);
	ASInterface*	getRootMovieByName(wstring movieName);
	void			setIsDuplicated(bool p) { isDuplicatedObject = p; }
	virtual void	computeBounds();
protected: 
	Script 			*script;
	ValueObject 	*objectInstance;
	Graphics 		*graphics;
	int				numPassedParams;
	long			objectUnid;
	int				currentFrame;
	bool			isDuplicatedObject;

	MATRIX			transformMatrix;
	CXFORMWITHALPHA cxFormWithAlpha;
	bool			hasTransformMatrix;
	bool			hasCxFormWithAlpha;
	RECT			bounds;
	bool			isStopped;
	bool			isReferenced;

	Function*		getBaseFunction(wstring methodName);
	Function* 		finishFunctionSetup(wstring memberName, int baseObjectIndex,Function *f);
	ObjectStore *	getObjectStore() { return script->getObjectStore(); }
	void			noReturnValue() { script->push(new BooleanObject(true)); }
	void			execActionScript(UI8 *code);
	void			removeObject(ValueObject *v) { script->removeObject(v); }
private:
	float 			invMatrix[16];		// inverse transform matrix
	int	 			clipDepth;
	bool			isLoaded;
	ASInterface*	parent;
};

typedef stdext::hash_map<int, ASInterface*, HashInt,  HashIntCompare> IntASHashMap;
typedef stdext::hash_map<wstring, ASInterface*, HashUTF8String,  HashUTF8StringCompare> StringASHashMap;

#endif
