/**
Action Script library. Internal class
Copyright (C) 2005 Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_internal.hh"
#include "../lib/event_manager.hh"

typedef void (ASInternal::*funcPtrInternal)();
static funcPtrInternal callbacks[]= { 
	&ASInternal::init,						// 0
	&ASInternal::ASSetPropFlags,			// 1
	&ASInternal::setInterval,				// 2
	// Need to be at the end.
	&ASInternal::addProperty,		
	&ASInternal::registerClass,	
	&ASInternal::AStoString,		
	&ASInternal::unwatch,			
	&ASInternal::valueOf,			
	&ASInternal::watch			

};

/**
 * The constructor
 **/
ASInternal::ASInternal(Script *s, Graphics *g) : ASObject(s,g,L"Internal") {
}

/**
 * The destructor
 **/
ASInternal::~ASInternal() {
}

/**
 * Return a function descriptor
 */
Function *ASInternal::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"Internal") 			{func->setFunctionIndex(0); func->setNumParams(0); }	// Should not be called!
	if (memberName == L"ASSetPropFlags")	{func->setFunctionIndex(1); func->setNumParams(0); }
	if (memberName == L"setInterval")		{func->setFunctionIndex(2); func->setNumParams(10); }	// unlimited
	func = finishFunctionSetup(memberName,3,func);
	return func;
}
 
/**
 * Call class built in functions
 * exotic c++ call, but fast
 **/
void ASInternal::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 *
 **/
void ASInternal::init() {
}

/**
 * Process internal ASSetPropFlags function
 * This function is used internally during the initializion stage for the predefined actionscript objects. It is used to hide an objects children from the for..in loop construct. The for..in loop construct iterates over all children of an object, this means it exposes both methods and properties of an object. It is also used to protect the predefined actionscript objects children from being over-written by another action with the same name and to protect the predefined actionscript objects from being deleted. The usefulness of this function can also be harnessed by the developer and that is when this function gets interesting...
 *Value BitFlag CanOverWrite CanDelete IsHidden
 * 0	000		no				no		no
 * 1	001		no 				no		yes
 * 2	010		no				yes		no
 * 3	011		no				yes		yes
 * 4	100		yes				no		no
 * 5	101		yes				no		yes
 * 6	110		yes				yes		no
 * 7	111		yes				yes		yes
 *
 * http://www.flashguru.co.uk/000037.php
**/
void ASInternal::ASSetPropFlags() {
	//printf("Internal function ASSetPropFlags ignored!");
	//IntegerObject *numArgs = (IntegerObject*)script->pop();
	// numArgs should be 3 (could be 4)
	//IntegerObject *hideMask = (IntegerObject*)script->pop();
	//StringObject  *props    = (StringObject*) script->pop();
	//StringObject  *obj		= (StringObject*) script->pop();
}

/**
 *
 **/
void ASInternal::setInterval() {
	if (numPassedParams > 1) {
		wstring functionName = script->getRegister(1)->toString();
		int interval = int(script->getRegister(2)->toNumeric());
		list<ValueObject*> l;
		for (int i=3; i< numPassedParams; i++) l.push_back(script->getRegister(i)->clone());
		CallBackRequest *c = new CallBackRequest(getObjectStore(),functionName,interval,l);
		script->getEventManager()->addTimerCallbackEntry(c);
	}
	noReturnValue();
}
