/**
Action Script library. MovieClip class definition
Copyright (C) 2005 Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Bston, MA  02111-1307  USA
**/
#ifndef _AS_MOVIE_CLIP_HH
#define _AS_MOVIE_CLIP_HH
#include "as_object.hh"

class ASMovieClip : public ASObject{
public:
	ASMovieClip(ASMovieClip *);
	ASMovieClip(Script *s, Graphics *g);
	ASMovieClip(Script *s, Graphics *g, wstring className);
	~ASMovieClip();
	Function *getFunctionMember(wstring memberName);
	void 	callFunction(Function *f);
	virtual void 	init();
	virtual ASInterface*getNewInstance(Script *s,Graphics *g) { return new ASMovieClip(s,g); }
	virtual void 	setMainMember(ValueObject  *v);
	virtual ValueObject*getMainMember();
	virtual ASInterface* getObjectDeepCopy(bool deep);
	virtual ValueObject*getMember(wstring memberName);

	virtual void	attachAudio();
	virtual void	attachMovie();
	virtual void	createEmptyMovieClip();
	virtual void	createTextField();
	virtual void	duplicateMovieClip();
	virtual void	ASgetBounds();
	virtual void	getBytesLoaded();
	virtual void	getBytesTotal();
	virtual void	ASgetDepth();
	virtual void	getInstanceAtDepth();
	virtual void	getNextHighestDepth();
	virtual void	getSWFVersion();
	virtual void	getTextSnapshot();
	virtual void	getURL();
	virtual void	globalToLocal();
	virtual void	gotoAndPlay();
	virtual void	gotoAndStop();
	virtual void	hitTest();
	virtual void	loadMovie();
	virtual void	loadVariables();
	virtual void	localToGlobal();
	virtual void	nextFrame();
	virtual void	play();
	virtual void	prevFrame();
	virtual void	removeMovieClip();
	virtual void	setMask();
	virtual void	startDrag();
	virtual void	stop();
	virtual void	stopDrag();
	virtual void	swapDepths();
	virtual void	unloadMovie();
	virtual void	beginFill();
	virtual void	beginGradientFill();
	virtual void	clear();
	virtual void	curveTo();
	virtual void	endFill();
	virtual void	lineStyle();
	virtual void	lineTo();
	virtual void	moveTo();


	// Events
	void	onData();
	void	onDragOut();
	void	onDragOver();
	void	onEnterFrame();
	void	onKeyDown();
	void	onKeyUp();
	void	onKillFocus();
	void	onLoad();
	void	onMouseDown();
	void	onMouseMove();
	void	onMouseUp();
	void	onPress();
	void	onRelease();
	void	onReleaseOutside();
	void	onRollOut();
	void	onRollOver();
	void	onSetFocus();
	void	onUnload();

	void	setShape(ShapeWithStyle *p);
	virtual void	toGraphic();
	void	setRatio(float p) { ratio = p; needRedraw = true;}
	void 	setMorphShape(DefineMorphShapeTag* t);
	virtual void	changeColor(RGB c);
	virtual void	computeBounds();
protected:
	ShapeWithStyle  *shape;
	void	commonInit();
private:
	SI32	lastX,lastY;
	bool	isMorphShape;
	float	ratio;
	DefineMorphShapeTag* morphShape;
};

#endif
