/**
Action Script library. Base Action Script Object
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_object.hh"

typedef void (ASObject::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASObject::init,			// 0
	&ASObject::addProperty,		// 1
	&ASObject::registerClass,	// 2
	&ASObject::AStoString,		// 3
	&ASObject::unwatch,			// 4
	&ASObject::valueOf,			// 5
	&ASObject::watch			// 6
};

/**
 * The constructor
 **/
ASObject::ASObject(Script *s, Graphics *g) : ASInterface(s,g,L"Object") {
}
ASObject::ASObject(Script *s, Graphics *g, wstring className) : ASInterface(s,g,className) {
}
ASObject::ASObject(ASObject *p) : ASInterface(p) {
	this->getFuncs = p->getFuncs;
	this->setFuncs = p->setFuncs;
}


/**
 * The destructor
 **/
ASObject::~ASObject() {
	getFuncs.clear();
	setFuncs.clear();
}

/**
 * Return a function descriptor
 */
Function *ASObject::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"Object")	{func->setFunctionIndex(0); func->setNumParams(1); }
	func = finishFunctionSetup(memberName,1,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASObject::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 *
 **/
void ASObject::init() {
	printf("ASObject init");
}

/**
 * Creates a getter/setter property on an object.
 **/
void ASObject::addProperty() {
	if (numPassedParams > 1) {
		wstring prop = script->getRegister(1)->toString();
		ScriptObject *o = (ScriptObject*)script->getRegister(2);
		Function *func;
		if (o->getObjectTypeName() == L"LiteralFunc") func = ((FunctionLiteral*)o)->getFunction();
		else func = getObjectStore()->getFunction(L"_global",o->toString());
		getFuncs[prop] = func;
	}
	if (numPassedParams > 2) {
		wstring prop = script->getRegister(1)->toString();
		ScriptObject *o = (ScriptObject*)script->getRegister(3);
		Function *func;
		if (o->getObjectTypeName() == L"LiteralFunc") func = ((FunctionLiteral*)o)->getFunction();
		else func = getObjectStore()->getFunction(L"_global",o->toString());
		setFuncs[prop] = func;
	}
	noReturnValue();
}

/**
 * Associates a movie clip symbol with an ActionScript object class.
 **/
void ASObject::registerClass() {
	puts("registerClass TODO");
	noReturnValue();
}

/**
 * Converts the specified object to a string and returns it.
 **/
void ASObject::AStoString() {
	puts("AStoString TODO");
	noReturnValue();
}

/**
 * Removes the watchpoint that Object.watch() created.
 **/
void ASObject::unwatch() {
	puts("unwatch TODO");
	noReturnValue();
}

/**
 * Returns the primitive value of an object.
 **/
void ASObject::valueOf() {
	puts("valueOf TODO");
	noReturnValue();
}

/**
 * Registers an event handler to be invoked when a specified property of an ActionScript object changes.
 **/
void ASObject::watch() {
	puts("watch TODO");
	noReturnValue();
}
