/**
Action Script library. Selection class implementation.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net) 

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_selection.hh"

typedef void (ASSelection::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASSelection::init,				// 0
	&ASSelection::addListener,		// 1
	&ASSelection::getBeginIndex,	// 2
	&ASSelection::getCaretIndex,	// 3
	&ASSelection::getEndIndex,		// 4
	&ASSelection::getFocus,			// 5
	&ASSelection::removeListener,	// 6
	&ASSelection::setFocus,			// 7
	&ASSelection::setSelection,		// 8
	// Need to be at the end.
	&ASSelection::addProperty,		
	&ASSelection::registerClass,	
	&ASSelection::AStoString,		
	&ASSelection::unwatch,			
	&ASSelection::valueOf,			
	&ASSelection::watch
};

/**
 * The constructor
 **/
ASSelection::ASSelection(Script *s, Graphics *g) : ASObject(s,g,L"Selection") {
	caret.isActive = false;
}

/**
 * The destructor
 **/
ASSelection::~ASSelection() {
}

/**
 * Return a function descriptor
 */
Function *ASSelection::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"Selection") 	{func->setFunctionIndex(0); func->setNumParams(0); }
	if (memberName == L"addListener") 	{func->setFunctionIndex(1); func->setNumParams(1); }
	if (memberName == L"getBeginIndex") {func->setFunctionIndex(2); func->setNumParams(0); }
	if (memberName == L"getCaretIndex") {func->setFunctionIndex(3); func->setNumParams(0); }
	if (memberName == L"getEndIndex") 	{func->setFunctionIndex(4); func->setNumParams(0); }
	if (memberName == L"getFocus") 		{func->setFunctionIndex(5); func->setNumParams(0); }
	if (memberName == L"removeListener"){func->setFunctionIndex(6); func->setNumParams(1); }
	if (memberName == L"setFocus") 		{func->setFunctionIndex(7); func->setNumParams(1); }
	if (memberName == L"setSelection") 	{func->setFunctionIndex(8); func->setNumParams(2); }
	finishFunctionSetup(memberName,9,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASSelection::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 * Selection singleton instance initialization.
 **/
void ASSelection::init() {
	puts("TODO");
}

/**
 * Selection singleton instance initialization.
 **/
void ASSelection::setFocusInternal(ASInterface *as, void *data) {
	puts("ASSelection::setFocusInternal");
	printWstring(as->getClassName());
	if (as->getClassName() == L"TextField") {
		EventDataCallback *e = (EventDataCallback*)data;
		float xt,yt;
		as->mouseTransform(e->x,e->y,xt,yt);
		//printf("%d %d \n",e->x,e->y);
	}
	puts("");
}

/**
 * Registers an object to receive notification when onSetFocus is invoked.
 **/
void ASSelection::addListener() {
	puts("TODO");
	noReturnValue();
}

/**
 * Returns the index at the beginning of the selection span. Returns -1 if there is no index or currently selected field.
 **/
void ASSelection::getBeginIndex() {
	puts("TODO");
	noReturnValue();
}

/**
 * Returns the current caret (insertion point) position in the currently focused selection span. 
 * Returns -1 if there is no caret position or currently focused selection span.
 **/
void ASSelection::getCaretIndex() {
	puts("TODO");
	noReturnValue();
}

/**
 * Returns the index at the end of the selection span. Returns -1 if there is no index or currently selected field.
 **/
void ASSelection::getEndIndex() {
	puts("TODO");
	noReturnValue();
}

/**
 * Returns the name of the variable for the currently focused text field. Returns null if there is no currently focused text field.
 **/
void ASSelection::getFocus() {
	puts("TODO");
	noReturnValue();
}

/**
 * Removes an object that was registered with addListener().
 **/
void ASSelection::removeListener() {
	puts("TODO");
	noReturnValue();
}

/**
 * Focuses the text field associated with the specified variable.
 **/
void ASSelection::setFocus() {
	puts("TODO");
	noReturnValue();
}

/**
 * Sets the beginning and ending indexes of the selection span.
 **/
void ASSelection::setSelection() {
	puts("TODO");
	noReturnValue();
}

/*
	REFACT THIS CODE

	IntInteractionPointerHashMap *m = objectStore->getInteractions();
	caret.isActive = false;
	for (IntInteractionPointerHashMap::iterator it = m->begin(); !found && it != m->end(); it++) {
		InteractionPointer *i = (*it).second;
		if (x > i->rect.xmin && x < i->rect.xmax && y > i->rect.ymin && y < i->rect.ymax) {
			found = true;
			int type = i->tag->getType();
			switch(type) {
				case stagDefineEditText: {
					DefineEditTextTag *t = (DefineEditTextTag*) i->tag;
					string key = string((char*)t->getVariableName()->getString());
					string s = ((StringObject*)objectStore->getLocalVariable(key))->getValue();
					caret = graphics->getCaretLocation(t,s,xt,yt);
					doTransform(caret.x,caret.y,i->matrix);
					caret.isActive = true;
					caret.tag = i->tag;
					printf("CLICK on edit text detected. Screen coordinates :%f %f\n+",caret.x, caret.y);
					break;
				}
				case stagDefineButton:
				case stagDefineButton2:
					puts("CLICK on button detected");
					break;
			}
		}
	}

	if (caret.isActive) {
		DefineEditTextTag *t = (DefineEditTextTag*) caret.tag;
		string skey = string((char*)t->getVariableName()->getString());
		string s = ((StringObject*)objectStore->getLocalVariable(skey))->getValue();
		switch (key) {
			case 0: { // shift
				break;
			}
			case 8: {
				if (caret.charOffset>0) {
					LocalFontEntry *fontEntry = localFontManager->getChararacterTexture(t->getFontID(),s[caret.charOffset-1]);
					s = s.substr(0,caret.charOffset-1) + s.substr(caret.charOffset);
					if (fontEntry != NULL) caret.x -= fontEntry->advance[0]*caret.height/64;
					caret.charOffset--;
				}
				break;
			}
			case 127:
				if (caret.charOffset < s.size()) s = s.substr(0,caret.charOffset) + s.substr(caret.charOffset+1);
				break;
			default:
				char c[2];
				c[0] = key;
				c[1] = 0;
				s = s.substr(0,caret.charOffset) + string(c) + s.substr(caret.charOffset);
				LocalFontEntry *fontEntry = localFontManager->getChararacterTexture(t->getFontID(),c[0]);
				if (fontEntry != NULL) caret.x += fontEntry->advance[0]*caret.height/64;
				caret.charOffset++;
				if (key == 13) {
					caret.y += caret.height;
					// caret.x = ...... todo
				}
				break;
		}
		objectStore->setVariable(skey, new StringObject(s));
		graphics->registerDynamicString(t,s);		
		//printf("key: %d\n",key);				
	}
*/
