/**
Action Script library.This represents the internal definition of a Sprite object
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#ifndef _AS_SPRITE_HH
#define _AS_SPRITE_HH
#include "as_object.hh"
#include "as_movie_clip.hh"
#include <map>

typedef map<int, ASObject*> IntASObjectMap;

class ASSprite : public ASMovieClip {
public:
	ASSprite(ASSprite*);
	ASSprite(Script *s, Graphics *g);
	~ASSprite();

	void	init() {puts("Dont use it!");};
	ASInterface*getNewInstance(Script *s,Graphics *g) { return new ASSprite(s,g); }
	void	init(DefineSpriteTag *t);
	list<BaseTag*>::iterator getCurrentTag();
	int		getCurrentTagType();
	void	goNextTag(bool nextFrame = false);
	bool	isAtEnd();
	void	setDepth(int d) { depth = d; }
	int 	getDepth() { return depth; }
	void	addCharacter(int depth,ASObject *p);
	void 	removeCharacter(int depth);
	void 	_removeCharacter(int depth);
	IntASObjectMap *getCharacters() { return characterDepth; }
	ASObject *getCharacter(int depth) { return (*characterDepth)[depth]; }
	void	actionGotoFrame(int frameNum);
	void	changeColor(RGB c);
	void	changeTransform(CXFORMWITHALPHA c);
	void	toGraphic();
	void 	setMainMember(ValueObject  *v);
	ValueObject*getMainMember();
	ValueObject*getMember(wstring memberName);
	int		getSpriteId() { return spriteID; }
	void	setMember(wstring memberName, ValueObject *v);
	int 	getFrameCount() { return frameCount; }
	ASInterface* getObjectDeepCopy(bool deep);
	void 	setChildrenProperty(wstring member, ValueObject *value);
	void	setModifiedByScript(bool p) { modifiedByScript = p; }
	bool	getModifiedByScript() { return modifiedByScript; }
	ScriptObject*getMovieByNameOrPath(wstring namePath);
	void	_addCharacter(int depth,ASObject *p);
	void	addCharacterNoRecurs(int depth,ASObject *p) { (*characterDepth)[depth] = p; }
	void	clearCharacters() { characterDepth->clear(); }

	//  MovieClip overloaded functions
	void	ASgetBounds();
	void	duplicateMovieClip();
	void	gotoAndPlay();
	void	gotoAndStop();
	void	hitTest();
	void	removeMovieClip();
	void	createEmptyMovieClip();
private:
	int		spriteID;
	int		frameCount;
	list<BaseTag*> *tags;
	long	_zz_bug;	// Memory fault if not here
	list<BaseTag*>::iterator currentTag;
	int 	depth;
	IntASObjectMap *characterDepth;
	RGB		color;
	bool	changeColorRequired;
	wstring	mainMember;
	float	v0,vf;
	bool	modifiedByScript;

	void	doChangeColor();
	float	getWidth(ASSprite *sprite);
	float	getHeight(ASSprite *sprite);
	float	getXmin(ASSprite *sprite);
	float	getXmax(ASSprite *sprite);
	float	getYmin(ASSprite *sprite);
	float	getYmax(ASSprite *sprite);
	
};

#endif
