/**
Action Script library. TextField class
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)  

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_text_field.hh"
#include "as_selection.hh"

typedef void (ASTextField::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASTextField::addListener,					// 0
	&ASTextField::getFontList,					// 1
	&ASTextField::getDepth,						// 2
	&ASTextField::getNewTextFormat,				// 3
	&ASTextField::getTextFormat,				// 4
	&ASTextField::removeListener,				// 5
	&ASTextField::removeTextField,				// 6
	&ASTextField::replaceSel,					// 7
	&ASTextField::setNewTextFormat,				// 8
	&ASTextField::setTextFormat,				// 9
	&ASTextField::replaceText,					// 10
	&ASTextField::onPress,						// 11
	&ASTextField::onReleaseOutside,				// 12
	// Need to be at the end.
	&ASTextField::addProperty,		
	&ASTextField::registerClass,	
	&ASTextField::AStoString,		
	&ASTextField::unwatch,			
	&ASTextField::valueOf,			
	&ASTextField::watch
};

/**
 * The constructor
 **/
ASTextField::ASTextField(Script *s, Graphics *g) : ASObject(s,g,L"TextField") {
	textFieldStatic = false;
	textFormat = new ASTextFormat(s,g);
	eventData = NULL;
}

ASTextField::ASTextField(ASTextField *p) : ASObject(p) {
}

/**
 * The destructor
 **/
ASTextField::~ASTextField() { 
	//puts("~ASTextField");
	//ScriptObject *so = (ScriptObject*)textFormat;
	//delete so; 
}

/**
 * Deep copy constructor
 **/
ASInterface* ASTextField::getObjectDeepCopy(bool deep) {
	ASTextField *newObj = new ASTextField(this);
	newObj->textFieldStatic = textFieldStatic;
	newObj->staticTag = staticTag;
	newObj->textFormat = textFormat;
	newObj->eventData = eventData;
	return newObj;
}

/**
 * Return a function descriptor
 */
Function *ASTextField::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"addListener") 		{func->setFunctionIndex(0); func->setNumParams(1); }
	if (memberName == L"getFontList") 		{func->setFunctionIndex(1); func->setNumParams(0); }
	if (memberName == L"getDepth") 			{func->setFunctionIndex(2); func->setNumParams(0); }
	if (memberName == L"getNewTextFormat") 	{func->setFunctionIndex(3); func->setNumParams(0); }
	if (memberName == L"getTextFormat") 	{func->setFunctionIndex(4); func->setNumParams(2); }
	if (memberName == L"removeListener") 	{func->setFunctionIndex(5); func->setNumParams(1); }
	if (memberName == L"removeTextField") 	{func->setFunctionIndex(6); func->setNumParams(0); }
	if (memberName == L"replaceSel") 		{func->setFunctionIndex(7); func->setNumParams(1); }
	if (memberName == L"setNewTextFormat") 	{func->setFunctionIndex(8); func->setNumParams(1); }
	if (memberName == L"setTextFormat") 	{func->setFunctionIndex(9); func->setNumParams(3); }
	if (memberName == L"replaceText") 		{func->setFunctionIndex(9); func->setNumParams(3); }
	if (memberName == L"onPress") 			{func->setFunctionIndex(10); func->setNumParams(1); }
	if (memberName == L"onReleaseOutside") 	{func->setFunctionIndex(11); func->setNumParams(1); }
	func = finishFunctionSetup(memberName,12,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASTextField::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 * Populate a static text field
 **/
void ASTextField::setStaticTextField(DefineTextTag *tag) {
	textFieldStatic = true;
	//TODO: setMember("length",tag-> sum of text records
	staticTag = tag;
	needRedraw = true;
}

/**
 * Set dynamic text
 **/
void ASTextField::setDynamicText(DefineEditTextTag *t) {
	StringObject *s;
	if (t->getHasText()) s = new StringObject(t->getInitialText()->getString());
	else s = new StringObject(L"");
	setMemberNoClone(L"text",s);
	setMemberNoClone(L"length", new IntegerObject(s->toString().size()));
	setMemberNoClone(L"wordWrap",new BooleanObject(t->getWordWrap()));
	setMemberNoClone(L"multiline",new BooleanObject(t->getMultiline()));
	setMemberNoClone(L"password",new BooleanObject(t->getPassword()));
	setMemberNoClone(L"type",t->getReadOnly() ? new StringObject(L"dynamic") : new StringObject(L"input"));
	setMemberNoClone(L"textColor",t->getHasTextColor() ? new IntegerObject(RGBA2INT(t->getTextColor())) : new IntegerObject(0));
	setMemberNoClone(L"_alpha",t->getHasTextColor() ? new IntegerObject(t->getTextColor()->alpha) : new IntegerObject(255));
	setMemberNoClone(L"maxChars",t->getHasMaxLength() ? new IntegerObject(t->getMaxLength()) : new IntegerObject(0));
	// This property is not very clear....
	wstring autos = L"none";
	if (t->getAutoSize() && t->getHasLayout()) {
		switch (t->getAlign()) {
			case 0: autos = L"left"; break;
			case 1: autos = L"right"; break;
			case 2: autos = L"center"; break;
			case 3: autos = L"justify"; break;
		}
	}
	setMemberNoClone(L"autoSize", new StringObject(autos));
	setMemberNoClone(L"selectable",new BooleanObject(t->getNoSelect()));
	setMemberNoClone(L"border",new BooleanObject(t->getBorder()));
	setMemberNoClone(L"html",new BooleanObject(t->getHTML()));
	setMemberNoClone(L"embedFonts",new BooleanObject(t->getUseOutlines()));
	setMemberNoClone(L"align",new BooleanObject(t->getAlign())); // ?????
	setMemberNoClone(L"fontId",new IntegerObject(t->getFontID())); // non specified member
	textFormat->setMemberNoClone(L"size",t->getHasFont() ? new IntegerObject(t->getFontHeight()): new IntegerObject(0));
	textFormat->setMemberNoClone(L"leftMargin",t->getHasLayout() ? new IntegerObject(t->getLeftMargin()) : new IntegerObject(0));
	textFormat->setMemberNoClone(L"rightMargin",t->getHasLayout() ? new IntegerObject(t->getRightMargin()) : new IntegerObject(0));
	textFormat->setMemberNoClone(L"indent",t->getHasLayout() ? new IntegerObject(t->getIndent()) : new IntegerObject(0));
	textFormat->setMemberNoClone(L"leading",t->getHasLayout() ? new IntegerObject(t->getLeading()) : new IntegerObject(0));
	bounds = t->getBounds();
	textFieldStatic = false;
	needRedraw = true;
}

/**
 *
 **/
void ASTextField::init() {
	puts("TODO");
}

/**
 * Registers an object to receive notification when the onChanged and onScroller event handlers are invoked.
 **/
void ASTextField::addListener() {
	puts("TODO");
	noReturnValue();
}

/**
 * Returns names of fonts on the player’s host system as an array.
 **/
void ASTextField::getFontList() {
	puts("TODO");
	noReturnValue();
}

/**
 * Returns the depth of a text field.
 **/
void ASTextField::getDepth() {
	puts("TODO");
	noReturnValue();
}

/**
 * Gets the default text format assigned to newly inserted text.
 **/
void ASTextField::getNewTextFormat() {
	puts("TODO");
	noReturnValue();
}

/**
 * Returns a TextFormat object containing formatting information for some or all text in a text field.
 **/
void ASTextField::getTextFormat(){
	script->push(textFormat);
};

/**
 *  Removes a listener object.
 **/
void ASTextField::removeListener() {
	puts("TODO");
	noReturnValue();
}

/**
 * Removes a text field that was created with MovieClip.createTextField().
 **/
void ASTextField::removeTextField() {
	puts("TODO");
	noReturnValue();
}

/**
 * Replaces the current selection.
 **/
void ASTextField::replaceSel() {
	puts("TODO");
	noReturnValue();
}

/**
 * Sets a TextFormat object for text that is inserted by a user or by a method.
 **/
void ASTextField::setNewTextFormat() {
	puts("TODO");
	noReturnValue();
}

/**
 *  Sets a TextFormat object for a specified range of text in a text field.
 **/
void ASTextField::setTextFormat() {
	puts("TODO");
	noReturnValue();
}

/**
 *
 **/
void ASTextField::replaceText() {
	puts("TODO");
	noReturnValue();
}

/**
 * Convert to TextField to graphic instructions. 
 **/
void ASTextField::toGraphic() {
	if (needRedraw) {
		//printf("ASTextField::toGraphic %x %d\n",objectUnid,needRedraw);getMember("text")->print();
		if (graphics->getSequence(objectUnid) != NULL)	graphics->clearSequence(objectUnid);
		if (textFieldStatic) {
			graphics->registerStaticString(objectUnid,staticTag);
		} else {
			graphics->registerDynamicString(this);
		}
		needRedraw = false;
	}
}

/**
 * Set the main object property
 **/
void ASTextField::setMainMember(ValueObject  *s) {
	setMember(L"text",s);
	needRedraw = true;
}


/**
 * Get the main object property
 **/
ValueObject* ASTextField::getMainMember() {
	return getMember(L"text");
}

/**
 * Dynamic text field event call back
 **/
bool ASTextField::callInternalEvent(string eventName, void *eventData) {
	if (eventData != NULL) {
		this->eventData = eventData;
	}
	funcPtr f = eventDispatcher(eventName);
	if (f != NULL) {
		(this->*f)();
		needRedraw = true;
	}
	return needRedraw;
}
/**
 * Handler event dispatcher
 **/
funcPtrASTextField	ASTextField::eventDispatcher(string eventName) {
	if (eventName == "onPress") 	return &ASTextField::onPress;
	if (eventName == "onReleaseOutside") return &ASTextField::onReleaseOutside;
	return NULL;
}

/**
 * Mouse click inside the text area
 **/
void ASTextField::onPress(){
	puts("ASTextField::onPress");
	ASSelection *as = (ASSelection*)getObjectStore()->getGlobalVariable(L"_SelectionSingleton");
	if (as == NULL) {
		as = new ASSelection(script,graphics);
		as->init();
		printf("+Create Selection Instance\n");
		getObjectStore()->setGlobalVariable(L"_SelectionSingleton",as);
	}
	as->setFocusInternal(this,eventData);
}

/**
 * Mouse click outside text area
 **/
void ASTextField::onReleaseOutside() {
	puts("ASTextField::onReleaseOutside()");
}

/**
 * Returns the string as ... as string
 **/
wstring ASTextField::toString() const{
	ValueObject *v =members->getMember(L"text");
	wstring res = v->toString();
	script->removeObject(v);
	return res;
}



