/**
Macromedia Flash Processed Object Store.
Copyright (C) 2005 Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include <iostream>
#include "object_store.hh"
#include "../ASLib/as_interface.hh"
#include "../ASLib/as_sprite.hh"

extern bool debug;
typedef multimap <long,ScriptObject*>::iterator RCIT; 
typedef pair<RCIT, RCIT> RRange;

/**
 * The contructor
 **/
ObjectStore::ObjectStore(Graphics *graphics) {
	currentTargetPath = L"/";
	this->graphics = graphics;
	globalRegisters = new RegisterSet(4);
	rootScript = new ScriptObject(L"_global");
	currentActionContext = rootScript;
	poolActionContext = L"_global";
	objectAtDepth.clear();
}

/**
 * Destructor
 **/
ObjectStore::~ObjectStore(){
	//puts("~ObjectStore");
	// clear functions
	for (StringFunctionHashMap::iterator it=functions.begin(); it != functions.end(); it++) delete (*it).second;
	functions.clear();

	// Clear local variables
	for (StringValueObjectHashMapHashMap::iterator it=variables.begin(); it != variables.end(); it++) {
		StringValueObjectHashMap *m = (*it).second;
		if (m != NULL) {
			for (StringValueObjectHashMap::iterator it2=m->begin(); it2 != m->end(); it2++) {
				ValueObject *v = (*it2).second;
				removeObject(v);
			}
			m->clear();
			delete m;
		}
	}
	variables.clear();

	// Clear constants
	for (StringIntValueObjectHashMapHashMap::iterator it=constantPools.begin(); it != constantPools.end(); it++) {
		IntValueObjectHashMap *m = (*it).second;
		if (m!=NULL) {
			for (IntValueObjectHashMap::iterator it2=m->begin(); it2 != m->end(); it2++) removeObject((*it2).second);
			m->clear();
			delete m;
		}
	}
	constantPools.clear();

	// clear rendered objects
	for (map<ScriptObject*,ScriptObject*>::const_iterator it=allocatedObjects.begin(); it != allocatedObjects.end(); it++) {
		ScriptObject *m = (*it).second;
// CRASH !!!! 		if (m != NULL) delete m;
	}

	delete globalRegisters;
	delete rootScript;
	allocatedObjects.clear();
}

/**
 * Go to a new level for local variables
 * Use object pointer address as a unique identifier, 
 * since the movie name may be used multiple times, 
 * if this object is placed more than once.
 **/
void ObjectStore::targetPathDown(ScriptObject *newLevel) {
	if (currentTargetPath == L"/") currentTargetPath = L"/" + buildKey(newLevel);
	else currentTargetPath = currentTargetPath +  L"/" + buildKey(newLevel);
}

/**
 * Go UP one level for local variables
 **/
void ObjectStore::targetPathUp() {
	int po = currentTargetPath.find_last_of('/');
	if (po != string::npos) currentTargetPath = currentTargetPath.substr(0,po);
	if (currentTargetPath == L"") currentTargetPath = L"/" ;
}

/**
 * Find the context of a variable
 * 
 * Try at current local scope, the goes up to /
 * if it is not found, try on the Sprite object
 */
wstring	ObjectStore::findVariableContext(ScriptObject *so, wstring key) {
	StringValueObjectHashMap *s = variables[currentTargetPath];
	if (s != NULL && (*s)[key] != NULL) {
		return currentTargetPath;
	}
	wstring sc = currentTargetPath;
	int po = sc.find_last_of('/');
	if (po != string::npos) sc = sc.substr(0,po);
	while (po != string::npos) {
		s = variables[sc];
		if (s != NULL && (*s)[key] != NULL) {
			return sc;
		}
		po = sc.find_last_of('/');
		if (po != string::npos && po > 0) sc = sc.substr(0,po-1);
		if (po == 0) po =string::npos; 
	}
	// if not found, return SpriteObject context
	return L"/"+buildKey(so);
}

/**
 * Add a local variable to the current local level (ie: /A/B/foo)
 **/
void ObjectStore::addLocalVariable(wstring key, ValueObject *v) {
	StringValueObjectHashMap *s = variables[currentTargetPath];
	if (s == NULL) {
		s = new StringValueObjectHashMap();
		variables[currentTargetPath] = s;
	} 
	(*s)[key] = v;
}

/**
 * Set a variable.
 **/
void ObjectStore::setVariable(ScriptObject *so, wstring key, ValueObject *v) {
	// Check if we reference a global variable
	int pos = key.find(L"_level0/");
	if (pos != string::npos) {			// That's global variable
		key = key.substr(pos+8);
		// Check if the variable contains the ":" string. In that case, it's eqyuvalent to a setProperty.
		pos = key.find(L":");
		if (pos != string::npos) {
			wstring var = key.substr(0,pos);
			wstring member = key.substr(pos+1);
			wstring uvar = var;
			pos = uvar.find(L"/");
			if (pos != string::npos) {
				uvar = uvar.substr(0,pos);
				member = var.substr(pos)+L"/"+member;
			}
			if (debug) wcout << L"set global member " << uvar << L"[" << member<<L"]";
			ASInterface *g = (ASInterface*)getGlobalVariable(uvar);
			g->setChildrenProperty(member,v);								// Set a property on a contained object
		} else {
			puts("setVariable :: FIXME!!!!");
		}
	} else {
		wstring ctx = findVariableContext(so,key);
		StringValueObjectHashMap *s = variables[ctx];
		if (s == NULL) {
			s = new StringValueObjectHashMap();
			variables[ctx] = s;
		} 
		ValueObject *ve = (*s)[key];
		if (ve != NULL) {
			if (ve->getObjectType() != ScriptObjectType) delete ve;
			else {
				ASObject * a = (ASObject*) ve;
				if (a->getClassName() != L"Sprite" && a->getClassName() != L"MovieClip") {
					delete a;
				}
			}
		}
		if(debug) wcout << L"Add local: " << ctx << L"/"<<key <<L"="<<*v;
		(*s)[key] = v->clone();
	}
}

/**
 * Remove a local variable
 **/
void ObjectStore::removeLocalVariable(wstring key) {
	StringValueObjectHashMap *s = variables[currentTargetPath];
	if (s != NULL) {
		ValueObject *v = (*s)[key];
		if (v !=NULL) removeObject(v);
		(*s).erase(key);
	}
}

/**
 * Read a variable 
 **/
ValueObject *ObjectStore::getVariable(ScriptObject *so, wstring key) {
	//printf("context : %x\n",so);
	wstring scope = findVariableContext(so,key);
	StringValueObjectHashMap *s = variables[scope];
	if (s != NULL) {
		ValueObject *v = (*s)[key];
		if (v != NULL) return v->clone();
	}
	ValueObject *v = so->getMembers()->getMember2(key);
	if (v != NULL) return v->clone();
	
	ASInterface *amc = (ASInterface *)so;
	if (amc->getParent() != NULL) {
		if (debug)puts("Search for PARENTS");
		// Find All parents. TODO
		ASInterface *parent = amc->getParent();
		v = parent->getMembers()->getMember2(key);
		if (v != NULL) return v->clone();
	}
	return NULL;
}

/**
 * Returns a Global variable
 **/
ValueObject *ObjectStore::getGlobalVariable(wstring key) {
	if (key.find('/') == 0) key = key.substr(1);
	StringValueObjectHashMap *s = variables[L"/"];
	ValueObject *v = (*s)[key];
	if (v != NULL) return v->clone();
	return NULL;
}

/**
 * Set a Global variable
 **/
void ObjectStore::setGlobalVariable(wstring key,ValueObject *v) {
	StringValueObjectHashMap *s = variables[L"/"];
	if (s == NULL) {
		s = new StringValueObjectHashMap();
		variables[L"/"] = s;
	} 
	(*s)[key] = v->clone();
}

/**
 * Clean up current level (when we leave a function)
 **/
void ObjectStore::cleanCurrentLevel() {
/*
FIXME: what is the policy of variable cleaning?

wcout << L"clean " << currentTargetPath << L"\n";
	if (currentTargetPath == L"/") return;
	StringValueObjectHashMap *s = variables[currentTargetPath];
	if (s != NULL) {
		for (StringValueObjectHashMap::iterator it2=s->begin(); it2 != s->end(); it2++) {
			ValueObject *v = (*it2).second;
			if (v != NULL) removeObject(v);
		}
		s->clear();
	}
	variables.erase(currentTargetPath);
*/
}

/**
 * Set a constant value (in the Constant Pool).
 */
void ObjectStore::setConstant(int entry, ValueObject *p) {
	IntValueObjectHashMap *m = constantPools[poolActionContext];
	if (m==NULL) {
		m = new IntValueObjectHashMap();
		constantPools[poolActionContext] = m;
	}
	ValueObject *c = (*m)[entry];
	if (c != NULL) delete c;
	(*m)[entry] = p;
}

/**
 * Get constant pool value
 */
ValueObject *ObjectStore::getConstant(int entry) {
	IntValueObjectHashMap *m = constantPools[poolActionContext];
	if (m != NULL) {
		ValueObject *c = (*m)[entry];
		if (c!=NULL) return c->clone();
		return new StringObject(L"undefined");
	} else {
		puts("SOMETHING WRONG in ObjectStore::getConstant!!!!");
		return new StringObject(L"undefined");
	}
}

/**
 * Removes an object testing if it can be removed (ie not a ScriptObject)
 **/
void ObjectStore::removeObject(ValueObject *v) {
	if (v!=NULL) {
		if (v->getObjectType() != ScriptObjectType) delete v;
		else {
			ScriptObject *so = (ScriptObject*) v;
			if (so->getShouldBeRemoved()) delete v;		// That's when the object is explicitly removed from actionScript
		}
	}
}

/**
 * Get a function. We make some class transforms if we have base classes.
 **/
Function* ObjectStore::getFunction(wstring className,wstring functionName) {
	if (className == L"_$$$String") className=L"String";
	if (className == L"_$$$Number") className=L"Number";
	return functions[className+L"."+functionName];
}

void ObjectStore::addFunction(wstring className,wstring functionName, Function *f) { 
	functions[className+L"."+functionName] = f;
}

/**
 * returns a movie given its name or its path
 *
 **/
ScriptObject*ObjectStore::getMovieByNameOrPath(wstring namePath) {
	if (namePath.find('/') == 0) namePath = namePath.substr(1);
	int po = namePath.find('/');
	ScriptObject*so = NULL;
	if (po == string::npos) {			// it's a simple movie name
		so = (ScriptObject*)getGlobalVariable(namePath);
	} else {
		wstring main = namePath.substr(0,po);
		ASSprite *mo = (ASSprite *)getGlobalVariable(main);
		if (mo != NULL) so = mo->getMovieByNameOrPath(namePath.substr(po+1));
	}
	if (so != NULL && so->getObjectType() != ScriptObjectType) so=NULL;
	return so;
}

/**
 * Set sprite scope variable
 **/
void ObjectStore::setSpriteVariable(ScriptObject *so,wstring key, ValueObject *v) {
	//if(debug)puts("setSpriteVariable\n");
	wstring targetPath = L"/"+buildKey(so);
	StringValueObjectHashMap *s = variables[targetPath];
	if (s == NULL) {
		s = new StringValueObjectHashMap();
		variables[targetPath] = s;
	} 
	(*s)[key] = v->clone();
}

/**
 * get sprite scope variable
 **/
ValueObject*ObjectStore::getSpriteVariable(ScriptObject *so, wstring key) {
	wstring ctx = L"/"+buildKey(so);
	StringValueObjectHashMap *s = variables[ctx];
	if (s == NULL) {
		s = new StringValueObjectHashMap();
		variables[ctx] = s;
	} 

	ValueObject *v = (*s)[key];
	if (v != NULL) return v->clone();
	else return NULL;
}

/**
 * remove a sprite variable
 **/
void ObjectStore::removeSpriteVariable(ScriptObject *so,wstring key){
	wstring ctx = L"/"+buildKey(so);
	StringValueObjectHashMap *s = variables[ctx];
	s->erase(key);
}

/**
 * Add a reference to a cloned Object
 **/
void ObjectStore::addClonedObject(int characterId, ScriptObject *so) {
	// Check if already associated.
	RRange range = clonedObjects.equal_range(characterId);
	if (range.first != clonedObjects.end()) {
		for(RCIT i=range.first; i!=range.second; ++i) {
			if (i->second == so) return;
		}
	}
	clonedObjects.insert(make_pair(characterId,so));
}

/**
 * Remove a cloned Object
 **/
void ObjectStore::removeClonedObject(int characterId, ScriptObject *so) {
	RRange range = clonedObjects.equal_range(characterId);
	if (range.first != clonedObjects.end()) {
		for(RCIT i=range.first; i!=range.second; ++i) {
			if (i->second == so) clonedObjects.erase(i);
		}
	}
}

/**
 * Returns the list of cloned object for a given characterId
 **/
list<ScriptObject*>ObjectStore::getClonedObjects(int characterId) {
	list<ScriptObject*> res;
	RRange range = clonedObjects.equal_range(characterId);
	if (range.first != clonedObjects.end()) {
		for(RCIT i=range.first; i!=range.second; ++i) {
			res.push_back(i->second);
		}
	}
	return res;
}

/**
 * Add a removed object - PlaceFlagMove && PlaceFlagHasCharacter
 * For future reuse
 **/
void ObjectStore::addDeletedObject(ScriptObject *parent,int characterId, ScriptObject *so){
	char tmp[100]; 
	sprintf(tmp,"%x:%d",parent,characterId); 
	string key = string(tmp);
	rubishStore[key] = so;
}

/**
 * Get a removed object
 **/
ScriptObject *ObjectStore::getDeletedObject(ScriptObject *parent,int characterId){
	char tmp[100]; 
	sprintf(tmp,"%x:%d",parent,characterId); 
	string key = string(tmp);
	ScriptObject *so = rubishStore[key];
	if (so!= NULL) rubishStore.erase(key);
	return so;
}

/**
 * Remove an object - call the destructor of the object
 **/
void ObjectStore::destroyObject(ScriptObject*s) {
	allocatedObjects.erase(s);
	delete s;
}

/**
 * Remove a global variable
 **/
void ObjectStore::removeGlobalVariable(wstring key) {
	StringValueObjectHashMap *s = variables[L"/"];
	s->erase(key);
}

// find something cleaner.
wstring	encodeCharacterName(int charId) { 
	char tmp[100]; 
	sprintf(tmp,"_nonAS:%d",charId); 
	return stringTowstring(tmp); 
}

