/**
Macromedia Flash File decoder.
Copyright (C) 2004  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "swf_decoder.hh"
#include <zlib.h>

const bool debug=true;

/**
 * Decode a bloc of data passed by the browser/flashplayer 
 **/
bool Swf_decoder::readStream(char *data, int size) {
	// First call
	if (needHeader && size > 21) {
		// Initialize the header
		memcpy(&header,data,8);
		UI8 *s = (UI8*)&header.fileLength;
		header.fileLength = (UI32) s[0] | ((UI32) s[1] << 8) | ((UI32) s[2] << 16) | ((UI32) s [3] << 24);
		decodeBuffer = (char *)malloc(header.fileLength-8);
		posBuff = 0;
		needHeader = false;
		if (header.signature1 == 'C') {
			// initialize the decompresssor and read first bloc.
			zstream.zalloc = Z_NULL;
			zstream.zfree = Z_NULL;
			zstream.opaque = Z_NULL;
			zstream.avail_in = 0;
			zstream.next_in = Z_NULL;
			int status = inflateInit(&zstream);
			if (status != Z_OK) return false;

			m_lastSize = 8;
			zstream.avail_in = size - m_lastSize;
			m_lastSize = size;
			zstream.next_in = (Bytef *)data+8;
			zstream.next_out = (Bytef *)decodeBuffer;
			zstream.avail_out = (uInt)header.fileLength-8;
			status = inflate( &zstream, Z_SYNC_FLUSH );
			if ( status != Z_OK && status != Z_STREAM_END ) return false; // Decode error, should never happens
			if (header.fileLength-8 > zstream.total_out) return false;		// Request more bytes
		} else {
			if(size < header.fileLength)	return false; // Request more bytes
			else memcpy(decodeBuffer,data+8,size-8);
		}
	}
	// Subsequent calls
	if (!needHeader) {
		if (header.signature1 == 'C') {
			zstream.next_in = (Bytef *)data + m_lastSize;
			zstream.avail_in = size - m_lastSize;
			m_lastSize = size;
			int status = inflate( &zstream, Z_SYNC_FLUSH );
			if ( status != Z_OK && status != Z_STREAM_END ) return false; // Decode error, should never happens
			posBuff = zstream.total_out;
			zstreamInPos = size - zstream.avail_in;
			if (header.fileLength-8 > posBuff) return false;	// Request more bytes
		} else {
			if(size < header.fileLength)	return false; // Request more bytes
			else memcpy(decodeBuffer,data+8,size-8);
		}	
	}

	// We have finished decoding.
	posBuff = 0;
	if (debug) {
		// open ouput file
		f = fopen("/tmp/decode.txt","w");
	}

	int nbits = getBitFieldUnsigned(5);
	header.frameSize.xmin = getBitFieldSigned(nbits);
	header.frameSize.xmax = getBitFieldSigned(nbits);
	header.frameSize.ymin = getBitFieldSigned(nbits);
	header.frameSize.ymax = getBitFieldSigned(nbits);
	header.frameRate = get8p8();
	header.frameCount = getWord();
	if (debug) dumpHeader(f);
	curSpriteId = 0;
	bool fin = false;

	while( !fin) {
		UI16 tagHeader = getWord();
   		long len = tagHeader & 0x3f;
   	 	long code = tagHeader >> 6;
    	if (len == 0x3f)  len = getDWord();
		if (debug) fprintf(f,"HEX Code: %x  dec code:%d, taglen:%d  DecodePos:%x/%x\n",tagHeader,code,len,posBuff>>3,header.fileLength-8);
		void * d;
  		fin = (d =dispatcher(&tags,len,code)) == NULL;
	}	
	if (header.signature1 == 'C') inflateEnd(&zstream);
	free(decodeBuffer);
	if (debug) fclose(f);
	return true;
}

/**
 *
 * STREAM readers functions.
 *
 **/
void Swf_decoder::alignNextByte(){
	posBuff = (posBuff+7)& ~0x7;
}

UI8 Swf_decoder::getByte(void) {
	alignNextByte();
	int bytePos = posBuff >>3;
	posBuff += 8;
    return decodeBuffer[bytePos];
}

UI16 Swf_decoder::getWord(void) {
	alignNextByte();    
	UI8 * s = (UI8*)decodeBuffer + (posBuff >>3);
    posBuff += 16;
    return (UI16) s[0] | ((UI16) s[1] << 8);
}

UI32 Swf_decoder::getDWord(void) {
	alignNextByte();
    UI8 * s = (UI8*)decodeBuffer + (posBuff >>3);
    posBuff += 32;
    return (UI32) s[0] | ((UI32) s[1] << 8) | ((UI32) s[2] << 16) | ((UI32) s [3] << 24);
}

UI32 Swf_decoder::getBitFieldUnsigned(int length) {
	UI32 res = 0;
	for (int i=0; i< length; i++) {
		int loc = posBuff>>3;
		int pos = 7 - (posBuff - (loc << 3));
		res = (res << 1) | ((decodeBuffer[loc]>>pos)&1);
		posBuff++;
	}
	return res;
}

SI32 Swf_decoder::getBitFieldSigned(int length) {
	UI32 res = getBitFieldUnsigned(length);
    if (res & (1L << (length - 1)))  {
        res |= -1L << length;
    }
	return (SI32) res;
}

SI32 Swf_decoder::getSWord() {
	alignNextByte();
	SI32 tmp = getWord();
    if (tmp & (1L << (16 - 1)))  {
        tmp |= -1L << 16;
    }
	return tmp;
}

float Swf_decoder::get8p8(void) {
	UI16 s = getWord();
	return (s>>8)+ ((float)(s&0xf))/16.0;
}

/**
 *
 * BASIC OBJECTS DECODING FUNCTIONS.
 * These functions are called during tag decoding.
 *
 **/ 
void Swf_decoder::getRect(RECT *r) {
	int nbits = getBitFieldUnsigned(5);
	r->xmin = getBitFieldSigned(nbits);
	r->xmax = getBitFieldSigned(nbits);
	r->ymin = getBitFieldSigned(nbits);
	r->ymax = getBitFieldSigned(nbits);
	alignNextByte();
}

void Swf_decoder::getMatrix(MATRIX *m) {
	int ns;
	m->hasScale = getBitFieldUnsigned(1);
	if (m->hasScale == 1) {
		ns = getBitFieldUnsigned(5);
		m->scaleX = getBitFieldSigned(ns);
		m->scaleY = getBitFieldSigned(ns);
	} else {
		m->scaleX = 65536;
		m->scaleY = 65536;
	}
	m->hasRotate = getBitFieldUnsigned(1);
	if (m->hasRotate == 1) {
		ns = getBitFieldUnsigned(5);
		m->rotateSkew0 = getBitFieldSigned(ns);
		m->rotateSkew1 = getBitFieldSigned(ns);
	} else {
		m->rotateSkew0 = 0;
		m->rotateSkew1 = 0;
	}
	ns = getBitFieldUnsigned(5);
	m->translateX = getBitFieldSigned(ns);
	m->translateY = getBitFieldSigned(ns);
	alignNextByte();
}

void Swf_decoder::getLineStyles(LINESTYLE* l, int n, bool isShape3) {
	for (int i=0; i< n ; i++) {
		l[i].width = getWord();
		l[i].color.red = getByte();
		l[i].color.green = getByte();
		l[i].color.blue  = getByte();
		l[i].color.alpha = isShape3 ? getByte() : 255;
	}
}

void Swf_decoder::getMorphLineStyles(MORPHLINESTYLE* l, int n) {
	for (int i=0; i< n ; i++) {
		l[i].startWidth = getWord();
		l[i].endWidth = getWord();
		l[i].startColor.red = getByte();
		l[i].startColor.green = getByte();
		l[i].startColor.blue  = getByte();
		l[i].startColor.alpha = getByte();
		l[i].endColor.red = getByte();
		l[i].endColor.green = getByte();
		l[i].endColor.blue  = getByte();
		l[i].endColor.alpha = getByte();
	}
}

void Swf_decoder::getGradient(GRADIENT *g,  bool isShape3) {
	g->setNumGradients(getByte());
	for (int i=0; i< g->getNumGradients(); i++) {
		GRADRECORD *gr = new GRADRECORD();
		gr->setRatio(getByte());
		RGBA	color;
		color.red = getByte();
		color.green = getByte();
		color.blue  = getByte();
		color.alpha = isShape3 ? getByte() : 255;
		gr->setColor(color);
		g->addGradientRecord(gr);
	}
}

void Swf_decoder::getMorphGradient(MORPHGRADIENT *g) {
	g->setNumGradients(getByte());
	for (int i=0; i< g->getNumGradients(); i++) {
		MORPHGRADRECORD *gr = new MORPHGRADRECORD();
		gr->startRatio = getByte();
		gr->startColor.red = getByte();
		gr->startColor.green = getByte();
		gr->startColor.blue  = getByte();
		gr->startColor.alpha = getByte();
		gr->endRatio = getByte();
		gr->endColor.red = getByte();
		gr->endColor.green = getByte();
		gr->endColor.blue  = getByte();
		gr->endColor.alpha = getByte();
		g->addGradientRecord(gr);
	}
}

void Swf_decoder::getFillStyles(FILLSTYLE* f, int n, bool isShape3) {
	for (int i=0; i< n ; i++) {
		f[i].setFillStyleType(getByte());
		int type = f[i].getFillStyleType();
		if (type == 0x0 ) {
			RGBA c;
			c.red = getByte();
			c.green = getByte();
			c.blue  = getByte();
			c.alpha = isShape3 ? getByte() : 255;
			f[i].setColor(c);
		}
		if (type == 0x10 || type == 0x12) {
			getMatrix(f[i].getGradientMatrix());
			GRADIENT *d = new GRADIENT();
			getGradient(d, isShape3);			
			f[i].setGradient(d);
		}
		if (type == 0x40 || type == 0x41 || type == 0x42 || type == 0x43) {
			f[i].setBitmapId(getWord());
			getMatrix(f[i].getBitmapMatrix());
		}
	}
}

void Swf_decoder::getMorphFillStyles(MORPHFILLSTYLE* f, int n) {
	for (int i=0; i< n ; i++) {
		f[i].setFillStyleType(getByte());
		int type = f[i].getFillStyleType();
		if (type == 0x0 ) {
			RGBA c;
			c.red = getByte();
			c.green = getByte();
			c.blue  = getByte();
			c.alpha = getByte();
			f[i].setStartColor(c);
			c.red = getByte();
			c.green = getByte();
			c.blue  = getByte();
			c.alpha = getByte();
			f[i].setEndColor(c);
		}
		if (type == 0x10 || type == 0x12) {
			getMatrix(f[i].getStartGradientMatrix());
			getMatrix(f[i].getEndGradientMatrix());
			MORPHGRADIENT *d = new MORPHGRADIENT();
			getMorphGradient(d);			
			f[i].setGradient(d);
		}
		if (type == 0x40 || type == 0x41 || type == 0x42 || type == 0x43) {
			f[i].setBitmapId(getWord());
			getMatrix(f[i].getStartBitmapMatrix());
			getMatrix(f[i].getEndBitmapMatrix());
		}
	}
}

void Swf_decoder::getRecords(Shape *shape, bool isShape3) {
	while (1) {
		bool typeFlag = getBitFieldUnsigned(1);
		if (typeFlag == 0)	{				//  Non edge
			int end = getBitFieldUnsigned(5);
			if (end ==0) {
				alignNextByte();
				return;			// End of Shape Record
			}
			posBuff -=5;
			StyleChangeRecord *s = new StyleChangeRecord(); 
			s->setTypeFlag(StyleChangeRecordType);
			s->setStateNewStyles(getBitFieldUnsigned(1));		// used if shape > 1
			s->setStateLineStyle(getBitFieldUnsigned(1));
			s->setStateFillStyle1(getBitFieldUnsigned(1));
			s->setStateFillStyle0(getBitFieldUnsigned(1));
			s->setStateMoveTo(getBitFieldUnsigned(1));
			if (s->getStateMoveTo() == 1) {
				int nb = getBitFieldUnsigned(5);
				s->setMoveDeltaX(getBitFieldSigned(nb));
				s->setMoveDeltaY(getBitFieldSigned(nb));
			}
			if (s->getStateFillStyle0() == 1) {
				s->setFillStyle0(getBitFieldUnsigned(shape->getNumFillBits()));
			}
			if (s->getStateFillStyle1() == 1) {
				s->setFillStyle1(getBitFieldUnsigned(shape->getNumFillBits()));
			}
			if (s->getStateLineStyle() == 1) {
				s->setLineStyle(getBitFieldUnsigned(shape->getNumLineBits()));
			}
			if (s->getStateNewStyles() == 1) {
				UI16 n = (UI16)getByte();
				if (n == 0xff) n = getWord();	
				s->setNumFillStyles(n);
				if (n > 0) {
					FILLSTYLE* 	fillStyles = new FILLSTYLE[n];
					getFillStyles(fillStyles,n,isShape3);
					s->setFillStyles(fillStyles);
				}
				n = (UI16)getByte();
				if (n == 0xff) n = getWord();	
				s->setNumLineStyles(n);
				if (n > 0) {
					LINESTYLE* 	lineStyles = new LINESTYLE[n];
					getLineStyles(lineStyles,n, isShape3);		
					s->setLineStyles(lineStyles);
				}
				shape->setNumFillBits(getBitFieldUnsigned(4));	// FIX!!!
				shape->setNumLineBits(getBitFieldUnsigned(4));
				s->setNumFillBits(shape->getNumFillBits());
				s->setNumLineBits(shape->getNumFillBits());
			}
			shape->addShapeRecord(s);
		} else { 							// edge type
			bool straitFlag = getBitFieldUnsigned(1);
			if (straitFlag) {				// STRAIGHTEDGERECORD
				StraightEdgeRecord *s = new StraightEdgeRecord();
				s->setTypeFlag(StraightEdgeRecordType);				
				s->setStraightFlag(straitFlag);
				int nb = getBitFieldUnsigned(4)+2;
				s->setGeneralLineFlag(getBitFieldUnsigned(1));
				if (s->getGeneralLineFlag() == 1) {
					s->setDeltaX(getBitFieldSigned(nb));
					s->setDeltaY(getBitFieldSigned(nb));
				} else {
					s->setVertLineFlag(getBitFieldUnsigned(1));
					s->setDeltaVert(getBitFieldSigned(nb));
				}
				shape->addShapeRecord(s);
			} else {						// CURVEDEDGERECORD
				CurvedEdgeRecord *s = new CurvedEdgeRecord();
				s->setTypeFlag(CurvedEdgeRecordType);
				s->setStraightFlag(straitFlag);
				int nb = getBitFieldUnsigned(4)+2;
				s->setControlDeltaX(getBitFieldSigned(nb));
				s->setControlDeltaY(getBitFieldSigned(nb));
				s->setAnchorDeltaX(getBitFieldSigned(nb));
				s->setAnchorDeltaY(getBitFieldSigned(nb));
				shape->addShapeRecord(s);								
			}
		}
	}
}

void Swf_decoder::getShape(Shape *shape, bool isShape3) {
	shape->setNumFillBits(getBitFieldUnsigned(4));
	shape->setNumLineBits(getBitFieldUnsigned(4));
	getRecords(shape,isShape3);
}

void Swf_decoder::getShapeWithStyle(ShapeWithStyle *shapeWithStyle, bool isShape3) {
	UI16 n = (UI16)getByte();
	if (n == 0xff) n = getWord();
	shapeWithStyle->setNumFillStyles(n);
	if (n > 0) {
		FILLSTYLE* 	fillStyles = new FILLSTYLE[n];
		getFillStyles(fillStyles,n,isShape3);
		shapeWithStyle->setFillStyles(fillStyles);
	}
	n = (UI16)getByte();
	if (n == 0xff) n = getWord();	
	shapeWithStyle->setNumLineStyles(n);
	if (n > 0) {
		LINESTYLE* 	lineStyles = new LINESTYLE[n];
		getLineStyles(lineStyles,n,isShape3);		
		shapeWithStyle->setLineStyles(lineStyles);
	}

	shapeWithStyle->setNumFillBits(getBitFieldUnsigned(4));
	shapeWithStyle->setNumLineBits(getBitFieldUnsigned(4));
	getRecords(shapeWithStyle,isShape3);
}

void Swf_decoder::getCxForm(CXFORM *r) {
	r->hasAddTerms = getBitFieldUnsigned(1);
	r->hasMultTerms = getBitFieldUnsigned(1);
	int nb = getBitFieldUnsigned(4);
	int shift = nb < 8 ? 8-nb : nb -8;

	if (r->hasMultTerms == 1) {
		r->redAddTerm = getBitFieldSigned(nb);
		r->greenAddTerm = getBitFieldSigned(nb);
		r->blueAddTerm = getBitFieldSigned(nb);
		if (nb < 8) {
			r->redMultTerm <<= shift;
			r->greenMultTerm <<= shift;
			r->blueMultTerm <<= shift;
		} 
	}
	if (r->hasAddTerms == 1) {
		r->redMultTerm = getBitFieldSigned(nb);
		r->greenMultTerm = getBitFieldSigned(nb);
		r->blueMultTerm = getBitFieldSigned(nb);
		if (nb < 8) {
			r->redAddTerm <<= shift;
			r->greenAddTerm <<= shift;
			r->blueAddTerm <<= shift;
		} 
	}
}

void Swf_decoder::getCxFormWithAlpha(CXFORMWITHALPHA *r) {
	r->hasAddTerms = getBitFieldUnsigned(1);
	r->hasMultTerms = getBitFieldUnsigned(1);
	int nb = getBitFieldUnsigned(4);
	int shift = nb < 8 ? 8-nb : nb -8;

	if (r->hasMultTerms == 1) {
		r->redMultTerm = getBitFieldSigned(nb);
		r->greenMultTerm = getBitFieldSigned(nb);
		r->blueMultTerm = getBitFieldSigned(nb);
		r->alphaMultTerm = getBitFieldSigned(nb);
		if (nb < 8) {
			r->redMultTerm <<= shift;
			r->greenMultTerm <<= shift;
			r->blueMultTerm <<= shift;
			r->alphaMultTerm <<= shift;
		} 
	}
	if (r->hasAddTerms == 1) {
		r->redAddTerm = getBitFieldSigned(nb);
		r->greenAddTerm = getBitFieldSigned(nb);
		r->blueAddTerm = getBitFieldSigned(nb);
		r->alphaAddTerm = getBitFieldSigned(nb);
		if (nb < 8) {
			r->redAddTerm <<= shift;
			r->greenAddTerm <<= shift;
			r->blueAddTerm <<= shift;
			r->alphaAddTerm <<= shift;
		} 
	}
	alignNextByte();
}

/**
 *	Read a string and convert to unicode.
 *
 * ****BUG**** here.don't know how to handle shift-JIS for flash <= 5
 * 
 **/
void Swf_decoder::getString(STRING *s) {
	char *buff = (char *) malloc(200);
	int idx = 0, alloc = 0, size = 200;
	
	bool fin;
	do {
		char c = (char)getByte();
		fin = c == 0;
		if (!fin) {
			if (alloc == 200) {
				size+= 200;
				buff = (char*)realloc(buff,size);
				alloc = 0;
			}
			buff[idx++] = c;
			alloc ++;
		}
	} while (!fin);

	buff[idx] = 0;
	s->setString(buff,header.version >= 6);
	s->setLen(idx);
	free(buff);
}

void Swf_decoder::getClipEventFlags(ClipEventFlags *r) {
	r->clipEventKeyUp = getBitFieldUnsigned(1);
	r->clipEventKeyDown = getBitFieldUnsigned(1);
	r->clipEventMouseUp = getBitFieldUnsigned(1);
	r->clipEventMouseDown = getBitFieldUnsigned(1);
	r->clipEventMouseMove = getBitFieldUnsigned(1);
	r->clipEventUnload = getBitFieldUnsigned(1);
	r->clipEventEnterFrame = getBitFieldUnsigned(1);
	r->clipEventLoad = getBitFieldUnsigned(1);
	r->clipEventDragOver = getBitFieldUnsigned(1);
	r->clipEventRollOut = getBitFieldUnsigned(1);
	r->clipEventRollOver = getBitFieldUnsigned(1);
	r->clipEventReleaseOutside = getBitFieldUnsigned(1);
	r->clipEventRelease = getBitFieldUnsigned(1);
	r->clipEventPress = getBitFieldUnsigned(1);
	r->clipEventInitialize = getBitFieldUnsigned(1);
	r->clipEventData = getBitFieldUnsigned(1);
	if (header.version >= 6) {
		getBitFieldUnsigned(5);
		r->clipEventConstruct = getBitFieldUnsigned(1);
		r->clipEventKeyPress = getBitFieldUnsigned(1);
		r->clipEventDragOut = getBitFieldUnsigned(1);
		getBitFieldUnsigned(8);
	}
	if(debug) dumpClipEventFlags(f,r);
}

/**
 * Clip actions are only valid for placing sprite character
 **/
void Swf_decoder::getClipActions(ClipActions *s) {
	int de = getWord();			// Reserved
	if (de !=0) { if (debug) fprintf(f,"DECODE ERROR: kO Swf_decoder::getClipActions %d\n",de); }
	ClipEventFlags 	clipEventFlags;
	getClipEventFlags(&clipEventFlags);
	s->setAllEventFlags(clipEventFlags);

	bool fin;
	do {
		UI32 len;
		if (header.version <= 5) {
			len = getWord();
			fin = len == 0;
			if (!fin) posBuff -= 16;
		} else {
			len = getDWord();
			fin = len == 0;
			if (!fin) posBuff -= 32;
		}
		if (!fin) {			// Read clip actions
			ClipActionRecord *car = new ClipActionRecord();
			ClipEventFlags c ;
			getClipEventFlags(&c);
			car->setClipEventFlags(c);
			car->setActionRecordSize(getDWord());
			int alen = car->getActionRecordSize();
			if (header.version >= 6 &&c.clipEventKeyPress == 1) {
				car->setKeyCode(getByte());
			}
			UI8* code = new UI8[alen];
			for (int i=0; i< alen ; i++) code[i] = getByte();
			car->setCode(code);
			s->addActionRecord(car);
		}
	} while (!fin);
}

void Swf_decoder::getTextRecord(TEXTRECORD *t, bool isDefineText2, int glyphBits, int advanceBits) {
	getBitFieldUnsigned(1);
	getBitFieldUnsigned(3);
	t->setStyleFlagsHasFont(getBitFieldUnsigned(1));
	t->setStyleFlagsHasColor(getBitFieldUnsigned(1));
	t->setStyleFlagsHasYOffset(getBitFieldUnsigned(1));
	t->setStyleFlagsHasXOffset(getBitFieldUnsigned(1));
	if (t->getStyleFlagsHasFont()) t->setFontID(getWord());
	if (t->getStyleFlagsHasColor()) {
		RGBA c;
		c.red = getByte();
		c.green = getByte();
		c.blue  = getByte();
		c.alpha = isDefineText2 ? getByte() : 255;
		t->setTextColor(c);
	}
	if (t->getStyleFlagsHasXOffset()) t->setXOffset(getSWord()); 
	if (t->getStyleFlagsHasYOffset()) t->setYOffset(getSWord()); 
	if (t->getStyleFlagsHasFont()) t->setTextHeight(getWord());
	t->setGlyphCount(getByte());
	GLYPHENTRY *g = new GLYPHENTRY[t->getGlyphCount()];
	for (int i=0; i< t->getGlyphCount(); i++) {
		g[i].glyphIndex = getBitFieldUnsigned(glyphBits);
		g[i].glyphAdvance = getBitFieldSigned(advanceBits);
	}
	t->setGlyphEntries(g);
}

void Swf_decoder::getBUTTONRECORD(BUTTONRECORD *r, bool isDefineButton2) {
	getBitFieldUnsigned(4);
	r->buttonStateHitTest=getBitFieldUnsigned(1);
	r->buttonStateDown=getBitFieldUnsigned(1);	
	r->buttonStateOver=getBitFieldUnsigned(1);
	r->buttonStateUp=getBitFieldUnsigned(1);
	r->characterID = getWord();
	r->placeDepth=getWord();
	getMatrix(&r->placeMatrix);
	if (isDefineButton2) {
		r->hasColorTransform = true;
		getCxFormWithAlpha(&r->colorTransform); 
	} else {
		r->hasColorTransform = false;
	}
}

bool Swf_decoder::getBUTTONCONDACTION(BUTTONCONDACTION *r, int maxSize) {
	r->setCondActionSize(getWord());
	r->setCondIdleToOverDown(getBitFieldUnsigned(1));
	r->setCondOutDownToIdle(getBitFieldUnsigned(1));
	r->setCondOutDownToOverDown(getBitFieldUnsigned(1));
	r->setCondOverDownToOutDown(getBitFieldUnsigned(1));
	r->setCondOverDownToOverUp(getBitFieldUnsigned(1));
	r->setCondOverUpToOverDown(getBitFieldUnsigned(1));
	r->setCondOverUpToIdle(getBitFieldUnsigned(1));
	r->setCondIdleToOverUp(getBitFieldUnsigned(1));
	r->setCondKeyPress(getBitFieldUnsigned(7));
	r->setCondOverDownToIdle(getBitFieldUnsigned(1));

	if ((posBuff >>3)+maxSize > header.fileLength-8) maxSize = header.fileLength-8 - (posBuff >>3);
	UI8* tmpCode = (UI8*) malloc(maxSize);
	char* addr = (char*) &decodeBuffer[posBuff >>3];
	memcpy(tmpCode,addr, maxSize);
	int tmpSize = 0;
	bool cont = true;
	while (cont) {
		UI8 a = getByte();
		tmpSize ++;
		UI16 dataLen=0;
		if ( (a>>7) != 0) {
			dataLen = getWord();
			for (int i=0; i< dataLen; i++) getByte();
			tmpSize += 2+dataLen;
		} 
		cont = a != 0;
	}
	UI8* code = new UI8[tmpSize];
	for (int i=0; i< tmpSize; i++) code[i]=tmpCode[i];
	free(tmpCode);
	r->setCode(code);
	return r->getCondActionSize() != 0;
}

void Swf_decoder::getSoundInfo(SOUNDINFO *s) {
	getBitFieldUnsigned(2);
	s->setSyncStop(getBitFieldUnsigned(1));
	s->setSyncNoMultiple(getBitFieldUnsigned(1));
	s->setHasEnvelope(getBitFieldUnsigned(1));
	s->setHasLoops(getBitFieldUnsigned(1));
	s->setHasOutPoint(getBitFieldUnsigned(1));
	s->setHasInPoint(getBitFieldUnsigned(1));
	if (s->getHasInPoint()) s->setInPoint(getDWord());
	if (s->getHasOutPoint()) s->setOutPoint(getDWord());
	if (s->getHasLoops()) s->setLoopCount(getWord());
	if (s->getHasEnvelope()) {
		s->setEnvPoints(getByte());
		SOUNDENVELOPE *se = new SOUNDENVELOPE[s->getEnvPoints()];
		for (int i=0; i< s->getEnvPoints(); i++) {
			se[i].pos44 = getDWord();
			se[i].leftLevel = getWord();
			se[i].rightLevel = getWord();
		}
		s->setEnvelopeRecords(se);
	}
}


/**
 * UTILS
 * 
 **/
DefineFontTag *Swf_decoder::getFont(UI16 fontId) {
	list<BaseTag*>::iterator it;
	for (it = tags.begin(); it != tags.end() ; it++) {
		if ((*it)->getType() == stagDefineFont) {
			DefineFontTag *s = (DefineFontTag*)*it;
			if (s->getFontID() == fontId) return s;
		}
	}
    // This should never happens...
	printf("ERROR: font %d not found\n",fontId);
	return NULL;
}



/**
 *
 * TAG decoding main dispatcher: it dispartch the decoder to proper decoding function
 *
 **/
BaseTag *Swf_decoder::dispatcher(list<BaseTag*> *tagList, int len, int code){
	long startPos = posBuff>>3;
	BaseTag *b;
	switch (code) {
		case stagEnd: 				b=decodeEnd(len) ; 				break;
		case stagShowFrame: 		b=decodeShowFrame(len) ; 		break;
		case stagDefineShape: 		b=decodeDefineShape(len) ; 		break;
		case stagFreeCharacter:	 	b=decodeFreeCharacter(len) ; 	break;
		case stagPlaceObject: 		b=decodePlaceObject(len) ; 		break;
		case stagRemoveObject: 		b=decodeRemoveObject(len) ; 	break;
		case stagDefineBits: 		b=decodeDefineBits(len) ; 		break;
		case stagDefineButton: 		b=decodeDefineButton(len) ;		break;
		case stagJPEGTables: 		b=decodeJPEGTables(len) ; 		break;
		case stagSetBackgroundColor:b=decodeSetBackgroundColor(len);break;
		case stagDefineFont: 		b=decodeDefineFont(len) ; 		break;
		case stagDefineText: 		b=decodeDefineText(len) ; 		break;
		case stagDoAction: 			b=decodeDoAction(len) ; 		break;
		case stagDefineFontInfo: 	b=decodeDefineFontInfo(len) ; 	break;
		case stagDefineSound: 		b=decodeDefineSound(len) ; 		break;
		case stagStartSound: 		b=decodeStartSound(len) ; 		break;
		case stagStopSound: 		b=decodeStopSound(len) ; 		break;
		case stagDefineButtonSound: b=decodeDefineButtonSound(len); break;
		case stagSoundStreamHead: 	b=decodeSoundStreamHead(len) ;  break;
		case stagSoundStreamBlock: 	b=decodeSoundStreamBlock(len) ; break;
		case stagDefineBitsLossless:b=decodeDefineBitsLossless(len);break;
		case stagDefineBitsJPEG2: 	b=decodeDefineBitsJPEG2(len);	break;
		case stagDefineShape2: 		b=decodeDefineShape2(len) ; 	break;
		case stagDefineButtonCxform:b=decodeDefineButtonCxform(len);break;
		case stagProtect: 			b=decodeProtect(len) ;			break;
		case stagPlaceObject2: 		b=decodePlaceObject2(len) ; 	break;
		case stagRemoveObject2: 	b=decodeRemoveObject2(len) ; 	break;
		case stagDefineShape3: 		b=decodeDefineShape3(len) ; 	break;
		case stagDefineText2: 		b=decodeDefineText2(len) ; 		break;
		case stagDefineButton2: 	b=decodeDefineButton2(len) ; 	break;
		case stagDefineBitsJPEG3: 	b=decodeDefineBitsJPEG3(len) ; 	break;
		case stagDefineBitsLossless2:b=decodeDefineBitsLossless2(len);break;
		case stagDefineEditText: 	b=decodeDefineEditText(len) ; 	break;
		case stagDefineSprite: 		b=decodeDefineSprite(len) ; 	break;
		case stagNameCharacter: 	b=decodeNameCharacter(len) ; 	break;
		case stagFrameLabel: 		b=decodeFrameLabel(len) ; 		break;
		case stagSoundStreamHead2: 	b=decodeSoundStreamHead2(len) ; break;
		case stagDefineMorphShape: 	b=decodeDefineMorphShape(len) ; break;
		case stagDefineFont2: 		b=decodeDefineFont2(len) ;		break;
		case stagExportAssets: 		b=decodeExportAssets(len) ; 	break;
		case stagImportAssets: 		b=decodeImportAssets(len) ; 	break;
		case stagEnableDebugger: 	b=decodeEnableDebugger(len) ; 	break;
		case stagDoInitAction: 		b=decodeDoInitActiontag(len) ; 	break;
		case stagDefineVideoStream: b=decodeDefineVideoStream(len) ;break;
		case stagVideoFrame: 		b=decodeVideoFrame(len) ; 		break;
		case stagDefineFontInfo2: 	b=decodeDefineFontInfo2(len) ; 	break;
		case stagEnableDebugger2: 	b=decodeEnableDebugger2(len) ; 	break;
		case stagScriptLimits: 		b=decodeScriptLimits(len) ; 	break;
		case stagSetTabIndex: 		b=decodeSetTabIndex(len) ; 		break;
		default: {
			printf("TAG code non recognized %d\n",code);
			return NULL;
		}
	}		
	if (b!=NULL) {
		b->setCurSpriteId(curSpriteId);
		tagList->push_back(b);
	}
	
	// Check for buggy decoder.
	long endPos = posBuff>>3;
	if (endPos - startPos != len) {
		int diff = len - (endPos-startPos);
		printf("ERROR adjust position: %d (code:%d)\n",diff,code);
		posBuff += diff<<3;
		fflush(f);
		exit(0); // TODO: remove that when code 100% tested
		return NULL;
	}
	return b;
}

/**
  *
  * 		Now we start decoding the TAGS. 
  *
  *			There is one entry per tag type.
  *
  **/

/**
 * The end of sequence.
 **/
BaseTag *Swf_decoder::decodeEnd(int len) { 
	EndTag *res = new EndTag();
	if (curSpriteId != 0) {
		curSpriteId = 0;	// End of sprite
		return res;
	}
	if (header.fileLength-8 == posBuff>>3) puts("Decode Successfull");
	else printf("Decode interrupted by an end command. Location: %d out of %d\n",posBuff>>3,header.fileLength);
	delete res;
	return NULL;
}

/**
 * The ShowFrame tag instructs Flash Player to display the contents of the display list.
 **/
BaseTag *Swf_decoder::decodeShowFrame(int len) { 
	ShowFrameTag *res = new ShowFrameTag();
	if (debug) dumpShowFrame(f,res);
	return res;
} 

/**
 * Decode a Shape
 **/
BaseTag *Swf_decoder::decodeDefineShape(int len) { 
	DefineShapeTag *res = new DefineShapeTag();
	res->setShapeId(getWord());
	RECT rect ;
	getRect(&rect);
	res->setShapeBounds(rect);
	ShapeWithStyle *shapeWithStyle = new ShapeWithStyle();
	getShapeWithStyle(shapeWithStyle,false);
	res->setShapeWithStyle(shapeWithStyle);
	if (debug) dumpDefineShapeTag(f,res);
	return res;
} 

/**
 * ??? Not in the specs, but in old gplFlash code
 **/
BaseTag *Swf_decoder::decodeFreeCharacter(int len) { 
	puts("Decoder non implemented:decodeFreeCharacter. What's that?"); exit(0);
} 

/**
 * The PlaceObject tag adds a character to the display list. 
 **/
BaseTag *Swf_decoder::decodePlaceObject(int len) { 
	PlaceObjectTag *res = new PlaceObjectTag ();
	res->setCharacterId(getWord());
	res->setDepth(getWord());
	getMatrix(res->getMatrix());
	getCxForm(res->getColorTransform());
	if (debug) dumpPlaceObject(f,res);
	return res;
} 

/**
 * The RemoveObject tag removes the specified character (at the specified depth) from the display list
 **/
BaseTag *Swf_decoder::decodeRemoveObject(int len) { 
	printf("WARNING: tag not tested yet!\nDoes it work? (decodeRemoveObject)");
	RemoveObjectTag *res = new RemoveObjectTag();
	res->setCharacterId(getWord());
	res->setDepth(getWord());
	if (debug) dumpRemoveObject(f,res);
	return res;
} 

/**
 * This tag defines a bitmap character with JPEG compression.
 **/
BaseTag *Swf_decoder::decodeDefineBits(int len) { 
	DefineBitsTag *res = new DefineBitsTag();
	res->setCharacterID(getWord());
	int size = len - 2 ;
	UI8 *jpegData = new UI8[size];
	for (int i=0; i< size; i++) jpegData[i] = getByte();
	res->setJPEGData(jpegData);
	if (debug) dumpDefineBitsTag(f,res);
	return res;
} 

/**
 * This tag defines a button character for later use by control tags such as PlaceObject.
 **/
BaseTag *Swf_decoder::decodeDefineButton(int len) { 
	printf("WARNING: tag not tested yet!\nDoes it work? (decodeDefineButton)");
	DefineButtonTag *res = new DefineButtonTag();
	res->setButtonId(getWord());
	bool cont;
	do {
		int i = getByte();
		cont = i!=0;
		if (cont) {
			posBuff -= 8;
			BUTTONRECORD *p = new BUTTONRECORD();
			getBUTTONRECORD(p, true);
			res->addCharacters(p); 
		}
	} while (cont);
	UI8* tmpCode = (UI8*) malloc(len);
	memcpy(tmpCode,&decodeBuffer[posBuff >>3], len);
	int tmpSize = 0;
	cont = true;
	while (cont) {
		UI8 a = getByte();
		tmpSize ++;
		UI16 dataLen=0;
		if ( (a>>7) != 0) {
			dataLen = getWord();
			for (int i=0; i< dataLen; i++) getByte();
			tmpSize += 2+dataLen;
		} 
		cont = a != 0;
	}
	UI8* code = new UI8[tmpSize];
	for (int i=0; i< tmpSize; i++) code[i]=tmpCode[i];
	free(tmpCode);
	res->setCode(code);
	if (debug) dumpDefineButton(f,res);
	return res;
} 

/**
 * This tag defines the JPEG encoding table (the Tables/Misc segment) for all JPEG images 
 **/
BaseTag *Swf_decoder::decodeJPEGTables(int len) { 
	JPEGTablesTag *res = new JPEGTablesTag();
	int size = len;
	UI8 *data = new UI8[size];
	for (int i=0; i< size; i++) data[i] = getByte();
	res->setJPEGData(data);
	if (debug) dumpJPEGTables(f,res);
	return res;
} 

/**
 * The SetBackgroundColor tag sets the background color of the display.
 * The minimum file format version is SWF 1.
 **/
BaseTag *Swf_decoder::decodeSetBackgroundColor(int len) {
	SetBackgroundColorTag *res = new  SetBackgroundColorTag();
	res->setType(stagSetBackgroundColor);
	UI8 red = getByte();
	UI8 green = getByte();
	UI8 blue = getByte();
	res->setBackgroundColor(red,green,blue);
	if (debug) dumpSetBackgroundColor(f,res);
	return res;
} 

/**
 * The DefineFont tag defines the shape outlines of each glyph used in a particular font.
 **/
BaseTag *Swf_decoder::decodeDefineFont(int len) { 
	DefineFontTag *res = new DefineFontTag();
	res->setFontID(getWord());
	int offset = getWord();
	UI16* offsetTable = new UI16[offset/2];
	offsetTable[0] = offset/2;
	for (int i=1; i< offset/2; i++)
		offsetTable[i] = getWord();
	res->setOffsetTable(offsetTable);
	Shape *shape = new Shape[offset/2];
	for (int i=0; i< offset/2; i++) {
		getShape(&shape[i],false);
	}
	res->setGlyphShapeTable(shape);
	if (debug) dumpDefineFont(f,res);
	return res;
} 


/**
 * The DefineText tag defines a block of static text
 **/
BaseTag *Swf_decoder::decodeDefineText(int len) {
	DefineTextTag *res = new DefineTextTag();
	res->setCharacterID(getWord());
	RECT r;
	getRect(&r);
	res->setTextBounds(r);
	MATRIX m;
	getMatrix(&m);
	res->setTextMatrix(m);
	res->setGlyphBits(getByte());
	res->setAdvanceBits(getByte());
	bool fin = false;
	while (!fin) {
		int o = getByte();
		fin = o == 0;
		if (!fin) {
			posBuff -= 8;
			TEXTRECORD *t = new TEXTRECORD();
			getTextRecord(t,false,res->getGlyphBits(),res->getAdvanceBits());
			res->addTextRecord(t);
		}
	}
	if (debug) dumpDefineText(f,res);
	return res;
} 

/**
 * Process action tags
 **/
BaseTag *Swf_decoder::decodeDoAction(int len) { 
	DoActiontag *res = new DoActiontag();
	UI8* code = new UI8[len];
	for (int i=0; i< len ; i++) code[i] = getByte();
	res->setCode(code);
	if (debug) dumpAction(f,res);
	return res;
}

/**
 * The DefineFontInfo tag defines a mapping from a glyph font (defined with DefineFont) to a device Font
 **/ 
BaseTag *Swf_decoder::decodeDefineFontInfo(int len) { 
	DefineFontInfoTag * res = new DefineFontInfoTag();
	res->setFontID(getWord());
	int sz = getByte();
	char *name = (char*) malloc(sz+1);
	for (int i=0; i< sz; i++) name[i] = getByte();
	name[sz] = 0;
	res->setFontName(name);
	getBitFieldUnsigned(2);
	res->setFontFlagsSmallText(getBitFieldUnsigned(1));
	res->setFontFlagsShiftJIS(getBitFieldUnsigned(1));
	res->setFontFlagsANSI(getBitFieldUnsigned(1));
	res->setFontFlagsItalic(getBitFieldUnsigned(1));
	res->setFontFlagsBold(getBitFieldUnsigned(1));
	res->setFontFlagsWideCodes(getBitFieldUnsigned(1));
	DefineFontTag *d = getFont(res->getFontID());
	UI16 *codeTable = new UI16[d->getNGlyphs()];
	for (int i=0; i< d->getNGlyphs(); i++) {
		codeTable[i] = res->getFontFlagsWideCodes() ? getWord() : getByte();
	}
	res->setCodeTable(codeTable);
	res->setCodeTableLen(d->getNGlyphs());
	if (debug)dumpDefineFontInfo(f,res);
	return res;
} 

/**
 * The DefineSound tag defines an event sound.
 **/
BaseTag *Swf_decoder::decodeDefineSound(int len) { 
	DefineSoundTag *res = new DefineSoundTag();
	res->setSoundId(getWord());
	res->setSoundFormat(getBitFieldUnsigned(4));
	res->setSoundRate(getBitFieldUnsigned(2));
	res->setSoundSize(getBitFieldUnsigned(1));
	res->setSoundType(getBitFieldUnsigned(1));
	res->setSoundSampleCount(getDWord());
	int size = len - 7;
	UI8 *data = new UI8[size];
	for (int i=0; i< size; i++) data[i] = getByte();
	res->setSoundData(data);
	res->setSoundDataSize(size);
	if (debug) dumpDefineSound(f,res);
	return res;
} 

/**
 * StartSound is a control tag that either starts (or stops) playing a sound defined by DefineSound.
 **/
BaseTag *Swf_decoder::decodeStartSound(int len) {
	StartSoundTag *res = new StartSoundTag();
	res->setSoundId(getWord());
	SOUNDINFO *s = new SOUNDINFO();
	getSoundInfo(s);
	res->setSoundInfo(s);
	if (debug) dumpStartSound(f,res);
	return res;
} 

/**
 * Not in the spec but in old gplFlash code.
 * TODO: ??
 **/
BaseTag *Swf_decoder::decodeStopSound(int len) { 
	puts("Decoder non implemented:decodeStopSound");
	return NULL;
} 

/**
 * The DefineButtonSound tag defines which sounds (if any) are played on state transitions.
 **/
BaseTag *Swf_decoder::decodeDefineButtonSound(int len) { 
	DefineButtonSoundTag *res = new DefineButtonSoundTag();
	res->setButtonId(getWord());
	res->setButtonSoundChar0(getWord());
	if (res->getButtonSoundChar0() != 0) {
		SOUNDINFO *s = new SOUNDINFO();
		getSoundInfo(s);
		res->setButtonSoundInfo0(s);
	}
	res->setButtonSoundChar1(getWord());
	if (res->getButtonSoundChar1() != 0) {
		SOUNDINFO *s = new SOUNDINFO();
		getSoundInfo(s);
		res->setButtonSoundInfo1(s);
	}
	res->setButtonSoundChar2(getWord());
	if (res->getButtonSoundChar2() != 0) {
		SOUNDINFO *s = new SOUNDINFO();
		getSoundInfo(s);
		res->setButtonSoundInfo2(s);
	}
	res->setButtonSoundChar3(getWord());
	if (res->getButtonSoundChar3() != 0) {
		SOUNDINFO *s = new SOUNDINFO();
		getSoundInfo(s);
		res->setButtonSoundInfo3(s);
	}
	if (debug) dumpDefineButtonSound(f,res);
	return res;
}
 
/**
 * Sound data streaming
 **/ 
BaseTag *Swf_decoder::decodeSoundStreamHead(int len) { 
	SoundStreamHeadTag *res = new SoundStreamHeadTag();
	getBitFieldUnsigned(4);
	res->setPlaybackSoundRate(getBitFieldUnsigned(2));
	res->setPlaybackSoundSize(getBitFieldUnsigned(1));
	res->setPlaybackSoundType(getBitFieldUnsigned(1));
	res->setStreamSoundCompression(getBitFieldUnsigned(4));
	res->setStreamSoundRate(getBitFieldUnsigned(2));
	res->setStreamSoundSize(getBitFieldUnsigned(1));
	res->setStreamSoundType(getBitFieldUnsigned(1));
	res->setStreamSoundSampleCount(getWord());
	if (res->getStreamSoundCompression() ==2) res->setLatencySeek(getSWord());
	if (debug) dumpSoundStreamHead(f,res);
	return res;
}

/**
 *  The SoundStreamBlock tag defines sound data that is interleaved with frame data so that sounds
 *  can be played as the movie is streamed over a network connection.
 **/
BaseTag *Swf_decoder::decodeSoundStreamBlock(int len) {
	SoundStreamBlockTag *res = new SoundStreamBlockTag();
	// TODO: call decode library
	int size = len ;
	UI8 *data = new UI8[size];
	for (int i=0; i< size; i++) data[i] = getByte();
	res->setStreamSoundData(data);
	if (debug) dumpSoundStreamBlock(f,res);
	return res;
} 

/**
 * Defines a lossless bitmap character that contains RGB bitmap data compressed with ZLIB
 *
 * Info: the bitmap is uncompressed. Should be decompressed at run time (buildFromZlibData).
 **/
BaseTag *Swf_decoder::decodeDefineBitsLossless(int len) {
	DefineBitsLosslessTag *res = new DefineBitsLosslessTag();
	res->setCharacterID(getWord());
	res->setBitmapFormat(getByte()); 
	res->setBitmapWidth(getWord()); 
	res->setBitmapHeight(getWord()); 
	if (res->getBitmapFormat() == 3) res->setBitmapColorTableSize(getByte());

	// TODO: call decode library
	int size = len - 7 - (res->getBitmapFormat() ? 1 : 0);
	UI8 *data = new UI8[size];
	for (int i=0; i< size; i++) data[i] = getByte();
	res->setZlibBitmapData(data) ;
	if (debug) dumpDefineBitsLossless(f,res);
	return res;
}
 
/**
 * This tag defines a bitmap character with JPEG compression.
 **/
BaseTag *Swf_decoder::decodeDefineBitsJPEG2(int len) { 
	DefineBitsJPEG2Tag *res = new DefineBitsJPEG2Tag();
	res->setCharacterID(getWord());
	//TODO: decode jpeg data 
	int size = len - 2 ;
	UI8 *jpegData = new UI8[size];
	for (int i=0; i< size; i++) jpegData[i] = getByte();
	res->setJPEGData(jpegData);
	if (debug) dumpDefineBitsJPEG2(f,res);
	return res;
} 

/**
 * DefineShape2 extends the capabilities of DefineShape with the ability to support more than 255
 *  styles in the style list and multiple style lists in a single shape.
 **/
BaseTag *Swf_decoder::decodeDefineShape2(int len) {
	DefineShapeTag *res = new DefineShapeTag();
	res->setShapeId(getWord());
	RECT rect;
	getRect(&rect);
	res->setShapeBounds(rect);
	ShapeWithStyle *shapeWithStyle = new ShapeWithStyle();
	getShapeWithStyle(shapeWithStyle,false);
	res->setShapeWithStyle(shapeWithStyle);
	if (debug) dumpDefineShapeTag2(f,res);
	return res;
}

/**
 * Defines the color transform for each shape and text character in a button. 
 **/
BaseTag *Swf_decoder::decodeDefineButtonCxform(int len) { 
	printf("WARNING: tag not tested yet!\nDoes it work? (decodeDefineButtonCxform)");
	DefineButtonCxformTag *res = new DefineButtonCxformTag();
	res->setButtonId(getWord());
	CXFORM c;
	getCxForm(&c);
	res->setButtonColorTransform(c);
	if (debug) dumpDefineButtonCxform(f,res);
	return res;
} 

/**
 * The Protect tag marks a file as not importable for editing in an authoring environment.
 **/
BaseTag *Swf_decoder::decodeProtect(int len) {
	ProtectTag *res = new ProtectTag();
	if (len !=0) {
		UI16 i = getWord();
		STRING *s = new STRING();
		getString(s);
		res->setPassword(s);
	}
	if (debug) dumpProtect(f,res);
	return res;
}

/**
 * PlaceObject2 can both add a character to the display list, and modify the attributes of a character that is already on the
 * display list. The PlaceObject2 tag changed slightly from Flash 4 to Flash 5. In Flash 5 clip actions were added.
 **/ 
BaseTag *Swf_decoder::decodePlaceObject2(int len) {
	PlaceObject2Tag *res = new PlaceObject2Tag ();
	res->setPlaceFlagHasClipActions(getBitFieldUnsigned(1));
	res->setPlaceFlagHasClipDepth(getBitFieldUnsigned(1));
	res->setPlaceFlagHasName(getBitFieldUnsigned(1));
	res->setPlaceFlagHasRatio(getBitFieldUnsigned(1));
	res->setPlaceFlagHasColorTransform(getBitFieldUnsigned(1));
	res->setPlaceFlagHasMatrix(getBitFieldUnsigned(1));
	res->setPlaceFlagHasCharacter(getBitFieldUnsigned(1));
	res->setPlaceFlagMove(getBitFieldUnsigned(1));
	res->setDepth(getWord());
	if (res->getPlaceFlagHasCharacter()) res->setCharacterId(getWord());
	if (res->getPlaceFlagHasMatrix()) getMatrix(res->getMatrix());
	if (res->getPlaceFlagHasColorTransform()) getCxFormWithAlpha(res->getColorTransform());
	if (res->getPlaceFlagHasRatio()) res->setRatio(getWord());
	if (res->getPlaceFlagHasName()) {
		STRING *s = new STRING();
		getString(s);
		res->setName(s);
	}
	if (res->getPlaceFlagHasClipDepth()) res->setClipDepth(getWord());
	if (res->getPlaceFlagHasClipActions()) {
		ClipActions *clipActions = new ClipActions();
		getClipActions(clipActions);
		res->setClipActions(clipActions);
	}
	if (debug) dumpPlaceObject2(f,res);
	return res;
}
 
/**
 * The RemoveObject2 tag removes the character at the specified depth from the display list.
 **/
BaseTag *Swf_decoder::decodeRemoveObject2(int len) { 
	RemoveObject2Tag *res= new RemoveObject2Tag();
	res->setDepth(getWord());
	if (debug) dumpRemoveObject2(f,res);
	return res;
} 

/**
 * DefineShape3 extends the capabilities of DefineShape2 by extending all of the RGB color fields
 *  to support RGBA with alpha transparency.
 **/
BaseTag *Swf_decoder::decodeDefineShape3(int len) {
	DefineShapeTag *res = new DefineShapeTag();
	res->setShapeId(getWord());
	RECT rect;
	getRect(&rect);
	res->setShapeBounds(rect);
	ShapeWithStyle *shapeWithStyle = new ShapeWithStyle();
	getShapeWithStyle(shapeWithStyle,true);
	res->setShapeWithStyle(shapeWithStyle);
	if (debug) dumpDefineShapeTag(f,res);
	return res;
} 

/**
 * Extends DefineText with RGBA
 **/
BaseTag *Swf_decoder::decodeDefineText2(int len) {
	DefineText2Tag *res = new DefineText2Tag();
	res->setCharacterID(getWord());
	RECT r;
	getRect(&r);
	res->setTextBounds(r);
	MATRIX m;
	getMatrix(&m);
	res->setTextMatrix(m);
	res->setGlyphBits(getByte());
	res->setAdvanceBits(getByte());
	bool fin = false;
	while (!fin) {
		int o = getByte();
		fin = o == 0;
		if (!fin) {
			posBuff -= 8;
			TEXTRECORD *t = new TEXTRECORD();
			getTextRecord(t,true,res->getGlyphBits(),res->getAdvanceBits());
			res->addTextRecord(t);
		}
	}
	if (debug) dumpDefineText2(f,res);
	return res;
}
 
/**
 * Text defined with DefineText2 is always rendered with glyphs
 **/
BaseTag *Swf_decoder::decodeDefineButton2(int len) {
	DefineButton2Tag *res = new DefineButton2Tag();
	res->setButtonId(getWord());
	getBitFieldUnsigned(7);
	res->setTrackAsMenu(getBitFieldUnsigned(1));
	res->setActionOffset(getWord());
	bool cont;
	do {
		int i = getByte();
		cont = i!=0;
		if (cont) {
			posBuff -= 8;
			BUTTONRECORD *p = new BUTTONRECORD();
			getBUTTONRECORD(p, true);
			res->addCharacters(p); 
		}
	} while (cont);
	if (res->getActionOffset() != 0) {
		do {
			BUTTONCONDACTION *p = new BUTTONCONDACTION();
			cont = getBUTTONCONDACTION(p,len);
			res->addAction(p); 
		} while (cont);
	}
	if (debug) dumpDefineButton2(f,res);
	return res;
}
 
/**
 * Defines a bitmap character with JPEG compression (alpha added)
 **/
BaseTag *Swf_decoder::decodeDefineBitsJPEG3(int len) {
	DefineBitsJPEG3Tag *res = new DefineBitsJPEG3Tag();
	res->setCharacterID(getWord());
	res->setAlphaDataOffset(getDWord());
	int size = len - 6;
	UI8 *jpegData = new UI8[size];
	for (int i=0; i< size; i++) jpegData[i] = getByte();
	res->setJPEGData(jpegData);
	if (debug) dumpDefineBitsJPEG3(f,res);
	return res;
} 

/**
 * DefineBitsLossless2 extends DefineBitsLossless with support for transparency (alpha values).
 **/
BaseTag *Swf_decoder::decodeDefineBitsLossless2(int len) {
	DefineBitsLossless2Tag *res = new DefineBitsLossless2Tag();
	res->setCharacterID(getWord());
	res->setBitmapFormat(getByte());
	res->setBitmapWidth(getWord());
	res->setBitmapHeight(getWord());
	if (res->getBitmapFormat()==3) res->setBitmapColorTableSize(getByte());
	int size = len - 7 - (res->getBitmapFormat()==3 ? 1 : 0);
	res->setSize(size);
	UI8 *data = new UI8[size];
	for (int i=0; i< size; i++) data[i] = getByte();
	res->setZlibBitmapData(data);
	if(debug) dumpDefineBitsLossless2(f,res);
	return res;
} 

/**
 * The DefineEditText tag defines a dynamic text object, or text field.
 **/
BaseTag *Swf_decoder::decodeDefineEditText(int len) {
	DefineEditTextTag *res = new DefineEditTextTag();
	res->setCharacterID(getWord());
	RECT b ;
	getRect(&b);
	res->setBounds(b);
	res->setHasText(getBitFieldUnsigned(1));
	res->setWordWrap(getBitFieldUnsigned(1));
	res->setMultiline(getBitFieldUnsigned(1));
	res->setPassword(getBitFieldUnsigned(1));
	res->setReadOnly(getBitFieldUnsigned(1));
	res->setHasTextColor(getBitFieldUnsigned(1));
	res->setHasMaxLength(getBitFieldUnsigned(1));
	res->setHasFont(getBitFieldUnsigned(1));
	getBitFieldUnsigned(1); // reserved
	res->setAutoSize(getBitFieldUnsigned(1));
	res->setHasLayout(getBitFieldUnsigned(1));
	res->setNoSelect(getBitFieldUnsigned(1));
	res->setBorder(getBitFieldUnsigned(1));
	getBitFieldUnsigned(1); // reserved
	res->setHTML(getBitFieldUnsigned(1));
	res->setUseOutlines(getBitFieldUnsigned(1));
	if (res->getHasFont()) {
		res->setFontID(getWord());
		res->setFontHeight(getWord());
	}
	if (res->getHasTextColor()) {
		RGBA a;
		a.red = getByte();
		a.green = getByte();
		a.blue = getByte();
		a.alpha = getByte();
		res->setTextColor(a);
	}
	if (res->getHasMaxLength()) res->setMaxLength(getWord());
	if (res->getHasLayout()) {
		res->setAlign(getByte());
		res->setLeftMargin(getWord());
		res->setRightMargin(getWord());
		res->setIndent(getWord());
		res->setLeading(getWord());
	}
	STRING *s = new STRING();
	getString(s);
	res->setVariableName(s);
	if (res->getHasText()) {
		STRING *s2  = new STRING();
		getString(s2);
		res->setInitialText(s2);
	}
	if (debug) dumpDefineEditTextTag(f,res);
	return res;
} 

/**
 * The DefineSprite tag defines a sprite character.
 **/
BaseTag *Swf_decoder::decodeDefineSprite(int len) { 
	long startPos = posBuff>>3;
	DefineSpriteTag *res = new DefineSpriteTag();
	res->setSpriteID(getWord());
	res->setFrameCount(getWord());
	list<BaseTag*> *tagList = res->getTags();
	if (debug) dumpDefineSprite(f,res);
	// Follows a serie of tags
	bool cont;
	curSpriteId = res->getSpriteID();
	do {
		UI16 tag = getWord() ;
   		long ilen = tag & 0x3f;
   	 	long code = tag >> 6;
    	if (ilen == 0x3f)  ilen = getDWord();
		if (debug) fprintf(f,"Sprite: id:%d  dec code:%d, taglen:%d  DecodePos:%d/%d\n",curSpriteId,code,ilen,posBuff>>3,header.fileLength-8);
  		BaseTag *curTag = dispatcher(tagList,ilen,code);
		cont = curTag !=NULL && curTag->getType() != stagEnd;
	} while (cont);

	// Check for buggy files
	long endPos = posBuff>>3;
	if (endPos - startPos != len) {
		int diff = len - (endPos-startPos);
		printf("WARNING adjust position: %d DefineSprite\n",diff);
		posBuff += diff<<3;
		fflush(f);
		return NULL;
	}

	curSpriteId = 0;
	if (debug) fprintf(f,"\n\n\n\n\n");
	return res;
} 

/**
 * TODO: find the spec
 * Buggy???
 **/
BaseTag *Swf_decoder::decodeNameCharacter(int len) {
	NameCharacterTag *res = new NameCharacterTag();
	res->setTagId(getWord());	
	STRING *s = new STRING();
	getString(s);
	res->setName(s);
	if (debug) dumpNameCharacter(f,res);
	return res;
} 

/**
 * The FrameLabel tag gives the specified Name to the current frame
 **/
BaseTag *Swf_decoder::decodeFrameLabel(int len) {
	FrameLabelTag *res = new FrameLabelTag();
	STRING *s = new STRING();
	getString(s);
	res->setName(s);
	if (debug) dumpFrameLabel(f,res);
	return res;
}

/**
 * Sound data streaming
 **/ 
BaseTag *Swf_decoder::decodeSoundStreamHead2(int len) { 
	SoundStreamHead2Tag *res = new SoundStreamHead2Tag();
	getBitFieldUnsigned(4);
	res->setPlaybackSoundRate(getBitFieldUnsigned(2));
	res->setPlaybackSoundSize(getBitFieldUnsigned(1));
	res->setPlaybackSoundType(getBitFieldUnsigned(1));
	res->setStreamSoundCompression(getBitFieldUnsigned(4));
	res->setStreamSoundRate(getBitFieldUnsigned(2));
	res->setStreamSoundSize(getBitFieldUnsigned(1));
	res->setStreamSoundType(getBitFieldUnsigned(1));
	res->setStreamSoundSampleCount(getWord());
	if (res->getStreamSoundCompression() ==2) res->setLatencySeek(getSWord());
	if (debug) dumpSoundStreamHead2(f,res);
	return res;
} 

/**
 * The DefineMorphShape tag defines the start and end states of a morph sequence
 **/
BaseTag *Swf_decoder::decodeDefineMorphShape(int len) { 
	DefineMorphShapeTag *res = new DefineMorphShapeTag();
	long startPos = posBuff>>3;

	res->setCharacterID(getWord());

	getRect(res->getStartBounds());
	getRect(res->getEndBounds());
	res->setOffset(getDWord());
	MORPHFILLSTYLES *ms = new MORPHFILLSTYLES();
	UI16 n = getByte(); 
	if (n == 0xff) n = getWord();
	ms->setFillStyleCount(n);
	MORPHFILLSTYLE *m = new MORPHFILLSTYLE[n];
	getMorphFillStyles(m,n);
	ms->setFillStyles(m);
	res->setMorphFillStyles(ms);
	MORPHLINESTYLES *ml = new MORPHLINESTYLES();
	n = getByte(); 
	if (n == 0xff) n = getWord();
	ml->setLineStyleCount(n);
	MORPHLINESTYLE *l = new MORPHLINESTYLE[n];
	getMorphLineStyles(l,n);
	ml->setLineStyles(l);
	res->setMorphLineStyles(ml);
	Shape *s = new Shape();
	getShape(s, false);
	res->setStartEdges(s);
	s = new Shape();
	getShape(s, false);
	res->setEndEdges(s);
	if (debug) dumpDefineMorphShape(f,res);

	// The tag may be empty: deleted?
	if (res->getOffset() ==0) {
		long endPos = posBuff>>3;
		int diff = len - (endPos-startPos);
		//printf("WARNING ajust position: %d\n",diff);
		posBuff += diff<<3;	
	}
	return res;
}

/**
 * The DefineFont2 tag extends the functionality of DefineFont
 **/
BaseTag *Swf_decoder::decodeDefineFont2(int len) { 
	DefineFont2Tag *res = new DefineFont2Tag();
	res->setFontID(getWord());
	res->setFontFlagsHasLayout(getBitFieldUnsigned(1));
	res->setFontFlagsShiftJIS(getBitFieldUnsigned(1));
	res->setFontFlagsSmallText(getBitFieldUnsigned(1));
	res->setFontFlagsANSI(getBitFieldUnsigned(1));
	res->setFontFlagsWideOffsets(getBitFieldUnsigned(1));
	res->setFontFlagsWideCodes(getBitFieldUnsigned(1));
	res->setFontFlagsItalic(getBitFieldUnsigned(1));
	res->setFontFlagsBold(getBitFieldUnsigned(1));
	res->setLanguageCode(getByte());
	res->setFontNameLen(getByte());
	UI8 *fName = new UI8[res->getFontNameLen()+1];
	for (int i=0; i< res->getFontNameLen(); i++) fName[i] = getByte();
	fName[res->getFontNameLen()] = 0;
	res->setFontName(fName);
	res->setNumGlyphs(getWord());
	UI32 *offsetTable = res->getNumGlyphs()>0 ? new UI32[res->getNumGlyphs()] : NULL;
	for (int i=0; i< res->getNumGlyphs(); i++) {
		offsetTable[i] = res->getFontFlagsWideOffsets() ? getDWord() : getWord();
	}
	res->setOffsetTable(offsetTable);
	res->setCodeTableOffset(res->getFontFlagsWideOffsets() ? getDWord() : getWord());
	Shape *shape = res->getNumGlyphs()>0 ? new Shape[res->getNumGlyphs()] : NULL;
	for (int i=0; i< res->getNumGlyphs(); i++)
		getShape(&shape[i],false);
	res->setGlyphShapeTable(shape);	
	UI16 *codeTable = res->getNumGlyphs()>0 ? new UI16[res->getNumGlyphs()] : NULL;
	for (int i=0; i< res->getNumGlyphs(); i++)
		codeTable[i] = res->getFontFlagsWideCodes() ? getWord() : getByte();
	res->setCodeTable(codeTable);
	if (res->getFontFlagsHasLayout()) {
		res->setFontAscent(getSWord());
		res->setFontDescent(getSWord());
		res->setFontLeading(getSWord());
		SI16 *fontAdvanceTable = new SI16[res->getNumGlyphs()];
		for (int i=0; i< res->getNumGlyphs(); i++) fontAdvanceTable[i] = getSWord();
		res->setFontAdvanceTable(fontAdvanceTable);
		RECT * fontBoundsTable = new RECT[res->getNumGlyphs()];
		for (int i=0; i< res->getNumGlyphs(); i++) getRect(&fontBoundsTable[i]);
		res->setFontBoundsTable(fontBoundsTable);
		res->setKerningCount(getWord());
		KERNINGRECORD *kr = new KERNINGRECORD[res->getKerningCount()];
		for (int i=0; i< res->getKerningCount(); i++) {
			kr[i].fontKerningCode1 = res->getFontFlagsWideCodes() ? getWord() : getByte();
			kr[i].fontKerningCode2 = res->getFontFlagsWideCodes() ? getWord() : getByte();
			kr[i].fontKerningAdjustment = getSWord();
		}
		res->setFontKerningTable(kr);
	}
	if (debug) dumpDefineFont2(f,res);
	return res;
}

/**
 *  ExportAssets makes portions of a SWF file available for import by other SWF files
 **/
BaseTag *Swf_decoder::decodeExportAssets(int len) { 
	ExportAssetsTag *res = new ExportAssetsTag();
	res->setCount(getWord());
	UI16 *p = new UI16[res->getCount()];
	STRING **s = (STRING**) malloc(res->getCount() * sizeof(STRING*));
	for (int i=0; i< res->getCount(); i++) {
		p[i] = getWord();
		s[i] = new STRING();
		getString(s[i]);
	}
	res->setTags(p);
	res->setNames(s);	
	if (debug) dumpExportAssets(f,res);
	return res;
} 

/**
 * The ImportAssets tag imports characters from another SWF file.
 **/
BaseTag *Swf_decoder::decodeImportAssets(int len) { 
	ImportAssetsTag *res = new ImportAssetsTag();
	STRING *s = new STRING();
	getString(s);
	res->setURL(s);
	res->setCount(getWord());
	UI16 *tags = new UI16[res->getCount()];
	STRING *names = new STRING[res->getCount()];
	for (int i=0; i< res->getCount(); i++) {
		tags[i] = getWord();
		getString(&names[i]);
	}
	res->setTags(tags);
	res->setNames(names);
	if (debug) dumpImportAssets(f,res);
	return res;
}
 
/**
 * The EnableDebugger tag enables debugging. 
 **/
BaseTag *Swf_decoder::decodeEnableDebugger(int len) { 
	EnableDebuggerTag *res = new EnableDebuggerTag();
	STRING *s = new STRING();
	getString(s);
	res->setPassword(s);
	if (debug) dumpEnableDebugger(f,res);
	return res;
} 

/**
 * The DoInitAction defines a series of bytecodes to be executed
 **/
BaseTag *Swf_decoder::decodeDoInitActiontag(int len) {
	DoInitActionTag *res = new DoInitActionTag();
	res->setSpriteID(getWord());
	UI8* code = new UI8[len-2];
	for (int i=0; i< len-2 ; i++) code[i] = getByte();
	res->setCode(code);
	if (debug) dumpDoInitAction(f,res);
	return res;
} 

/**
 * Defines a video character which can later be placed on the display list 
 **/
BaseTag *Swf_decoder::decodeDefineVideoStream(int len) {
	DefineVideoStreamTag *res = new DefineVideoStreamTag();
	res->setCharacterID(getWord());
	res->setNumFrames(getWord());
	res->setWidth(getWord());
	res->setHeight(getWord());
	getBitFieldUnsigned(5);
	res->setVideoFlagsDeblocking(getBitFieldUnsigned(2));
	res->setVideoFlagsSmoothing(getBitFieldUnsigned(1));
	res->setCodecID(getByte());
	if (debug) dumpDefineVideoStream(f,res);
	return res;
} 

/**
 * Provides a single frame of video data for a video character that has already been defined with DefineVideoStream.
 **/
BaseTag *Swf_decoder::decodeVideoFrame(int len) { 
	VideoFrameTag *res = new VideoFrameTag();
	res->setStreamID(getWord());
	res->setFrameNum(getWord());
	int size = len - 4;
	UI8 *data = new UI8[size];
	for (int i=0; i< size; i++) data[i] = getByte();
	res->setVideoData(data);
	res->setSize(size);
	if (debug) dumpVideoFrame(f,res);
	return res;
} 

/**
 * Extends decodeDefineFontInfo by adding LanguageCode
 **/
BaseTag *Swf_decoder::decodeDefineFontInfo2(int len) { 
	DefineFontInfo2Tag * res = new DefineFontInfo2Tag();
	res->setFontID(getWord());
	int sz = getByte();
	char *name = (char*) malloc(sz+1);
	for (int i=0; i< sz; i++) name[i] = getByte();
	name[sz] = 0;
	res->setFontName(name);
	getBitFieldUnsigned(2);
	res->setFontFlagsSmallText(getBitFieldUnsigned(1));
	res->setFontFlagsShiftJIS(getBitFieldUnsigned(1));
	res->setFontFlagsANSI(getBitFieldUnsigned(1));
	res->setFontFlagsItalic(getBitFieldUnsigned(1));
	res->setFontFlagsBold(getBitFieldUnsigned(1));
	res->setFontFlagsWideCodes(getBitFieldUnsigned(1));
	res->setLanguageCode(getByte());
	DefineFontTag *d = getFont(res->getFontID());
	UI16 *codeTable = new UI16[d->getNGlyphs()];
	for (int i=0; i< d->getNGlyphs(); i++) {
		codeTable[i] = res->getFontFlagsWideCodes() ? getWord() : getByte();
	}
	res->setCodeTable(codeTable);
	res->setCodeTableLen(d->getNGlyphs());
	if (debug) dumpDefineFontInfo2(f,res);
	return res;
} 

/**
 * The EnableDebugger2 tag enables debugging.
 **/
BaseTag *Swf_decoder::decodeEnableDebugger2(int len) { 
	printf("WARNING: tag not tested yet!\nDoes it work? (decodeEnableDebugger2)");
	EnableDebugger2Tag *res = new EnableDebugger2Tag();
	getWord();
	STRING *s = new STRING();
	getString(s);
	res->setPassword(s);
	if (debug) dumpEnableDebugger2(f,res);
	return res;
} 

/**
 * Maximum recursion depth and ActionScript time-out
 **/
BaseTag *Swf_decoder::decodeScriptLimits(int len) {
	printf("WARNING: tag not tested yet!\nDoes it work? (decodeScriptLimits)");
	ScriptLimitsTag *res = new ScriptLimitsTag();
	res->setMaxRecursionDepth(getWord());
	res->setScriptTimeoutSeconds(getWord());
	if (debug) dumpScriptLimits(f,res);
	return res;
} 

/**
  * Flash Player maintains a concept of tab order of the interactive and textual objects displayed.
  **/
BaseTag *Swf_decoder::decodeSetTabIndex(int len) {
	SetTabIndexTag *res = new SetTabIndexTag();
	res->setDepth(getWord());
	res->setTabIndex(getWord());
	if (debug) dumpSetTabIndex(f,res);
	return res;
} 
