/**
Macromedia Flash File decoder class header.
Copyright (C) 2004  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#ifndef _SWF_DECODER_
#define _SWF_DECODER_
#include "common.hh"
#include "swf.hh"
#include "swf_tags.hh"
#include "zlib.h"

class Swf_decoder {
public:
	bool	readStream(char *data, int len);
	HEADER* getHeader() { return &header; }
	list<BaseTag*> *getTagList() { return &tags; }
	~Swf_decoder() { 
		list<BaseTag*>::iterator it;
		for (it = tags.begin(); it != tags.end(); it++) 	delete (*it);
		tags.clear(); 
	}
	Swf_decoder() {
		needHeader = true;
	}

private:
	char 	*decodeBuffer;
	long 	posBuff;
	HEADER 	header;
	list<BaseTag*> tags;
	int 	curSpriteId;
	FILE    *f;
	bool	needHeader;
	z_stream zstream;
	long	zstreamInPos;
	long	m_lastSize;

	// Decoder helpers
	UI8 	getByte(void);
	UI16 	getWord(void);
	UI32 	getDWord(void);
	void 	alignNextByte();
	UI32 	getBitFieldUnsigned(int length);
	SI32 	getBitFieldSigned(int length);
	SI32	getSWord();
	float 	get8p8(void);
	BaseTag	*dispatcher(list<BaseTag*> *tags, int len, int code);
	void 	getRect(RECT *r);
	void 	getMatrix(MATRIX *m);
	void	getGradient(GRADIENT *g, bool isShape3);
	void	getMorphGradient(MORPHGRADIENT *g);
	void 	getFillStyles(FILLSTYLE* f, int n, bool isShape3);
	void	getMorphFillStyles(MORPHFILLSTYLE* f, int n);
	void	getLineStyles(LINESTYLE* l, int n, bool isShape3);
	void	getMorphLineStyles(MORPHLINESTYLE* l, int n);
	void	getShape(Shape *shape, bool isShape3);
	void 	getShapeWithStyle(ShapeWithStyle *shapeWithStyle, bool isShape3);
	void 	getRecords(Shape *shape, bool isShape3);
	void 	getCxForm(CXFORM *r);
	void	getCxFormWithAlpha(CXFORMWITHALPHA *r);
	void 	getString(STRING *s);
	void	getClipEventFlags(ClipEventFlags *r);
	void 	getClipActions(ClipActions *s);
	void	getTextRecord(TEXTRECORD *t, bool isDefineText2,int glyphBits, int advanceBits);
	void	getBUTTONRECORD(BUTTONRECORD *r, bool isDefineButton2);
	bool	getBUTTONCONDACTION(BUTTONCONDACTION *r, int maxLen);
	void	getSoundInfo(SOUNDINFO *s);

	// Dump methods
	void 	dumpHeader(FILE *);
	void 	dumpRect(FILE *, RECT *r);
	void 	dumpString(FILE *f , STRING *s);
	void 	dumpAction(FILE *f, DoActiontag *r);
	void 	dumpDefineShapeTag(FILE *f, DefineShapeTag *r);
	void 	dumpDefineShapeTag2(FILE *f, DefineShapeTag *r);
	void	dumpShapeWithStyle(FILE *f, ShapeWithStyle *r);
	void 	dumpShape(FILE *f, Shape *r);
	void 	dumpFillStyle(FILE *f, FILLSTYLE *r);
	void	dumpMorphFillStyle(FILE *f, MORPHFILLSTYLE *r);
	void 	dumpLineStyle(FILE *f, LINESTYLE *r);
	void	dumpMorphLineStyle(FILE *f, MORPHLINESTYLE *r);
	void 	dumpRecord(FILE *f, ShapeRecord *r);
	void	dumpMatrix(FILE *f, MATRIX *r);
	void 	dumpGradient(FILE *f, GRADIENT *r);
	void	dumpMorphGradient(FILE *f, MORPHGRADIENT *r) ;
	void	dumpCxForm(FILE *f, CXFORM *s);
	void	dumpCxFormWithAlpha(FILE *f, CXFORMWITHALPHA *s);
	void 	dumpPlaceObject(FILE *f, PlaceObjectTag *r);
	void 	dumpPlaceObject2(FILE *f, PlaceObject2Tag *r);
	void 	dumpDefineFont(FILE *f, DefineFontTag *r);
	void 	dumpDefineFontInfo(FILE *, DefineFontInfoTag *r);
	void	dumpDefineFontInfo2(FILE *f, DefineFontInfo2Tag *r);
	void	dumpTextRecord(FILE *f, TEXTRECORD *r);
	void 	dumpDefineText(FILE *f, DefineTextTag *r);
	void	dumpDefineSprite(FILE* f,DefineSpriteTag *r);
	void	dumpRemoveObject2(FILE *f, RemoveObject2Tag *r);
	void	dumpDefineBitsLossless(FILE *f,DefineBitsLosslessTag *res);
	void	dumpDefineFont2(FILE *f, DefineFont2Tag*r);
	void	dumpDefineEditTextTag(FILE *f, DefineEditTextTag *res);
	void	dumpBUTTONRECORD(FILE *f,BUTTONRECORD *p);
	void	dumpBUTTONCONDACTION(FILE *f,BUTTONCONDACTION *p);
	void	dumpDefineButton(FILE *f,DefineButtonTag *r);
	void	dumpDefineButton2(FILE *f,DefineButton2Tag *r);
	void	dumpFrameLabel(FILE *f, FrameLabelTag *r);
	void	dumpExportAssets(FILE *f, ExportAssetsTag *r);
	void	dumpDoInitAction(FILE *f, DoInitActionTag *r);
	void	dumpClipActions(FILE *f, ClipActions *r);
	void	dumpClipEventFlags(FILE *f, ClipEventFlags *r);
	void	dumpSetTabIndex(FILE* f,SetTabIndexTag *r);
	void	dumpShowFrame(FILE *f, ShowFrameTag *r);
	void	dumpDefineBitsLossless2(FILE *f,DefineBitsLossless2Tag *r);
	void	dumpDefineBits(FILE *f, DefineBitsTag *r);
	void	dumpDefineBitsJPEG2(FILE *f, DefineBitsJPEG2Tag *r);
	void	dumpDefineBitsJPEG3(FILE *f, DefineBitsJPEG3Tag *r);
	void	dumpSoundStreamHead(FILE *f, SoundStreamHeadTag *r);
	void	dumpSoundStreamHead2(FILE *f, SoundStreamHead2Tag *r);
	void	dumpDefineSound(FILE *f,DefineSoundTag *r);
	void	dumpSoundInfo(FILE *f, SOUNDINFO *s);
	void	dumpDefineButtonSound(FILE *f, DefineButtonSoundTag *r);
	void	dumpStartSound(FILE *f, StartSoundTag *r);
	void	dumpJPEGTables(FILE *f, JPEGTablesTag*r);
	void	dumpDefineBitsTag(FILE *f,DefineBitsTag *r);
	void	dumpDefineMorphShape(FILE *f, DefineMorphShapeTag *r);
	void	dumpSoundStreamBlock(FILE *f, SoundStreamBlockTag* r);
	void	dumpProtect(FILE *f , ProtectTag *r);
	void	dumpNameCharacter(FILE *f, NameCharacterTag *r);
	void	dumpRemoveObject(FILE *f, RemoveObjectTag *r);
	void	dumpSetBackgroundColor(FILE *f, SetBackgroundColorTag *r);
	void	dumpDefineButtonCxform(FILE *f, DefineButtonCxformTag *r);
	void	dumpDefineText2(FILE *f , DefineText2Tag *r);
	void	dumpImportAssets(FILE *f, ImportAssetsTag *r);
	void	dumpEnableDebugger(FILE *f, EnableDebuggerTag *r);
	void	dumpEnableDebugger2(FILE *f, EnableDebugger2Tag *r);
	void	dumpDefineVideoStream(FILE *f, DefineVideoStreamTag *r);
	void	dumpVideoFrame(FILE *f , VideoFrameTag *r);
	void	dumpScriptLimits(FILE *f, ScriptLimitsTag *r);
	void	dumpActionCode(FILE *f, UI8* code);

	// utils
	DefineFontTag *getFont(UI16 fontId);

	// Tag decode methods
	BaseTag *  decodeEnd(int len) ;
	BaseTag *  decodeShowFrame(int len) ; 
	BaseTag *  decodeDefineShape(int len) ; 
	BaseTag *  decodeFreeCharacter(int len) ; 
	BaseTag *  decodePlaceObject(int len) ; 
	BaseTag *  decodeRemoveObject(int len) ; 
	BaseTag *  decodeDefineBits(int len) ; 
	BaseTag *  decodeDefineButton(int len) ; 
	BaseTag *  decodeJPEGTables(int len) ; 
	BaseTag *  decodeSetBackgroundColor(int len) ; 
	BaseTag *  decodeDefineFont(int len) ; 
	BaseTag *  decodeDefineText(int len) ; 
	BaseTag *  decodeDoAction(int len) ; 
	BaseTag *  decodeDefineFontInfo(int len) ; 
	BaseTag *  decodeDefineSound(int len) ; 
	BaseTag *  decodeStartSound(int len) ; 
	BaseTag *  decodeStopSound(int len) ; 
	BaseTag *  decodeDefineButtonSound(int len) ; 
	BaseTag *  decodeSoundStreamHead(int len) ; 
	BaseTag *  decodeSoundStreamBlock(int len) ; 
	BaseTag *  decodeDefineBitsLossless(int len) ; 
	BaseTag *  decodeDefineBitsJPEG2(int len) ; 
	BaseTag *  decodeDefineShape2(int len) ; 
	BaseTag *  decodeDefineButtonCxform(int len) ; 
	BaseTag *  decodeProtect(int len) ; 
	BaseTag *  decodePlaceObject2(int len) ; 
	BaseTag *  decodeRemoveObject2(int len) ; 
	BaseTag *  decodeDefineShape3(int len) ; 
	BaseTag *  decodeDefineText2(int len) ; 
	BaseTag *  decodeDefineButton2(int len) ; 
	BaseTag *  decodeDefineBitsJPEG3(int len) ; 
	BaseTag *  decodeDefineBitsLossless2(int len) ; 
	BaseTag *  decodeDefineEditText(int len) ; 
	BaseTag *  decodeDefineSprite(int len) ; 
	BaseTag *  decodeNameCharacter(int len) ; 
	BaseTag *  decodeFrameLabel(int len) ; 
	BaseTag *  decodeSoundStreamHead2(int len) ; 
	BaseTag *  decodeDefineMorphShape(int len) ; 
	BaseTag *  decodeDefineFont2(int len) ; 
	BaseTag *  decodeExportAssets(int len) ; 
	BaseTag *  decodeImportAssets(int len) ; 
	BaseTag *  decodeEnableDebugger(int len) ; 
	BaseTag *  decodeDoInitActiontag(int len) ; 
	BaseTag *  decodeDefineVideoStream(int len) ; 
	BaseTag *  decodeVideoFrame(int len) ; 
	BaseTag *  decodeDefineFontInfo2(int len) ; 
	BaseTag *  decodeEnableDebugger2(int len) ; 
	BaseTag *  decodeScriptLimits(int len) ; 
	BaseTag *  decodeSetTabIndex(int len) ; 
};
#endif
