/**
Macromedia Flash Texture management.
Copyright (C) 2004  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "texture.hh"
#include "maths.hh"

/**
 * The constructor
 **/
Texture::Texture() { 
	generatedTexSize = 256;
	glColorTexture = 0;
	nbColorEntries = 1;
}

/**
 * The destructor
 **/
Texture::~Texture() {
	textures.clear();
	texScaleX.clear();
	texScaleY.clear();
	texDimX.clear();
	texDimY.clear();
	ramps.clear();
	colorEntries.clear();
	for (map<GLuint,RGBA*>::const_iterator it= textureCache.begin(); it != textureCache.end(); ++it) delete []it->second;
	textureCacheDimx.clear();
	textureCacheDimy.clear();
}

/**
 * Creates a Gradient Linear texture
 **/
void Texture::registerLinearGradient(int characterId, int fillIndex, FILLSTYLE *f) {
	char name[100];
	sprintf(name,"SHAPE:%d_%d",characterId,fillIndex);
    int totalMemory = generatedTexSize ;
    RGBA* textureMemory = new RGBA[totalMemory];

	for (int i=0; i< generatedTexSize; i++) {
		float radius = 255.0 * float(i)/float(generatedTexSize);
		RGBA a,b;
		float ratio1 = 0, ratio2 = 0;
		bool stop = false;
		list<GRADRECORD*>::iterator it;
		int cpt=0;
		for (it = f->getGradient()->getGradientRecords()->begin(); it != f->getGradient()->getGradientRecords()->end() && !stop; it++) {
			GRADRECORD *d = *it;
			b = a;
			ratio2 = ratio1;
			a = *d->getColor();
			ratio1 = d->getRatio(); 
			if (ratio1 > radius) stop = true;
			cpt++;
		}
		if (cpt==1) b=a;
		float t = ratio1==ratio2 ? 1 :(radius-ratio2)/(ratio1-ratio2);
		if (ratio1 < radius) t=1;
		int red   = int(a.red*t + (1-t)*b.red);	
		int green = int(a.green*t + (1-t)*b.green);					
		int blue  = int(a.blue*t + (1-t)*b.blue);					
		int alpha = int(a.alpha*t + (1-t)*b.alpha);					
		textureMemory[i].red = red;
		textureMemory[i].green = green;
		textureMemory[i].blue = blue;
		textureMemory[i].alpha = alpha;
	}
    GLuint textID;
    glGenTextures( 1, (GLuint*)&textID);

    glBindTexture( GL_TEXTURE_1D, textID);
    glEnable(GL_TEXTURE_1D);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage1D( GL_TEXTURE_1D, 0, GL_RGBA, generatedTexSize, 0, GL_RGBA, GL_UNSIGNED_BYTE, textureMemory);
    glBindTexture(GL_TEXTURE_1D, 0);
    glDisable(GL_TEXTURE_1D);
    ramps[name] = textID;
	if (hasPixelShaderSupport) {
		delete []textureMemory;
	} else {
		textureCacheDimx[textID] =  generatedTexSize;
		textureCacheDimy[textID] =  0; 
		textureCache[textID] = textureMemory;
	}
}

/**
 * Create a Gradient Radial texture
 **/
void Texture::registerRadialGradient(int characterId, int fillIndex, FILLSTYLE *f) {
	char name[100];
	sprintf(name,"SHAPE:%d_%d",characterId,fillIndex);
    int totalMemory = generatedTexSize;
    RGBA* textureMemory = new RGBA[totalMemory];

	for (int i=0; i< generatedTexSize; i++) {
		float radius = 255.0 * float(i)/float(generatedTexSize);
		RGBA a,b;
		float ratio1=0,ratio2=0;
		bool stop = false;
		int cpt=0;
		list<GRADRECORD*>::iterator it;
		for (it = f->getGradient()->getGradientRecords()->begin(); it != f->getGradient()->getGradientRecords()->end() && !stop; it++) {
			GRADRECORD *d = *it;
			b = a;
			ratio2 = ratio1;
			a = *d->getColor();
			ratio1 = d->getRatio(); 
			if (ratio1 > radius) stop = true;
			cpt++;
		}
		if (cpt==1) b=a;
		float t =  ratio1==ratio2 ? 1 : (radius-ratio2)/(ratio1-ratio2);
		if (ratio1 < radius) t=1;
		int red   = int(a.red*t + (1-t)*b.red);	
		int green = int(a.green*t + (1-t)*b.green);					
		int blue  = int(a.blue*t + (1-t)*b.blue);					
		int alpha = int(a.alpha*t + (1-t)*b.alpha);
		textureMemory[i].red = red;
		textureMemory[i].green = green;
		textureMemory[i].blue = blue;
		textureMemory[i].alpha = alpha;
	}
    GLuint textID;
    glGenTextures( 1, (GLuint*)&textID);

    glBindTexture( GL_TEXTURE_1D, textID);
    glEnable(GL_TEXTURE_1D);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage1D( GL_TEXTURE_1D, 0, GL_RGBA, generatedTexSize, 0, GL_RGBA, GL_UNSIGNED_BYTE, textureMemory);
    glBindTexture(GL_TEXTURE_1D, 0);
    glDisable(GL_TEXTURE_1D);
    ramps[name] = textID;
	if (hasPixelShaderSupport) {
		delete []textureMemory;
	} else {
		textureCacheDimx[textID] =  generatedTexSize;
		textureCacheDimy[textID] =  0; 
		textureCache[textID] = textureMemory;
	}
}


/**
 * Register a Bitmap texture
 **/
GLuint Texture::registerCharacterBitmap(int characterId, Image *img) {
	char name[100];
	sprintf(name,"CHARACTER:%d_0",characterId);
	int width = nextPowerOf2(img->getWidth());
	int height = nextPowerOf2(img->getHeight());    
	//printf("Texture size: %dx%d\n",width,height);
	int totalMemory = width*height;
    RGBA* textureMemory = new RGBA[totalMemory];
	for (int i=0; i< width; i++) {
		for (int j=0; j< height; j++) {
			if (i< img->getWidth() && j<img->getHeight()) {
				UI8 p = img->getPixels()[j*img->getWidth()+i];
				RGBA c = img->getColormap()[p];
				textureMemory[j*width+i].red   = c.red;
				textureMemory[j*width+i].green = c.green;
				textureMemory[j*width+i].blue  = c.blue;
				textureMemory[j*width+i].alpha = c.alpha;
			} else {
				textureMemory[j*width+i].red   = 0;
				textureMemory[j*width+i].green = 0;
				textureMemory[j*width+i].blue  = 0;
				textureMemory[j*width+i].alpha = 0;
			}
		}
	}
    GLuint textID;
    glGenTextures( 1, (GLuint*)&textID);
    glBindTexture(GL_TEXTURE_2D, textID);
    glEnable(GL_TEXTURE_2D);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, textureMemory);
	glBindTexture(GL_TEXTURE_2D, 0);
    glDisable(GL_TEXTURE_2D);
    delete []textureMemory;
    textures[name] = textID;
	texScaleX[name] =  width; 
	texScaleY[name] =  height;
	return textID;
}

/**
 * Register a Bitmap texture
 **/
GLuint Texture::registerShapeBitmap(int shapeId, Image *img) {
	char name[100];
	sprintf(name,"SHAPE:%d",shapeId);
	int width = nextPowerOf2(img->getWidth());
	int height = nextPowerOf2(img->getHeight());    
	int totalMemory = width*height;
    RGBA* textureMemory = new RGBA[totalMemory];
	for (int j=0; j< height; j++) {
		UI8* pixels = &img->getPixels()[j*img->getWidth()];
		memcpy(&textureMemory[j*width],pixels,img->getWidth()*sizeof(RGBA));
	}

    GLuint textID;
    glGenTextures( 1, (GLuint*)&textID);
    glBindTexture(GL_TEXTURE_2D, textID);
    glEnable(GL_TEXTURE_2D);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, textureMemory);
	glBindTexture(GL_TEXTURE_2D, 0);
    glDisable(GL_TEXTURE_2D);

    textures[name] = textID;
	texScaleX[name] =  float(img->getWidth())/float(width); 
	texScaleY[name] =  float(img->getHeight())/float(height);
	if (hasPixelShaderSupport) {
		delete []textureMemory;
	} else {
		textureCacheDimx[textID] =  width;
		textureCacheDimy[textID] =  height; 
		textureCache[textID] = textureMemory;
	}
	return textID;
}

/**
 * Register a an external Bitmap texture.
 * The OpenGL bitmap is created when the surface is created (the shape is processed).
 **/
void Texture::registerExternalBitmap(int bitmapId, Image *img) {
	char name[100];
	sprintf(name,"BITMAP:%d",bitmapId);
	int width = nextPowerOf2(img->getWidth());
	int height = nextPowerOf2(img->getHeight());    
	int totalMemory = width*height;
    RGBA* textureMemory = new RGBA[totalMemory];
	for (int i=0; i< width; i++) {
		for (int j=0; j< height; j++) {
			if (i< img->getWidth() && j<img->getHeight()) {
				UI8 p = img->getPixels()[j*img->getWidth()+i];
				RGBA c = img->getColormap()[p];
				textureMemory[j*width+i].red   = c.red;
				textureMemory[j*width+i].green = c.green;
				textureMemory[j*width+i].blue  = c.blue;
				if (img->getAlphaBuf() == NULL)	textureMemory[j*width+i].alpha = c.alpha;
				else textureMemory[j*width+i].alpha = img->getAlphaBuf()[j*img->getWidth()+i];
			} else {
				textureMemory[j*width+i].red   = 0;
				textureMemory[j*width+i].green = 0;
				textureMemory[j*width+i].blue  = 0;
				textureMemory[j*width+i].alpha = 0;
			}
		}
	}
	GLuint textID;
	glGenTextures( 1, (GLuint*)&textID);
	glBindTexture(GL_TEXTURE_2D, textID);
	glEnable(GL_TEXTURE_2D);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, textureMemory);
	glBindTexture(GL_TEXTURE_2D, 0);
    glDisable(GL_TEXTURE_2D);
	ramps[name] = textID;
	texScaleX[name] =  width; 
	texScaleY[name] =  height;
	texDimX[name] =  width; 
	texDimY[name] =  height;

	// If we have Pixel shader support, we don't need to get the texture value to apply color transformations.
	// For non pixel sharer enabled graphic cards, we keep the texture in CPU memory, and apply the color transform manually.
	if (hasPixelShaderSupport) {
		delete []textureMemory;
	} else {
		textureCacheDimx[textID] =  width;
		textureCacheDimy[textID] =  height; 
		textureCache[textID] = textureMemory;
	}
}

/**
 *  Allocate a texture for local fonts
 **/
GLuint Texture::allocateTexture(int textureWidth, int textureHeight) {   
    int totalMemory = textureWidth * textureHeight;
    unsigned char* textureMemory = new unsigned char[totalMemory];
    memset( textureMemory, 0, totalMemory);

    GLuint textID;
    glGenTextures( 1, (GLuint*)&textID);
    glBindTexture( GL_TEXTURE_2D, textID);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D( GL_TEXTURE_2D, 0, GL_ALPHA, textureWidth, textureHeight, 0, GL_ALPHA, GL_UNSIGNED_BYTE, textureMemory);
    glDisable(GL_TEXTURE_2D);

    delete [] textureMemory;
	//printf("Allocate texture %dX%d - 8bits ->id:%d\n",textureWidth, textureHeight,textID);
    return textID;
}

/**
 *  Allocate a texture for Video Rendering (RGB)
 **/
GLuint Texture::allocateTexture24(int textureWidth, int textureHeight) {   
	//printf("Allocate texture %dX%d - 24bits\n",textureWidth, textureHeight);
    int totalMemory = textureWidth * textureHeight * 3;
    unsigned char* textureMemory = new unsigned char[totalMemory];
    memset( textureMemory, 0xff, totalMemory);

    GLuint textID;
    glGenTextures( 1, (GLuint*)&textID);
    glBindTexture( GL_TEXTURE_2D, textID);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, textureWidth, textureHeight, 0, GL_RGB, GL_UNSIGNED_BYTE, textureMemory);
    glDisable(GL_TEXTURE_2D);

    delete [] textureMemory;
    return textID;
}

/**
 * To allow full material & ligthing effect (CxFormWithAlpha) colors are  handled as texture coordinates.
 **/
int Texture::registerColor(RGBA *c) {
	if (glColorTexture ==0) {
	    int totalMemory = 256*256*4;
	    unsigned char* textureMemory = new unsigned char[totalMemory];
	    memset( textureMemory, 0, totalMemory);	
	    glGenTextures( 1, (GLuint*)&glColorTexture);
	    glBindTexture( GL_TEXTURE_2D, glColorTexture);
	    glEnable(GL_TEXTURE_2D);
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, 256, 256, 0, GL_RGBA, GL_UNSIGNED_BYTE, textureMemory);
	    glDisable(GL_TEXTURE_2D);
	    delete [] textureMemory;
	}
	UI32 index = (((((c->red<<8)+c->green)<<8)+c->blue)<<8)+c->alpha;
	if (colorEntries[index] == 0) {
		//printf("%d  (%d,%d,%d,%d) Tid:%d\n",nbColorEntries,c->red,c->green,c->blue,c->alpha,glColorTexture);
		colorEntries[index] = nbColorEntries;
		invColorEntries[nbColorEntries] = index;
	    glEnable(GL_TEXTURE_2D);
		glPushClientAttrib( GL_CLIENT_PIXEL_STORE_BIT);
		glPixelStorei( GL_UNPACK_LSB_FIRST, GL_FALSE);
		glPixelStorei( GL_UNPACK_ROW_LENGTH, 0);
		glPixelStorei( GL_UNPACK_ALIGNMENT, 1);
		glBindTexture( GL_TEXTURE_2D, glColorTexture);
		glTexSubImage2D( GL_TEXTURE_2D, 0, nbColorEntries%256, nbColorEntries/256, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE, c);
	    glDisable(GL_TEXTURE_2D);
		glPopClientAttrib();
		nbColorEntries ++;
	}	
	return colorEntries[index];
}

/**
 * Transfer video data to texture
 **/
void Texture::video2Texture(int textureId, int dimx, int dimy, char *data) {
    glEnable(GL_TEXTURE_2D);
	glPushClientAttrib( GL_CLIENT_PIXEL_STORE_BIT);
	glPixelStorei( GL_UNPACK_LSB_FIRST, GL_FALSE);
	glPixelStorei( GL_UNPACK_ROW_LENGTH, 0);
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1);
	glBindTexture( GL_TEXTURE_2D, textureId);
	glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, dimx, dimy, GL_RGB, GL_UNSIGNED_BYTE, data);
    glDisable(GL_TEXTURE_2D);
}

/**
 * Returns a color (non texture) by it's index
 **/
RGBA Texture::getColorByIndex(int index) { 
	unsigned int color = invColorEntries[index]; 
	RGBA t;
	t.red   = (color>>24)&0xff;
	t.green = (color>>16)&0xff;
	t.blue  = (color>>8)&0xff;
	t.alpha = color&0xff;
	return t;
}

/**
 * get gradient parameters
 **/
GLuint Texture::getGradient(int characterId, int fillIndex) { 	
	char name[100];
	sprintf(name,"SHAPE:%d_%d",characterId,fillIndex);
	if (ramps.find(name) != ramps.end()) return ramps[name];
	else return 0;
}

/**
 * Get external texture parameters
 **/
GLuint Texture::getExternalTexture(int bitmapId,float &scaleX,float &scaleY,int &dimx, int &dimy) {
	char name[100];
	sprintf(name,"BITMAP:%d",bitmapId);
	if (ramps.find(name) != ramps.end()) {
		scaleX = texScaleX[name];
		scaleY = texScaleY[name];
		dimx = texDimX[name];
		dimy = texDimY[name];
		return ramps[name];
	}
	else return 0;
}

/**
 * Get external texture parameters
 **/
GLuint Texture::getCharacterBitmap(int characterId, float &scaleX, float &scaleY) { 	
	char name[100];
	sprintf(name,"CHARACTER:%d_0",characterId);
	scaleX = texScaleX[name];
	scaleY = texScaleY[name];
	return textures[name];
}

/**
 * Get Shape texture parameters
 **/
GLuint Texture::getShapeBitmap(int shapeId, float &scaleX, float &scaleY) { 	
	char name[100];
	sprintf(name,"SHAPE:%d",shapeId);
	scaleX = texScaleX[name];
	scaleY = texScaleY[name];
	return textures[name];
}

/**
 * Get bitmap texture parameters
 **/
bool Texture::bitmapDefined(int bitmapId) {
	char name[100];
	sprintf(name,"BITMAP:%d",bitmapId);
	return ramps.find(name) != ramps.end();
}

/**
 * Get character texture parameters
 **/
bool Texture::characterDefined(int characterId) {
	char name[100];
	sprintf(name,"CHARACTER:%d_0",characterId);
	return textures[name] != 0;
}

/**
 * set if character defined
 **/
void Texture::setCharacterDefined(int characterId) {
	char name[100];
	sprintf(name,"CHARACTER:%d_0",characterId);
	textures[name] = 1;
}

/**
 * Returns the color of a Bitmap texture point
 **/
RGBA Texture::getColorByTextureCoords(int texNum, float x, float y) {
	int ix = int(x*textureCacheDimx[texNum]);
	int iy = int(y*textureCacheDimy[texNum]);
	RGBA *texture = textureCache[texNum];
	RGBA res = texture[iy*textureCacheDimx[texNum]+ix];
	return res;
}
