/**
Macromedia Flash Texture management.
Copyright (C) 2004  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#ifndef _TEXTURE_H_
#define _TEXTURE_H_
#include "common.hh"
#include "swf.hh"
#include "swf_tags.hh"
#include <GL/glx.h>
#include <GL/glu.h>
#include <map>

typedef stdext::hash_map<string, GLuint, HashString, HashStringCompare> StringBufferHashMap;

class Texture {
public:
	Texture();
	~Texture();
	void   	registerLinearGradient(int characterId, int fillIndex, FILLSTYLE *f);
	void   	registerRadialGradient(int characterId, int fillIndex, FILLSTYLE *f);
	GLuint 	registerCharacterBitmap(int characterId, Image *img);
	GLuint 	registerShapeBitmap(int shapeId, Image *img);
	void   	registerExternalBitmap(int shapeId, Image *img);
	GLuint 	allocateTexture(int textureWidth, int textureHeight);
	GLuint 	allocateTexture24(int textureWidth, int textureHeight);
	void   	removeTexture(GLuint textureId) { glDeleteTextures(1,&textureId); }
	void   	video2Texture(int textureId, int dimx, int dimy, char *data);
	int	   	registerColor(RGBA *c);
	int	   	getColorTexture() { return glColorTexture; }
	int	   	getColorTextureIndex(RGBA *c) { return colorEntries[(((((c->red<<8)+c->green)<<8)+c->blue)<<8)+c->alpha]; }
	RGBA   	getColorByIndex(int index);
	RGBA 	getColorByTextureCoords(int texNum, float x, float y);
	GLuint 	getGradient(int characterId, int fillIndex);
	GLuint 	getExternalTexture(int bitmapId,float &scaleX,float &scaleY,int &dimx, int &dimy);
	GLuint 	getCharacterBitmap(int characterId, float &scaleX, float &scaleY);
	GLuint 	getShapeBitmap(int shapeId, float &scaleX, float &scaleY);
	bool	bitmapDefined(int bitmapId);
	bool 	characterDefined(int characterId);
	void 	setCharacterDefined(int characterId);
	void	setHasPixelShaderSupport(bool p) { hasPixelShaderSupport = p; }
private:
	ASCIIStringIntHashMap textures, texDimX, texDimY;
	ASCIIStringFloatHashMap texScaleX, texScaleY;
	int 	generatedTexSize;
	StringBufferHashMap ramps;
	GLuint	glColorTexture;
	IntIntHashMap	colorEntries, invColorEntries;
	int 	nbColorEntries;
	map<GLuint,RGBA*>	textureCache;
	map<GLuint,int>	textureCacheDimx,textureCacheDimy;
	bool	hasPixelShaderSupport;
};
#endif
