/**
gplflash Player entry point.
Copyright (C) 2005

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "../lib/common.hh"
#include "../lib/swf.hh"
#include "../lib/swf_decoder.hh"
#include "../lib/swf_engine.hh"
#include "../lib/localFonts.hh"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <GL/glx.h>
#include <X11/cursorfont.h>
#include <sys/stat.h>

#define SCALE (1.0/20.0)
#define VALGRIND false

static int attributeList_noFSAA[] = { GLX_RGBA, GLX_DOUBLEBUFFER, GLX_STENCIL_SIZE, 1, None };
static int attributeList_FSAA[] = { GLX_RGBA, GLX_DOUBLEBUFFER, GLX_STENCIL_SIZE, 1, GLX_SAMPLE_BUFFERS_ARB, 1,GLX_SAMPLES_ARB, 1, None };

static int currentX11Cursor;
int 	newX11Cursor;
bool debug = true;

/**
 * Opens the  file, get its property and call the decoder
 **/
int openDecodeStream(char *fileName, Swf_decoder * decoder) {
	struct stat buf;
	lstat(fileName, &buf);
	FILE *f = fopen(fileName,"rb");
	if (f == NULL) {
		wcout << "FILE NOT FOUND!";
		return false;
	}
	char *buffer = (char*) malloc(buf.st_size);
	fread(buffer,buf.st_size,1,f);
	fclose(f);
	int status = decoder->readStream(buffer,buf.st_size);
	free(buffer);
	return status;
}

/**
 * Open X11 rendering window & process event loop
 **/
void render(Swf_decoder * decoder,bool normalSpeed) {
	Display *dpy;
	XVisualInfo *vi;
	Colormap cmap;
	XSetWindowAttributes swa;
	Window win;
	GLXContext cx;
	XEvent event;

	dpy = XOpenDisplay(0);
	vi = glXChooseVisual(dpy, DefaultScreen(dpy), attributeList_FSAA);
	if (vi == NULL) vi = glXChooseVisual(dpy, DefaultScreen(dpy), attributeList_noFSAA);
	cx = glXCreateContext(dpy, vi, 0, GL_TRUE);
	cmap = XCreateColormap(dpy, RootWindow(dpy, vi->screen), vi->visual, AllocNone);
	swa.colormap = cmap;
	swa.border_pixel = 0;
	swa.event_mask = StructureNotifyMask | ExposureMask | KeyPressMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask;
	HEADER *header = decoder->getHeader();
	float dimx = (header->frameSize.xmax-header->frameSize.xmin);
	float dimy = (header->frameSize.ymax-header->frameSize.ymin);
	win = XCreateWindow(dpy, RootWindow(dpy, vi->screen), 0, 0,
				 int(dimx*SCALE), int(dimy*SCALE),
                 0, vi->depth, InputOutput, vi->visual,
                 CWBorderPixel|CWColormap|CWEventMask, &swa);
    XMapWindow(dpy, win);
	currentX11Cursor = XC_left_ptr;

    glXMakeCurrent(dpy, win, cx);
	glViewport(0, 0,int(dimx*SCALE),int(dimy*SCALE));
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0,dimx,dimy,0,1,-1);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

	SwfEngine *engine = new SwfEngine(decoder);

	int frame = 1;
	engine->next();
	bool redraw = true;
	int timeBetweenFrames = int(1000.0/header->frameRate);

	struct timeval now;
	gettimeofday(&now,0);
	long nowl = now.tv_sec *1000 + now.tv_usec/1000;
	int timePos = 0;

#if VALGRIND
	for (int i=0; i< 50; i++) {
#else
    while(1) {
#endif
		if (XPending(dpy)) {
	       	XNextEvent(dpy, &event); 
			switch (event.type) { 
	        	case KeyPress: {
					KeySym key;
    	            XLookupString((XKeyEvent *)&event, NULL, 0, &key, NULL); 
					char code[5];
					XLookupString((XKeyEvent *)&event, code, 5, NULL, NULL); 
					//if (key == 65361) { if (frame>0) frame--; engine->goToMainTrackFrame(frame); }
					//else if (key == 65363) { frame++; engine->next(); }
					if (key == 65307) exit(0);
					else engine->keyboardCallback(key,code);
					redraw = true;
					break;
				}
				case ButtonPress: {
					int x = event.xbutton.x;
					int y = event.xbutton.y;
					// convert to twips
					x = int (x / SCALE);
					y = int (y / SCALE);
					redraw = engine->mouseClickCallback(x,y,MouseActionPress);
					break;
				}
			    case ButtonRelease: {
					int x = event.xbutton.x;
					int y = event.xbutton.y;
					// convert to twips
					x = int (x / SCALE);
					y = int (y / SCALE);
					redraw = engine->mouseClickCallback(x,y,MouseActionRelease);
					break;
				}
				case MotionNotify: {
					int x = event.xbutton.x;
					int y = event.xbutton.y;
					// convert to twips
					x = int (x / SCALE);
					y = int (y / SCALE);
					redraw = engine->mouseClickCallback(x,y,MouseActionOver);
					break;
				}
				case Expose: {
					redraw = true;
					break;
				}
			}
		}
		if (normalSpeed && timePos > timeBetweenFrames ) { frame++; engine->next(); redraw = true; timePos-=timeBetweenFrames; }
		if (redraw) {
			engine->display();
			glXSwapBuffers(dpy, win);
			redraw = false;
		}
		if (currentX11Cursor != newX11Cursor) { 		// X11 Cursor change detected
			Cursor c = XCreateFontCursor (dpy, newX11Cursor);
			XDefineCursor (dpy, win, c);
			currentX11Cursor = newX11Cursor;
		}
		usleep(1000*10);	// 10 ms loop.
		struct timeval now;
		gettimeofday(&now,0);
		int tmp = now.tv_sec *1000 + now.tv_usec/1000;
		timePos += tmp - nowl;
		nowl = tmp;
	}
	delete engine;
	delete decoder;
	free(vi);
}


/**
 * Entry point
 **/
int main(int argc, char ** argv) {
	bool normalSpeed = false;
	char *fileName;
	if (argc > 1) {
		fileName = argv[1];
		if (argc > 2) normalSpeed = true;
	} else {
		cout << "Usage: flashplayer fileName [autoPlay]\n";
		exit(0);
	}
	
	Swf_decoder * decoder = new Swf_decoder();
	int ret = openDecodeStream(fileName, decoder);
	if (ret == 0) {
		cout << "Decoding fails!\n. Bye.";
	}
	render(decoder,normalSpeed);
	return 0;
}
