/**
Plugin interface. Mozilla / Konqueror.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#ifndef _FLASH_INTERFACE_HH_
#define _FLASH_INTERFACE_HH_

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/Core.h>
#include <X11/CoreP.h>

#define FLASH_VERSION_STRING "Version 0.99.0"
#define PLUGIN_NAME "Shockwave Flash"

#if defined(__cplusplus) || defined(c_plusplus)
extern "C" {
#endif
extern int newX11Cursor;

typedef void *FlashHandle;
typedef struct {
	FlashHandle flashHandle;
	const char *data;
	int 	len;
	int		offset;
} tNSBufferDescriptor;

FlashHandle flashNew();
int		addBytesToDecoder(void *data);
int		setWindow(FlashHandle flashHandle,Window win,Display* dpy);
void	flashTest(FlashHandle flashHandle);
int		getTimerInterval(FlashHandle flashHandle);
void	*flashRender(void*data);
void	closeFlash(FlashHandle flashHandle);
int		mouseClickCallback(FlashHandle flashHandle,int x, int y,int mouseAction);
void	keyboardCallback(FlashHandle flashHandle,KeySym key, char*code);
//void FlashSetGetUrlMethod(FlashHandle flashHandle, void (*getUrl)(char *, char *, void *), void *);
//void FlashSetGetSwfMethod(FlashHandle flashHandle, void (*getSwf)(char *url, int level, void *clientData), void *clientData);

#if defined(__cplusplus) || defined(c_plusplus)
};
#endif

#if defined(__cplusplus) || defined(c_plusplus) 	// C++ definition
class PluginInterface {
public:
	PluginInterface();
	~PluginInterface();
	int		addBytesToDecoder(const char *data,int offset, int len);
	int 	setWindow(Window win,Display* dpy);
	void	test();
	int		getTimerInterval() { return timeBetweenFrames; }
	void	flashRender();
	SwfEngine *getSwfEngine() { return engine; }
	int		mouseClickCallback(int x, int y,int mouseAction);
	void	keyboardCallback(int key, char *code);

private:
	Swf_decoder * decoder;
	Window 		win;
	Display* 	dpy;
	GLXContext  context;
	int 		dimx, dimy;
	char		*internalBuffer;
	int			internalBufferSize;
	int			timeBetweenFrames;
	SwfEngine 	*engine;
	bool		renderInProgress;
	HEADER * 	header;

	void 		initGraphic();
};

#endif

#endif // _FLASH_INTERFACE_HH_
